/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/

#include "qar.h"

extern "C" QplugReturnStatus QarVECPROD(QplugArgList *);

QplugReturnStatus QarVECPROD  ( QplugArgList *call_list )

{
QplugReturnStatus	status;
int qar_status;

QdObject_var in1_ptr;
QdObject_var in2_ptr;
QdObject_var  out_ptr;

char * attr_ptr;
char * string_buf;
char * out_name;

status = QPLUG_SUCCESS;


//	Unpack the input data objects

 in1_ptr = (*call_list)[0];

 in2_ptr = (*call_list)[1];


// Get name of output object

 char * slot_ptr = GetSlotText(0);
 if ( slot_ptr[0] != '\0' ) {
    out_name = QplugNewSTR(slot_ptr);
 }
 else{
    out_name = QplugNewSTR("VectorProduct");
    SetSlotText(0,out_name);
 }
  
// Test validity of inputs...

qar_status =  QarTestConformal(in1_ptr, in2_ptr);
  if(qar_status == QAR_ERR)  
  {
    QplugAppendTextDisplay ("Arithmetic>> input sequences not joined properly\n");
     return QPLUG_FAILURE; 
  }
  
// test special for multiply ...
qar_status =  QarTestConformalForVecProds(in1_ptr, in2_ptr);

  if(qar_status == QAR_ERR)  
  {
    QplugAppendTextDisplay ("Arithmetic>> input not a vector or frames don't match\n");
     return QPLUG_FAILURE; 
  }

// do arithmetic...

  out_ptr = outer(in1_ptr, in2_ptr);

if (out_ptr.is_nil()) {
    QplugAppendTextDisplay (
        "Arithmetic>> Not valid for these input types\n");
    return QPLUG_FAILURE; 
}


 // Add timeline: Either timeline is safe if it exists
 
 QdTimeSeq_var tt = get_timetags(in1_ptr);
 if (tt.is_nil()) tt = get_timetags(in2_ptr);
 if (!tt.is_nil()) set_timetags(out_ptr, tt);

 // Set QSAS essential attriutes
 
 // SI_conversion
 
 
 if ( (attr_ptr = QarSIConv_product(in1_ptr, in2_ptr)) != NULL) {
   QuSetTxtAttr("SI_conversion", attr_ptr, out_ptr);
   delete[] attr_ptr;
 }

 // FIELDNAM

 // ignore fieldnames as concatenating these rapidly gets too long

   QuSetTxtAttr("FIELDNAM", out_name, out_ptr);
   free (out_name);
 
 // Frame

  QuSetFrame(out_ptr, QuGetFrameAttr(  in1_ptr, in2_ptr ));
 
 // UNITS
 
 if ( (attr_ptr = QarUNITS_product(in1_ptr, in2_ptr)) != NULL) {
   QuSetTxtAttr("UNITS", attr_ptr, out_ptr);
   delete[] attr_ptr;
 } 
 
 // DEPEND_i etc.
 
 bool OK = copyDimXrefs(in1_ptr, out_ptr);
 // only try other object if NO dimensioned xrefs found for first
 if(!OK) OK = copyDimXrefs(in2_ptr, out_ptr);
 if (OK) QplugAppendTextDisplay ("Dimensional Xrefs copied to new object\n");


  //	..add  the output series to the list

  call_list->push_back(out_ptr);

  QplugAppendTextDisplay ("Vector Product: Done\n");
  
return status;

}

