#include "qOperators.h"

QdObject_var QdFills( QString& rqstrErrorMsg, QdObject_var dobj, QdObject_var userValue )
{
   QdRScalar_var newVal_ptr;
   QdObject_var out_ptr;
   
   if(dobj.is_nil() || userValue.is_nil()  ){
     QcalcAppendTextDisplay( "Invalid input object in Fillval Conversion \n" );
     return out_ptr; 
   }
 
   newVal_ptr = QdRScalar_var::narrow(userValue);
   if( newVal_ptr.is_nil() ){
     QcalcAppendTextDisplay( "Invalid input target value for Fillval Conversion \n" );
     return out_ptr; 
   }
 
  // create output object

  QdObject_var fill_xref;
  double fill;
  bool hasTT = false;
  QdObject_var obj_out;
  
  double newValue = (double) *newVal_ptr;
  QdRScalar_var fill_out = new QdRScalar( newValue );

    if (dobj->xref_exists("FILLVAL"))
      fill_xref = dobj->get_xref("FILLVAL");

    fill = -10.e+30; // in case xref not set
    
    QdTimeSeq_var ts_in=get_timetags(dobj);
    hasTT = !(ts_in.is_nil());

    // is sequence a matrix sequence
    QdRMatrixSeq_var ms_in=QdRMatrixSeq_var::narrow(dobj);
    if(!ms_in.is_nil())
    {
      if(ms_in->empty())
          return obj_out;

      // if so, do we have a single fill-value or a matrix of different fill
      // values for each dimension

      QdRMatrix_var fill_m=QdRMatrix_var::narrow(fill_xref);

      // ...either way create matrix of fill values
      // this assumes the MatrixSeq is homogeneous i.e. each matrix
      // element has the same no. and sizes of dimensions

      if(fill_m.is_nil())
      {
          QdRScalar_var fill_v=QdRScalar_var::narrow(fill_xref);
          if(fill_v.is_nil())
              fill_v = new QdRScalar(fill);

          fill_m=new QdRMatrix(ms_in[0].dimdata(),fill_v->scalar());

      }

      // initialise matrix output sequence and reserve max space
      QdRMatrixSeq_var ms_out = Factory::clone(ms_in);
	 
      // get underlying data for fill matrix
      valarray<double> v_fill(fill_m->data()[fill_m->getgslice()]);

      for(register int i=0;i<ms_out->size();i++)
      { 
	      // create valarray<double> of results for assigning into
           // matrix (gslice_array)
           bool update=false;       
           valarray<double> v_out(ms_out[i].data()[ms_out[i].getgslice()]);
           for(register int j=0;j<v_out.size();j++)
           {
               const double& dfill = v_fill[j%v_fill.size()];
               if( QuIsMatch( v_out[j], dfill) ) {
                   update=true;
                   v_out[j] = newValue;  // newValue is a double
               }
           }
           // assign results into gslice array (only if need to)
           if(update)
             ms_out[i].data()[ms_out[i].getgslice()]=v_out;
      }

      ms_out->copy_xrefs_from(ms_in);
	 ms_out->change_xref(FILL_VALUE, fill_out);
	 

      if( hasTT){
         set_timetags(ms_out,ts_in);
      }
      return ms_out;
   }

    // scalar sequence case

   QdRScalarSeq_var ss_in=QdRScalarSeq_var::narrow(dobj);
   if(!ss_in.is_nil())
   {
     if(ss_in->empty())
          return obj_out;

      QdRScalar_var fill_v=QdRScalar_var::narrow(fill_xref);
      if( !fill_v.is_nil()) fill = (double) *fill_v;

      QdRScalarSeq_var ss_out = Factory::clone(ss_in);
	 
      for(int i=0;i<ss_in->size();i++){
          if( QuIsMatch(ss_out[i], fill) ) ss_out[i] = newValue;
      }

      ss_out->copy_xrefs_from(ss_in);
	 ss_out->change_xref(FILL_VALUE, fill_out);

      if( hasTT ){
        set_timetags(ss_out, ts_in);
      }
      return ss_out;
   }


    QdRMatrix_var m_in=QdRMatrix_var::narrow(dobj);
    if(!m_in.is_nil())
    {

      // if so, do we have a single fill-value or a matrix of different fill
      // values for each dimension

      QdRMatrix_var fill_m=QdRMatrix_var::narrow(fill_xref);

      // ...either way create matrix of fill values
      // this assumes the MatrixSeq is homogeneous i.e. each matrix
      // element has the same no. and sizes of dimensions

      if(fill_m.is_nil())
      {
          QdRScalar_var fill_v=QdRScalar_var::narrow(fill_xref);
          if(fill_v.is_nil()) fill_v = new QdRScalar(fill);

          fill_m=new QdRMatrix(m_in->dimdata(),fill_v->scalar());

      }


      // iterator for fill values
      RMatrix::iterator fitr=fill_m->begin();
      
	 QdRMatrix_var m_out = Factory::clone(m_in);
      m_out->copy_xrefs_from(m_in);
	 m_out->change_xref(FILL_VALUE, fill_out);

      if( hasTT){
         set_timetags(m_out,ts_in);
      }
	 
      // iterator for matrix
      RMatrix::iterator itr=m_out->begin();
      int j=0;
      while(itr!=m_out->end() )
      {
          double& d=*itr++;
          if( QuIsMatch(d, *(fitr+j++) ) ) d = newValue;
      }

      return m_out;    // return original object

    }

   QdRScalar_var s_in=QdRScalar_var::narrow(dobj);
   if(!s_in.is_nil())
   {
      QdRScalar_var fill_v=QdRScalar_var::narrow(fill_xref);
      if( !fill_v.is_nil()) fill = (double) *fill_v;
	 
	 QdRScalar_var s_out = Factory::clone(s_in);
      s_out->copy_xrefs_from(s_in);
	 s_out->change_xref(FILL_VALUE, fill_out);

      if( hasTT){
         set_timetags(s_out,ts_in);
      }

      if( QuIsMatch( *s_out, fill)) {
	   *s_out = newValue;
	 }


      return s_out; 
   }

   // string and other input types not cleaned - return input object

   QcalcAppendTextDisplay( " Fillval Conversion not handled for this data type\n" );
   return dobj;

}



