#include "qOperators.h"


QdObject_var QdSub( QString& rqstrErrorMsg, QdObject_var qdOperand1, QdObject_var qdOperand2 )
{
    char*   attrPtr;
    char*   stringBuf;
    QdObject_var  qdResult;

    if ( ( qdOperand1.is_nil() ) || ( qdOperand2.is_nil() ) )
    {
        // The object will be empty - this should never happen
	// The qdResult object is NULL at this point
        return ( qdResult );
    }
    
 	// Test validity of inputs...
	int nQarStatus =  QarTestConformal( qdOperand1, qdOperand2 );
	if ( nQarStatus == QAR_ERR )
	{
		QcalcAppendTextDisplay( "Sequence size or join error: Notify csc_support" );
		return qdResult;
	}
	
       // some validity tests are done in isOpAllowed() called by DoOperation() before (pairwise) joining

        if( QarAreUnitsSame( qdOperand1, qdOperand2) == QAR_ERR ){
	  // try to convert into same units 
	  try{
             QcalcAppendTextDisplay( "Trying to convert to same units" );
             qdOperand2 = QarConvertUnitsTo(qdOperand2, QuGetAttrText("SI_conversion", qdOperand1) ); 
              // this creates new converted object, copies xrefs across and sets SI_conversion
          }
          catch(Exception &e){
               QcalcAppendTextDisplay( e.msg().c_str() );
               return qdResult;  
          }  
	}
	

	// do arithmetic ...
	try
	{
		qdResult = qdOperand1 - qdOperand2;
	}
	catch( Exception &e )
	{
		QcalcAppendTextDisplay ( e.msg().c_str() );
		QcalcAppendTextDisplay( "The operation substraction is not yet supported between these input types");
		rqstrErrorMsg = QString( "The operation substraction is not yet supported between these input types");
		return ( qdResult );
	}


	if ( qdResult.is_nil() )
	{
	    QcalcAppendTextDisplay( "Arithmetic>> This Operation is not yet supported between these input types" );
		rqstrErrorMsg = QString( "Arithmetic>> This Operation is not yet supported between these input types" );
		return ( qdResult );
	}

	// Add timeline: Either timeline is safe if it exists
	QdTimeSeq_var tt = get_timetags( qdOperand1 );
	if ( tt.is_nil() )
	{
		tt = get_timetags( qdOperand2 );
	}

	if ( !tt.is_nil() )
	{
		set_timetags( qdResult, tt );
	}

	// Set QSAS essential attriutes
	// for subtraction they stay the same as inputs
	// SI_conversion
	attrPtr = QuGetAttrText( "SI_conversion", qdOperand1 );
	if( strcmp(attrPtr, "") != 0 )
	{
		QuSetTxtAttr( "SI_conversion", attrPtr, qdResult );
	}
	else
	{
		attrPtr = QuGetAttrText( "SI_conversion", qdOperand2 );
		QuSetTxtAttr( "SI_conversion", attrPtr, qdResult );
	}


	// Frame
	QuSetFrame( qdResult, QuGetFrameAttr(  qdOperand1, qdOperand2 ) );

	// UNITS
	attrPtr = QuGetAttrText( "UNITS", qdOperand1 );
	if( strcmp(attrPtr, "") != 0)
	{
		QuSetTxtAttr( "UNITS", attrPtr, qdResult );
	}
	else
	{
		attrPtr = QuGetAttrText( "UNITS", qdOperand2 );
		QuSetTxtAttr( "UNITS", attrPtr, qdResult );
	}

	// DEPEND_i
    if (qdOperand1->xref_exists(DEPEND_1)) 
    {
        QdObject_var d1_xref = qdOperand1->get_xref(DEPEND_1);
        qdResult->set_xref( DEPEND_1, d1_xref  );
    }
    else if (qdOperand2->xref_exists(DEPEND_1))
    {
        QdObject_var d1_xref = qdOperand2->get_xref(DEPEND_1);
        qdResult->set_xref( DEPEND_1, d1_xref  );
    }

    if (qdOperand1->xref_exists(DEPEND_2)) 
    {
        QdObject_var d1_xref = qdOperand1->get_xref(DEPEND_2);
        qdResult->set_xref( DEPEND_2, d1_xref  );
    }
    else if (qdOperand2->xref_exists(DEPEND_2))
    {
        QdObject_var d1_xref = qdOperand2->get_xref(DEPEND_2);
        qdResult->set_xref( DEPEND_2, d1_xref  );
    }

    if (qdOperand1->xref_exists(DEPEND_3)) 
    {
        QdObject_var d1_xref = qdOperand1->get_xref(DEPEND_3);
        qdResult->set_xref( DEPEND_3, d1_xref  );
    }
    else if (qdOperand2->xref_exists(DEPEND_3))
    {
        QdObject_var d1_xref = qdOperand2->get_xref(DEPEND_3);
        qdResult->set_xref( DEPEND_3, d1_xref  );
    }


    return ( qdResult );
}

