#include "qcalc.h"
#include "qdutil.h"
#include "qwui.h"

//#ifdef QT_DEFINED
#include <qapplication.h>
#include <qprogressdialog.h>
#include <qmessagebox.h>
//#endif



using namespace QSAS;


bool QcalcWindow::CheckSelfLinkedVirtualObjs()
{
	// Check if the Virtual objects are linked to themselves (directly or indirectly)
	QcalcListVirtualItem* qNextItem  = ( QcalcListVirtualItem * ) this->_join_listview->firstChild();
	bool                  bReturn    = true;

	while ( qNextItem )
	{
		if ( qNextItem->m_eObjType == VIRTUAL_OBJECT )
		{
			QString qstrAlias = QString("");
			bReturn = qNextItem->CheckOperatorSelfLink( qstrAlias );
			if ( !bReturn )
			{
				QString qstrError = 
				         QString( "Error: Item [ %1 ] is linked to itself (directly or indirectly)" ).arg( qNextItem->text(Alias_));
				QcalcAppendTextDisplay( qstrError.latin1() );
				DisplayEndMessage();
				QApplication::restoreOverrideCursor();
				return ( true );
			}
		}
		qNextItem = ( QcalcListVirtualItem* ) qNextItem->QguGetNext( );
	}

	return ( false );
}

bool QcalcWindow::JoinFirst()
{
    QdObjectSeq_var          qdJoinedList = new QdObjectSeq();
	if ( qdJoinedList.is_nil() )
	{
		QcalcAppendTextDisplay( "Error: Cannot start calculation - Memory cannot be allocated for intermediary objects" );
		DisplayEndMessage();
		return ( false );
	}

	// Set up the target time tags here
	// these are used for all joins:- join first, join for operations and join after
	
	this->_join_listview->qdtargetTT = getTimeTags();  // this is member of QcalcListView as it applies to all items

	if ( this->_join_listview->qdtargetTT.is_nil() )
	{
		QcalcAppendTextDisplay( "No time tags found for the object list" );
		if ( multiJn_check1->isChecked() ) return (false); // no tt so cannot join first
		                                   return (true); // no tt but join may be unnecessary
	}

	vector< KVDataBase_var > qvOptionList;
	QcalcListVirtualItem* qNextItem = NULL;

	if ( multiJn_check1->isChecked() )
	{
		QcalcAppendTextDisplay( "All time-tagged objects will be multi-joined before calculation" );
		
		// do Multi-join
		
		if ( !joinObjects( qdJoinedList, qvOptionList, TRUE ) ) // TRUE skips virtul objects
		{
			QApplication::restoreOverrideCursor();
			QcalcAppendTextDisplay( "Multi-join: Error joining time-tagged objects before calculation" );
			DisplayEndMessage();
			return ( false );
		}

		// ensure each WL item has a joined return value
		qNextItem  = ( QcalcListVirtualItem * ) this->_join_listview->firstChild();
		while ( qNextItem )
		{
			if ( qNextItem->m_eObjType == WORKING_LIST_OBJECT &&  qNextItem->entryType == ListTypeEntry )
			{		
				// go through joined list and attach each joined object back to item m_qdObjJoinedFirst pointer
				for ( int nIndex = 0; nIndex < qdJoinedList->size(); nIndex++ )
				{
					QdObject_var qObj = qdJoinedList[ nIndex ];
					if ( qObj.is_nil() ) continue;

					// Get the name of the objects being joined
					QdString_var qdName = QdString_var::narrow( ( qvOptionList[ nIndex ] )->get( NEW_NAME ) );
					if ( qdName.is_nil() ) continue;

					if ( qNextItem->text(Alias_) == QString( qdName->c_str() ) )
					{
						qNextItem->m_qdObjJoinedFirst = qObj;
					}
				}
				// put constants onto m_qdObjJoinedFirst pointer
				if(qNextItem->m_qdObjJoinedFirst.is_nil())
				{
					qNextItem->m_qdObjJoinedFirst = qNextItem->m_pcalcTab->operandStack[0]->slot->getSlotObject();
					if( qNextItem->m_qdObjJoinedFirst->xref_exists(TIME_TAGS) )
					{
						QApplication::restoreOverrideCursor();
						QcalcAppendTextDisplay( " time-tagged object failed to join before calculation" );
						DisplayEndMessage();
						return ( false );
					}
					
				}
			}
			qNextItem = ( QcalcListVirtualItem* ) qNextItem->QguGetNext( );
		}
	}

	return ( true );

}


bool QcalcWindow::JoinLast()
{
	if ( multiJn_check2->isChecked() )
	{
		QdObjectSeq_var          qdJoinedList = new QdObjectSeq();
		if ( qdJoinedList.is_nil() )
		{
			QcalcAppendTextDisplay( "Error: Cannot start calculation - Memory cannot be allocated for intermediary objects" );
			DisplayEndMessage();
			return ( false );
		}
	
		vector< KVDataBase_var > qvOptionList;
		QcalcListVirtualItem* qNextItem = NULL;


		QcalcAppendTextDisplay( "All time-tagged objects will be multi-joined after calculation" );
		
		// do Multi-join
		
		if ( !joinObjects( qdJoinedList, qvOptionList, FALSE ) ) // FALSE does not skip virtul objects - they should exist
		{
			QApplication::restoreOverrideCursor();
			QcalcAppendTextDisplay( "Multi-join: Error joining time-tagged objects after calculation" );
			DisplayEndMessage();
			return ( false );
		}

		// ensure each WL item has a joined return value
		qNextItem  = ( QcalcListVirtualItem * ) this->_join_listview->firstChild();
		while ( qNextItem )
		{
			if (  qNextItem->entryType == ListTypeEntry )
			{		
				// go through joined list and attach each joined object back to item m_qdObjJoinedFirst pointer
				for ( int nIndex = 0; nIndex < qdJoinedList->size(); nIndex++ )
				{
					QdObject_var qObj = qdJoinedList[ nIndex ];
					if ( qObj.is_nil() ) continue;

					// Get the name of the objects being joined
					QdString_var qdName = QdString_var::narrow( ( qvOptionList[ nIndex ] )->get( NEW_NAME ) );
					if ( qdName.is_nil() ) continue;

					if ( qNextItem->text(Alias_) == QString( qdName->c_str() ) )
					{
						qNextItem->m_qdObjResult = qObj;
					}
				}
				// put constants onto m_qdObjJoinedFirst pointer
				if(qNextItem->m_qdObjResult.is_nil())
				{
					qNextItem->m_qdObjResult = qNextItem->m_pcalcTab->operandStack[0]->slot->getSlotObject();
					if( qNextItem->m_qdObjResult->xref_exists(TIME_TAGS) )
					{
						QApplication::restoreOverrideCursor();
						QcalcAppendTextDisplay( " time-tagged object failed to join before calculation" );
						DisplayEndMessage();
						return ( false );
					}
				}
			}
			qNextItem = ( QcalcListVirtualItem* ) qNextItem->QguGetNext( );
		}
	}
	
	return ( true );
}


void QcalcWindow::DoCalculation()
{
	QdObjectSeq_var          qdJoinedList = new QdObjectSeq();
	QdObject_var nilObj;
	vector< KVDataBase_var > qvOptionList;
	char   bufLog[ 1024 ];
	QDate  qdate = QDate::currentDate();
	QTime  qTime = QTime::currentTime();

	newCalcBtn->setEnabled( false );
	// Display start message log
	
#ifdef QT_VERSION
#if QT_VERSION <0x030000
	sprintf( bufLog, "%s Calculator Started", qTime.toString( ).latin1() );
#else
	sprintf( bufLog, "%s Calculator Started", qTime.toString( "hh:mm:ss.zzz" ).latin1() );
#endif // end of #if
#endif // end of #ifdef


	QcalcAppendTextDisplay( bufLog );

	if ( qdJoinedList.is_nil() )
	{
		QcalcAppendTextDisplay( "Error: Cannot start calculation - Memory cannot be allocated for intermediary objects" );
		DisplayEndMessage();
		newCalcBtn->setEnabled( true );
		return;
	}

#ifdef QT_VERSION
#if QT_VERSION <0x030000
	QApplication::setOverrideCursor( QCursor( WaitCursor ) );
#else
	QApplication::setOverrideCursor( QCursor( Qt::WaitCursor ) );
#endif // end of #if
#endif // end of #ifdef
	


	qApp->processEvents();
	if ( CheckSelfLinkedVirtualObjs() )
	{
		newCalcBtn->setEnabled( true );
		QApplication::restoreOverrideCursor();
		return;
	}


	// Reset the claculation done flag and check that SOMETHING gets saved to WL
	bool noResultSaved = true;
	QcalcListVirtualItem* qNextItem = ( QcalcListVirtualItem * ) this->_join_listview->firstChild();
	while ( qNextItem )
	{
		if (qNextItem->m_pcalcTab && qNextItem->m_pcalcTab->saveWL_check->isChecked() ) noResultSaved = false;
		
		if (qNextItem->getEntryType() == ListTypeEntry)
		{
			qNextItem->m_bCalcDone = false;
			qNextItem->m_qdObjJoinedFirst = nilObj;
			qNextItem->m_qdObjResult = nilObj;
		}
		qNextItem = ( QcalcListVirtualItem* ) qNextItem->QguGetNext( );
	}

	if(noResultSaved) {
		QcalcAppendTextDisplay( "No item has 'Save to WL' ticked" );
		QcalcAppendTextDisplay( "Calculations ignored" );
		DisplayEndMessage();
		messageHandler( "All operations virtual: No results would be saved", true );
		newCalcBtn->setEnabled( true );
		QApplication::restoreOverrideCursor();
		return;
	}
	
	qApp->processEvents();

	// Check if joining first is required
	if ( !JoinFirst() )
	{
		// JoinFirst returned an error
		QcalcAppendTextDisplay( "Problem in joining first" );
		newCalcBtn->setEnabled( true );
		QApplication::restoreOverrideCursor();
		return;
	}
	// Do the actual calculation
	qNextItem = ( QcalcListVirtualItem * ) this->_join_listview->firstChild();
	if ( qNextItem == NULL )
	{
		QApplication::restoreOverrideCursor();
		QcalcAppendTextDisplay( "Calculation: Null pointer encountered" );
		DisplayEndMessage();
		newCalcBtn->setEnabled( true );
		QApplication::restoreOverrideCursor();
		return;
	}

	int nStep = 0;
	qNextItem = ( QcalcListVirtualItem * ) this->_join_listview->firstChild();
	while ( qNextItem )
	{
		// Calculate the number of steps
		nStep++;
		qNextItem = ( QcalcListVirtualItem* ) qNextItem->QguGetNext( );
	}

#if QT_VERSION < 0x040000
	QProgressDialog dlgProg( "Calculating...", "Abort", nStep, this, "progress", TRUE );
#else
        QProgressDialog dlgProg( "Calculating...", "Abort", 0, nStep, this, Qt::Dialog );
#endif
	dlgProg.setCaption( "Calculator" );
	dlgProg.show();
	qApp->processEvents();

	this->_join_listview->clearSelectedList();

	int nCounter = 0;

	qNextItem = ( QcalcListVirtualItem * ) this->_join_listview->firstChild();
	while( qNextItem )
	{
#if QT_VERSION < 0x040000
		dlgProg.setProgress( nCounter );
#else
                dlgProg.setValue( nCounter );
#endif
	 	qApp->processEvents();
#if QT_VERSION < 0x040000
		if ( dlgProg.wasCancelled() )
#else
                    if ( dlgProg.wasCanceled() )
#endif
		{
			switch	(	QMessageBox::warning(
			                                  0,
			                                  "Calculator",
			                                  "Abort?",
					                          QMessageBox::No | QMessageBox::Default,
					                          QMessageBox::Yes,
											  QMessageBox::Cancel | QMessageBox::Escape
											)	 )
			{
				case QMessageBox::Yes:
						QApplication::restoreOverrideCursor();
						newCalcBtn->setEnabled( true );
						QcalcAppendTextDisplay( "Calculation: Calculator aborted" );
						DisplayEndMessage();
						//break;
						return;

				case QMessageBox::No:
				case QMessageBox::Cancel:
				default:
						dlgProg.reset();
						dlgProg.setLabelText("Calculating...");
#if QT_VERSION < 0x040000
						dlgProg.setProgress( nCounter );
#else
                                                dlgProg.setValue( nCounter );
#endif
						dlgProg.show();
						qApp->processEvents();
						break;
	   		}
	 }

		// Show which item is being processed

		this->_join_listview->setSelected( qNextItem, true);
		this->_join_listview->ensureItemVisible( qNextItem );

		qNextItem->DoOperation();
		qNextItem = ( QcalcListVirtualItem* ) qNextItem->QguGetNext( );
		nCounter++;

	}
	//this->_join_listview->localList.clear();
#if QT_VERSION < 0x040000
	dlgProg.setProgress( nStep );
#else
        dlgProg.setValue( nStep );
#endif

	qApp->processEvents();
	// Check if joinig last is required
	if ( !JoinLast() )
    {
		// Debugging info are within the calling function
		newCalcBtn->setEnabled( true );
		QApplication::restoreOverrideCursor();
		return;
	}


	// Write the objects into the working list
	qNextItem = ( QcalcListVirtualItem * ) this->_join_listview->firstChild();

	while ( qNextItem )
	{
		qApp->processEvents();
		QcalcTab *qTab = qNextItem->m_pcalcTab;
		if(qTab == NULL) continue;
		
		if ( qTab->saveWL_check->isChecked() )
		{
			if (  !( qNextItem->m_qdObjResult.is_nil() )  )
			{
				try
				{
					QString newObjectName = qTab->_nameWL_field->text();
					if( newObjectName.isEmpty() ) newObjectName = QString(qNextItem->text(Alias_));
					
					messageHandler( "Calculator: Placing object " + newObjectName + " In the working list", false );
					QcalcAppendTextDisplay( "Calculator: Placing object " + newObjectName + " In the working list" );
					
					// set FIELDNAM to be object name
					QuSetTxtAttr( "FIELDNAM", newObjectName.latin1(), qNextItem->m_qdObjResult );

					QwuiPutOnWL(  qNextItem->m_qdObjResult,
								newObjectName.latin1(),
								"Data" , false);
				}
				catch( Exception &e )
				{
					messageHandler( QString( e.msg().c_str() ), true );
					QcalcAppendTextDisplay( e.msg().c_str() );
					DisplayEndMessage();
					QApplication::restoreOverrideCursor();
					newCalcBtn->setEnabled( true );
					return;
				}
			}
		}

		qNextItem = ( QcalcListVirtualItem* ) qNextItem->QguGetNext( );
	}


	QApplication::restoreOverrideCursor();
	QcalcWindow::DisplayEndMessage();
	QcalcAppendTextDisplay( " " );
	newCalcBtn->setEnabled( true );
}

void QcalcWindow::DisplayEndMessage()
{
	QDate qdate = QDate::currentDate();
	QTime qTime = QTime::currentTime();
	char  bufLog[ 1024 ];

#ifdef QT_VERSION
#if QT_VERSION <0x030000
	sprintf( bufLog, "%s %s Calculator Ended",  qdate.toString().latin1(), qTime.toString().latin1() );
#else
	sprintf( bufLog, "%s %s Calculator Ended",  qdate.toString().latin1(), qTime.toString( "hh:mm:ss.zzz" ).latin1() );
#endif // end of #if
#endif // end of #ifdef


	QcalcAppendTextDisplay( bufLog );
}

