
/**********************************************************************
* SCCS id: @(#)CorCoefPlug.cc	1.1 14:32:23 12/17/98
************************************************************************
* PROJECT:     Cluster UK CDHF
* COMPONENT:   Correlation Coefficient Plugin 
* MODULE:      CorCoefPlug.cc
* LANGUAGE:	   C++
* FUNCTION:
* PURPOSE:     Compute Correlation Coefficient between 2 scalar time series 
* DESCRIPTION:
* ARGUMENTS: [passed via call_list] - see qtpl and CorCoef.c for details
*
*   INPUT  ANY_TIME_SERIES xts_in, 
*   INPUT  ANY_TIME_SERIES yts_in, 
*   INPUT  TIME_INTERVAL start_end,
*   INPUT  FLOAT dt_lag_max,
*   OUTPUT SCALAR_SERIES dt_lag,
*   OUTPUT SCALAR_SERIES rho_out,
*   OUTPUT VECTOR_SERIES rho_vec,
*
* RETURN:
*    TYPE:     QplugReturnStatus
*    VALUES:   see qplug_if.h
* FUNCTIONS_USED: CorCoef, stats, CorCoefPlug_utils, IndexTSSubset
* ALGORITHM:   see qtpl and CorCoef.c
* HISTORY:	   Steve Schwartz  Dec 98
*               vector input/ouput added M.Fraenz Feb 2001
***********************************************************************/

#ifndef __cplusplus
#define __cplusplus
#endif

#include "CorCoefPlug.h"       // this then sucks in all other headers

QplugReturnStatus  CorCoefStart  ( QplugArgList *call_list )
{

CorCoefPlugReturnStatus CCP_return;     // used to return from sub-routines
CCutilsSuccessorFailure       ok;
CorCoefReturnStatus     CC_return;

CorCoefPlug_ptrs       ccp,*ccpp=&ccp;  // holds pointers to data arrays
CorCoefpacket          ccpack;

double half_spacing, tol;
int			ixstart,iystart;
                                        

QplugAppendTextDisplay("Correlation Coefficient. 6 Feb 2001\n");

CCP_return=CorCoefPlug_ini(ccpp);   // initialise NULL ptrs

//	...first, check the types of the input arguments, and load the data

CCP_return =CorCoefPlug_checkload_input (call_list, ccpp);
if (CCP_return != CORCOEF_PLUG_SUCCESS){
    QplugAppendTextDisplay(CorCoefPlugReturnMessage[CCP_return]);
    return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);
}

#ifdef VERBOSE_PLUG
QplugAppendTextDisplay("CorCoefPlug input types ok and loaded successfully\n");
#endif

CCP_return = CorCoefPlug_SubsetXY (ccpp);
if (CCP_return != CORCOEF_PLUG_SUCCESS){
    QplugAppendTextDisplay(CorCoefPlugReturnMessage[CCP_return]);
    return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);
}
ixstart = ccpp->ixstart;
iystart = ccpp->iystart;
// ... test timetags for regular and joined. Set tag spacing in ms

ok = Test_regular (&(ccpp->tx[ixstart]), ccpp->nxsub, &(ccpp->dt_spacing));
if (ok != CCUTILSSUCCESS){
 QplugAppendTextDisplay("CorCoefPlug ERROR>> x timetags not regular\n");
 return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);
}
ccpp->dt_spacing *= 2.0;    // Test_regular returns half_spacing

// Use spacing to deterine tag tolerance. Minumum of 10 ms or 10% spacing
tol = ccpp->dt_spacing > 100 ? 10 : 0.1*ccpp->dt_spacing;

ok = Test_join(&(ccpp->tx[ixstart]), ccpp->nxsub, 
               &(ccpp->ty[iystart]), ccpp->nysub, tol);
if (ok!=CCUTILSSUCCESS){
 QplugAppendTextDisplay("CorCoefPlug ERROR>> Subsetted tx and ty not Joined\n");
 return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);
}

#ifdef VERBOSE
  QplugAppendTextDisplay("CorCoefPlug: DO joined and regular tags\n");
#endif
 
QplugAppendTextDisplay(">>CorCoef >> Input checked\n");

// load packet for CorCoef c routine:
ccpack.npts = ccpp->nxsub;
ccpp->nlags =  (int) floor(ccpp->dt_lag_max / ccpp->dt_spacing);
ccpack.nlags = ccpp->nlags;
ccpack.x = ccpp->x + ixstart;
ccpack.y = ccpp->y + iystart;
ccpp->out_array_size = 2*(ccpp->nlags)+1;
ccpp->rho = new float[ccpp->out_array_size];
ccpack.corcoef_results =ccpp->rho;
ccpp->rvecset = ccpp->xvecset + ccpp->yvecset;



// ... do it
CC_return = CorCoef(&ccpack);
if (CC_return != CORCOEF_OK){
  QplugAppendTextDisplay(CorCoefErrorMessage[CC_return]);
  return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);  
}


// ... now do the vector components one by one
if (ccpp->rvecset != 0) {
 ccpp->rho_vx = new float[ccpp->out_array_size];
 ccpp->rho_vy = new float[ccpp->out_array_size];
 ccpp->rho_vz = new float[ccpp->out_array_size];
 
 ccpack.corcoef_results =ccpp->rho_vx;
 if (ccpp->xvecset != 0) {
  ccpack.x = ccpp->x_vx + ixstart;
  if (ccpp->yvecset != 0) ccpack.y = ccpp->y_vx + iystart;
 }
 else ccpack.y = ccpp->y_vx + iystart;
 
 CC_return = CorCoef(&ccpack);
 if (CC_return != CORCOEF_OK){
  QplugAppendTextDisplay(CorCoefErrorMessage[CC_return]);
  return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);  
 }
 
 ccpack.corcoef_results =ccpp->rho_vy;
 if (ccpp->xvecset != 0) {
  ccpack.x = ccpp->x_vy + ixstart;
  if (ccpp->yvecset != 0) ccpack.y = ccpp->y_vy + iystart;
 }
 else ccpack.y = ccpp->y_vy + iystart;

 CC_return = CorCoef(&ccpack);
 if (CC_return != CORCOEF_OK){
  QplugAppendTextDisplay(CorCoefErrorMessage[CC_return]);
  return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);  
 }
 
 ccpack.corcoef_results =ccpp->rho_vz;
 if (ccpp->xvecset != 0) {
  ccpack.x = ccpp->x_vz + ixstart;
  if (ccpp->yvecset != 0) ccpack.y = ccpp->y_vz + iystart;
 }
 else ccpack.y = ccpp->y_vz + iystart;

 CC_return = CorCoef(&ccpack);
 if (CC_return != CORCOEF_OK){
  QplugAppendTextDisplay(CorCoefErrorMessage[CC_return]);
  return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);  
 }

}


QplugAppendTextDisplay("CorCoefPlug>> Calculation of Cor Coef Finished OK\n");
//printf("CorCoefPlug>> Calculation of Cor Coef Finished OK\n");


// Create and fill output data object

CCP_return = CorCoefPlug_create_out_DOs(call_list, ccpp);
if (CCP_return != CORCOEF_PLUG_SUCCESS){
  QplugAppendTextDisplay(CorCoefPlugReturnMessage[CCP_return]);
  return CorCoefPlug_cleanup (QPLUG_FAILURE, ccpp);  
}  


//	..clean up and exit

QplugAppendTextDisplay ("CorCoefPlug>> - cleaning up ...\n");
return CorCoefPlug_cleanup (QPLUG_SUCCESS, ccpp);  
 
}  // end CorCoefPlug

/***********************************************************************
**************************************** CorCoefPlug_ini
***********************************************************************/

CorCoefPlugReturnStatus  CorCoefPlug_ini(CorCoefPlug_ptrs *ccpp){

ccpp->x = NULL;
ccpp->x_vx = NULL;
ccpp->x_vy = NULL;
ccpp->x_vz = NULL;

ccpp->tx=NULL;
ccpp->y = NULL;
ccpp->y_vx = NULL;
ccpp->y_vy = NULL;
ccpp->y_vz = NULL;
ccpp->ty=NULL;
ccpp->rho = NULL;
ccpp->rho_vx = NULL;
ccpp->rho_vy = NULL;
ccpp->rho_vz = NULL;
ccpp->dt_lag=NULL;

return CORCOEF_PLUG_SUCCESS;
}


/***********************************************************************
**************************************** CorCoefPlug_checkload_input
***********************************************************************/
//
// This routine:
//  o	Checks there are enough arguments on the list
//  o	Checks each input argument type
//  o	Puts x, y, interval data objects in the ccpp structure
//  o	Extracts ( x(i),tx(i) ) and ( y(i),ty(i) ) arrays into new'ed arrays
//  o	Extracts tstart and tend from subsetting interval and puts in ccpp
//  o	Extracts dt_lag_max and puts it into ccpp

CorCoefPlugReturnStatus  CorCoefPlug_checkload_input(QplugArgList *call_list,
        CorCoefPlug_ptrs *ccpp){


if ((call_list->size()) < 4){    // are there enough args?
   QplugAppendTextDisplay( "Not enough entries in argument list from GUI\n");
   return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}  //endif

// 1111111111111111111111111111111111111111111111111111111111111111111111

QdObject_var ts_in = (* call_list)[0];
QdRScalarSeq_var ss = QdRScalarSeq_var::narrow(ts_in);
if(!ss.is_nil() ){
   ccpp->xdo = (QdRScalarSeq_var) ss;
   ccpp->xvecset = 0;
}
else{
    QdRMatrixSeq_var ms = QdRMatrixSeq_var::narrow(ts_in);
    if(ms.is_nil() ){
      QplugAppendTextDisplay ("CorCoefPlug>> ERROR - sequence type not handled\n");
      return CORCOEF_PLUG_CHECK_INPUT_ERROR; 
    }
    if ( QuIsThreeVector(ms) ) {
     ccpp->xdo_vec = (QdRMatrixSeq_var) ms;
     ccpp->xvecset = 1;
     ccpp->xdo     = QdRScalarSeq_var::narrow(QuGetVectorComponent(ms, "r(mag)")); 
    }
    else {
      QplugAppendTextDisplay ("CorCoefPlug>> ERROR - sequence type not handled\n");
      return CORCOEF_PLUG_CHECK_INPUT_ERROR; 
    }
}


QdTimeSeq_var xtt = get_timetags(ts_in);
if(xtt.is_nil()){
   QplugAppendTextDisplay("CorCoefPlug ERROR>> X not a Time Series\n");	
   return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}

ccpp->nxdata_tot = (ccpp->xdo)->sequence_size();


if (ccpp->nxdata_tot <=0){
    QplugAppendTextDisplay("CorCoefPlug ERROR>> nxdata_tot <=0\n");	
    return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}
ccpp->x = new float[ccpp->nxdata_tot];
ccpp->tx = new double[ccpp->nxdata_tot];

for (int i =0; i<ccpp->nxdata_tot; i++){
  ccpp->x[i] = (float) (ccpp->xdo)[i];
  ccpp->tx[i] = cdfEpoch(xtt[i]);
}
if (ccpp->xvecset != 0) {
  ccpp->x_vx = new float[ccpp->nxdata_tot];
  ccpp->x_vy = new float[ccpp->nxdata_tot];
  ccpp->x_vz = new float[ccpp->nxdata_tot];
  for (int i =0; i<ccpp->nxdata_tot; i++){
   ccpp->x_vx[i] = (float) (ccpp->xdo_vec)[i][0];
   ccpp->x_vy[i] = (float) (ccpp->xdo_vec)[i][1];
   ccpp->x_vz[i] = (float) (ccpp->xdo_vec)[i][2];
  }
}

// 222222222222222222222222222222222222222222222222222222222222222222222



ts_in = (* call_list)[1];
ss = QdRScalarSeq_var::narrow(ts_in);
if(!ss.is_nil() ){
   ccpp->ydo = (QdRScalarSeq_var) ss;
   ccpp->yvecset = 0;
}
else{
    QdRMatrixSeq_var ms = QdRMatrixSeq_var::narrow(ts_in);
    if(ms.is_nil() ){
      QplugAppendTextDisplay ("CorCoefPlug>> ERROR - sequence type not handled\n");
      return CORCOEF_PLUG_CHECK_INPUT_ERROR; 
    }
    if ( QuIsThreeVector(ms) ) {
     ccpp->ydo_vec = (QdRMatrixSeq_var) ms;
     ccpp->yvecset = 1;
     ccpp->ydo     = QdRScalarSeq_var::narrow(QuGetVectorComponent(ms, "r(mag)"));
    } 
    else {
      QplugAppendTextDisplay ("CorCoefPlug>> ERROR - sequence type not handled\n");
      return CORCOEF_PLUG_CHECK_INPUT_ERROR; 
    }
}

QdTimeSeq_var ytt = get_timetags(ts_in);
if(ytt.is_nil()){
   QplugAppendTextDisplay("CorCoefPlug ERROR>> Y not a Time Series\n");	
   return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}

ccpp->nydata_tot = (ccpp->ydo)->sequence_size();


if (ccpp->nydata_tot <=0){
    QplugAppendTextDisplay("CorCoefPlug ERROR>> nydata_tot <=0\n");	
    return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}
ccpp->y = new float[ccpp->nydata_tot];
ccpp->ty = new double[ccpp->nydata_tot];

for (int i =0; i<ccpp->nydata_tot; i++){
  ccpp->y[i] = (float) (ccpp->ydo)[i];
  ccpp->ty[i] = cdfEpoch(ytt[i]);
}
if (ccpp->yvecset != 0) {
  ccpp->y_vx = new float[ccpp->nydata_tot];
  ccpp->y_vy = new float[ccpp->nydata_tot];
  ccpp->y_vz = new float[ccpp->nydata_tot];
  for (int i =0; i<ccpp->nydata_tot; i++){
   ccpp->y_vx[i] = (float) (ccpp->ydo_vec)[i][0];
   ccpp->y_vy[i] = (float) (ccpp->ydo_vec)[i][1];
   ccpp->y_vz[i] = (float) (ccpp->ydo_vec)[i][2];
  }
}


// 333333333333333333333333333333333333333333333333333333333333333333333

ccpp->interval_do = get_timeinterval((*call_list)[2]);

if ( (ccpp->interval_do).is_nil() ) { 
 QplugAppendTextDisplay("CorCoefPlug ERROR>> Third arg has no Time Interval\n");
   return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}

ccpp->tstart = cdfEpoch( (ccpp->interval_do)->start() );
ccpp->tend = cdfEpoch( (ccpp->interval_do)->end() );


// 444444444444444444444444444444444444444444444444444444444444444444444

QdRScalar_var lag_ptr = QdRScalar_var::narrow((*call_list)[3]);

if ( lag_ptr.is_nil() ) { 
   QplugAppendTextDisplay("CorCoefPlug ERROR>> Fourth arg not a float");
   return CORCOEF_PLUG_CHECK_INPUT_ERROR;
}

ccpp->dt_lag_max = (float)( *lag_ptr);

ccpp->dt_lag_max *= 1000.;   // convert secs to ms

if (ccpp->dt_lag_max > 0.5* fabs(ccpp->tend - ccpp->tstart)) {
  QplugAppendTextDisplay("CorCoefPlug: dt_lag_max greater than 1/2 subsetted interval\n");
  QplugAppendTextDisplay("dt_lag_max set to 1/4 subsetted interval\n");
  ccpp->dt_lag_max = 0.25*fabs(ccpp->tend - ccpp->tstart);
}

return CORCOEF_PLUG_SUCCESS;
}  //end CorCoefPlug_checkload_input


/***********************************************************************
**************************************** CorCoefPlug_SubsetXY
***********************************************************************/

CorCoefPlugReturnStatus  CorCoefPlug_SubsetXY (CorCoefPlug_ptrs *ccpp){

IndexTSSubsetErrorStatus index_ok;
char out_message[500];
int iend;

// subset x array:

index_ok = IndexTSSubset (ccpp->tstart, ccpp->tend, 
              ccpp->tx, ccpp->nxdata_tot, 
              &(ccpp->ixstart), &iend, out_message);
if (index_ok != INDEX_TSSUBSET_OK){
   QplugAppendTextDisplay ("CorCoefPlug Error Subsetting x. Message:\n");
   QplugAppendTextDisplay (out_message);
   return CORCOEF_PLUG_SUBSET_ERROR;
}
ccpp->nxsub = iend - (ccpp->ixstart) + 1;

index_ok = IndexTSSubset (ccpp->tstart, ccpp->tend, 
              ccpp->ty, ccpp->nydata_tot, 
              &(ccpp->iystart), &iend, out_message);
if (index_ok != INDEX_TSSUBSET_OK){
   QplugAppendTextDisplay ("CorCoefPlug Error Subsetting y. Message:\n");
   QplugAppendTextDisplay (out_message);
   return CORCOEF_PLUG_SUBSET_ERROR;
}
ccpp->nysub = iend - (ccpp->iystart) + 1;

return CORCOEF_PLUG_SUCCESS;
}


/***********************************************************************
**************************************** CorCoefPlug_create_out_DOs
***********************************************************************/
CorCoefPlugReturnStatus      CorCoefPlug_create_out_DOs(QplugArgList  *call_list,  
                               CorCoefPlug_ptrs *ccpp)
{

QdRScalarSeq_var rho_out;
QdObject_var rho_vecout;
QdRScalarSeq_var dt_lags_out;

float  *dtzerop, n2dtlag;
double rhomax=-1.0, rho_vecmax[3]={-1.0,-1.0,-1.0};
double rhomaxlag=0.0,rho_vecmaxlag[3]={0.0,0.0,0.0};
int i, nlags, jmax,jmaxx,jmaxy,jmaxz;

int n_min_attrs=5;    // attribute shufflers
char  *min_attr_ptr[]={"FIELDNAM", "Frame",    "SI_conversion", "UNITS",    "LABLAXIS"  };                 
char *rho_atts[]={"Correlation Coefficient", "scalar>na", "1>(cor coef)", " ","Cor Coef"};
char *rho_v_atts[]={"Correlation Coefficient Comp", "vector>na_xyz", "1>(cor coef)", " ","Cor Coef"};
char *dt_atts[]={"Lag time (Var 1 leads Var2)","scalar>na", "1>s", "seconds", "Lag (V1 leads V2)"};
char *att_str_buff,*tstart_string,*tend_string,tinterval_string[70];
char *fieldnamex, *fieldnamey,*fields_string,*dt_lags_out_fieldnam;
char	str_out[150];


nlags = ccpp->nlags;
ccpp->dt_lag = new float[ccpp->out_array_size];

dtzerop = ccpp->dt_lag + nlags;
n2dtlag = (ccpp->dt_spacing)/1000.0;
for (i=-nlags; i<nlags+1; i++){
   dtzerop[i]=i*n2dtlag;
}                                                                                                                                                                                                                                                                                                                                                                                                                                                      // Create the output data Series and fill them. Add attributes

rho_out     = new QdRScalarSeq( ccpp->out_array_size);
dt_lags_out = new QdRScalarSeq( ccpp->out_array_size );

for(int j =0; j<ccpp->out_array_size ; j++){
  rho_out[j] = (double) ccpp->rho[j];
  if (ccpp->rho[j] >= rhomax) rhomax=ccpp->rho[(jmax=j)];
  dt_lags_out[j] = (double) ccpp->dt_lag[j];
}
rhomaxlag=ccpp->dt_lag[jmax];
if (ccpp->rvecset !=0) {
  QdRMatrixSeq_var rho_vec = new QdRMatrixSeq( ccpp->out_array_size,RMatrix(Dim(3)));
  for(int j =0; j<ccpp->out_array_size ; j++) {
  rho_vec[j][0] = (double) ccpp->rho_vx[j];
  if (ccpp->rho_vx[j] >= rho_vecmax[0]) rho_vecmax[0]=ccpp->rho_vx[(jmaxx=j)];
  rho_vec[j][1] = (double) ccpp->rho_vy[j];
  if (ccpp->rho_vy[j] >= rho_vecmax[1]) rho_vecmax[1]=ccpp->rho_vy[(jmaxy=j)];
  rho_vec[j][2] = (double) ccpp->rho_vz[j];
  if (ccpp->rho_vz[j] >= rho_vecmax[2]) rho_vecmax[2]=ccpp->rho_vz[(jmaxz=j)];
  }
  rho_vecmaxlag[0]=ccpp->dt_lag[jmaxx];
  rho_vecmaxlag[1]=ccpp->dt_lag[jmaxy];
  rho_vecmaxlag[2]=ccpp->dt_lag[jmaxz]; 
  rho_vecout =  (QdObject_var) rho_vec;
}

// print maximal values to plugin window and append as xref:

sprintf(str_out,"CorCoef for Magnitudes>> max. correlation:%f, at  timelag: %f \n", 
            rhomax,rhomaxlag);
QplugAppendTextDisplay (str_out);
QdRScalar_var rhomax_var = new QdRScalar(rhomax);
rho_out->change_xref("MaxCor", rhomax_var);
QdRScalar_var rhomaxlag_var = new QdRScalar(rhomaxlag);
rho_out->change_xref("LagMax", rhomaxlag_var);

if (ccpp->rvecset !=0) {
 sprintf(str_out,"CorCoef for X-Component>> max. correlation:%f, at  timelag: %f \n", 
            rho_vecmax[0],rho_vecmaxlag[0]);
 QplugAppendTextDisplay (str_out); 
 sprintf(str_out,"CorCoef for Y-Component>> max. correlation:%f, at  timelag: %f \n", 
            rho_vecmax[1],rho_vecmaxlag[1]);
 QplugAppendTextDisplay (str_out); 
 sprintf(str_out,"CorCoef for Z-Component>> max. correlation:%f, at  timelag: %f \n", 
            rho_vecmax[2],rho_vecmaxlag[2]);
 QplugAppendTextDisplay (str_out);
 QdRMatrix_var rho_vecmax_var = new QdRMatrix(Dim(3),3,rho_vecmax);
 rho_vecout->change_xref("MaxCor", rho_vecmax_var);
 QdRMatrix_var rho_vecmaxlag_var = new QdRMatrix(Dim(3),3,rho_vecmaxlag);
 rho_vecout->change_xref("LagMax", rho_vecmaxlag_var);
}



for (i=0; i<n_min_attrs; i++){
   QuSetTxtAttr(min_attr_ptr[i], rho_atts[i], rho_out);
   if (ccpp->rvecset != 0) QuSetTxtAttr(min_attr_ptr[i], rho_v_atts[i], rho_vecout);
   QuSetTxtAttr(min_attr_ptr[i], dt_atts[i], dt_lags_out);
}

sprintf(tinterval_string,"Time Interval: %s to %s",
        cdfEpoch(ccpp->tstart).iso_srep().c_str(),
	cdfEpoch(ccpp->tend).iso_srep().c_str());
QuSetTxtAttr("Time_Interval", tinterval_string, rho_out);
if (ccpp->rvecset != 0) QuSetTxtAttr("Time_Interval", tinterval_string, rho_vecout);


if (ccpp->xvecset != 0)  fieldnamex = QuGetAttrText("FIELDNAM", ccpp->xdo_vec);
else fieldnamex = QuGetAttrText("FIELDNAM", ccpp->xdo);
if (ccpp->yvecset != 0)   fieldnamey = QuGetAttrText("FIELDNAM", ccpp->ydo_vec);
else fieldnamey = QuGetAttrText("FIELDNAM", ccpp->ydo);

if ( (strcmp(fieldnamex,"") ==0 ) || (strcmp(fieldnamey,"") ==0 ) ){
  QuSetTxtAttr("Analysis_Notes", "1st data series leads 2nd", rho_out);  
  if (ccpp->rvecset != 0) QuSetTxtAttr("Analysis_Notes", "1st data series leads 2nd", rho_vecout);
}
else{ 
  fields_string = new char[strlen(fieldnamex)+strlen(fieldnamey)+10];
  sprintf(fields_string,"%s leads %s",fieldnamex,fieldnamey);
  QuSetTxtAttr("Analysis_Notes", fields_string, rho_out);
  if (ccpp->rvecset != 0) QuSetTxtAttr("Analysis_Notes", fields_string, rho_vecout);  
  dt_lags_out_fieldnam = new char[strlen(fields_string)+15];
  sprintf(dt_lags_out_fieldnam,"Lag Time (%s)",fields_string);
  QuSetTxtAttr("FIELDNAM", dt_lags_out_fieldnam, dt_lags_out);  
  delete[] dt_lags_out_fieldnam;
  delete[] fields_string;
}



//  args to return   

  rho_out->change_xref("DEPEND_0",dt_lags_out); 
  if (ccpp->xvecset != 0)  rho_vecout->change_xref("DEPEND_0",dt_lags_out); 
  call_list->push_back(rho_out);
  call_list->push_back(rho_vecout);

return CORCOEF_PLUG_SUCCESS;
}


// *****************************************************************
// *****************************************************************
// ***************************************** CorCoefPlug_cleanup **
// *****************************************************************
// This routine cleans up after CorCoefPlug, deleting all new'ed memory
// See SJS Storage Management Notes
// *****************************************************************
//

QplugReturnStatus  CorCoefPlug_cleanup  (QplugReturnStatus returnstatus,
                           CorCoefPlug_ptrs *ccpp)
{

delete[]	ccpp->x;
delete[]	ccpp->x_vx;
delete[]	ccpp->x_vy;
delete[]	ccpp->x_vz;
delete[]	ccpp->tx;
delete[]	ccpp->y;
delete[]	ccpp->y_vx;
delete[]	ccpp->y_vy;
delete[]	ccpp->y_vz;
delete[]	ccpp->ty;
delete[]	ccpp->rho;
delete[]	ccpp->rho_vx;
delete[]	ccpp->rho_vy;
delete[]	ccpp->rho_vz;
delete[]	ccpp->dt_lag;

return returnstatus;

}  // end CorCoefPlug_cleanup


