#include <stdio.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"
#include <qdatetime.h>

#define MVAB = 0     // minimum variance of B-field
#define MFR  = 1     // minimum faraday residue
#define MMR  = 2     // minimum massflux residue
#define MVAJ = 3     // minimum variance of current density
#define MLMR = 4     // minimum linear momentum flux residue
#define MTER = 5     // minimum total energy residue
#define MER  = 6     // minimum entropy residue
#define MVAV  = 7     // minimum variance of plasma velocity

// --- prototypes ----

int SingleAnalysis(QdRMatrixSeq_var V, QdRMatrixSeq_var B, QdRMatrixSeq_var dB, QdRMatrixSeq_var P, 
		   QdRMatrixSeq_var E, QdRScalarSeq_var N, QdRMatrixSeq_var H, 
		   QdTimeSeq_var ttJoined, QdRMatrix_var ExWeights,
		   int constrained, double alpha, char *SCc_ID,
		   bool doType[10],int nestsize,
		   QplugArgList *call_list);

int NestedAnalysis(QdRMatrixSeq_var V_, QdRMatrixSeq_var B_, QdRMatrixSeq_var dB_, QdRMatrixSeq_var P_, 
		   QdRMatrixSeq_var E_, QdRScalarSeq_var N_, QdRMatrixSeq_var H_, 
   		   QdTimeSeq_var ttJoined, QdRMatrix_var ExWeights,
		   int constrained, double alpha, int min_nest, char *SCc_ID, 
		   bool doType[10],int nestsize);

QdRMatrix_var Q_MVAB_calc(QdRMatrixSeq_var B, QdRMatrixSeq_var V, QdRMatrix_var U, QdRMatrixSeq_var Valf,
			  QdRScalarSeq_var N, double *wl, double *cc, int nestsize, bool doHT);

QdRMatrix_var Q_MLMR_calc(QdRMatrixSeq_var V, QdRMatrixSeq_var B, QdRMatrixSeq_var P,
			  QdRScalarSeq_var N,  QdRMatrix_var U, int nestsize);

QdRMatrix_var Q_MER_calc(QdRMatrixSeq_var V, QdRMatrixSeq_var B, QdRMatrixSeq_var P,
			 QdRScalarSeq_var N, QdRMatrix_var U, int nestsize);

QdRMatrix_var Q_MFR_calc(QdRMatrixSeq_var B, QdRMatrixSeq_var V, QdRMatrix_var U, int nestsize);

QdRMatrix_var Q_MTER_calc(QdRMatrixSeq_var V, QdRMatrixSeq_var B, QdRMatrixSeq_var P, 
			  QdRMatrixSeq_var E, QdRScalarSeq_var N, QdRMatrixSeq_var H, 
			  QdRMatrix_var U, int nestsize);

QdRMatrix_var Q_MMR_calc(QdRMatrixSeq_var V,  QdRScalarSeq_var N, QdRMatrix_var U, int nestsize);

QdRMatrix_var Q_MVAV_calc(QdRMatrixSeq_var V, QdRMatrix_var U, int nestsize);

QdRMatrix_var Projection(QdRMatrixSeq_var B, QdRMatrixSeq_var V, QdRMatrix_var U,   
			 QdRMatrixSeq_var V_Alfven, QdRMatrix_var Evec, double alpha, 
			 int nestsize, int constrain, const double dens = 0.0);

double aveTotalPressure(QdRMatrixSeq_var B, QdRMatrixSeq_var P, int nestsize);
void Report(QdRMatrix_var Evalues, QdRMatrix_var Evec, bool outstd, const char *msg, int constrain, double alpha, char *SCc_ID, const bool changeOrder=false);

void ReportSiscoe(QdRMatrixSeq_var n, QdRMatrixSeq_var U, 
		 QdRMatrix_var Bdave, QdRMatrix_var Vdave, QdRMatrix_var Vadave, 
		 int nestsize, const char *msg, int constrain, double alpha, char *SCc_ID);

QdRMatrix_var AveVec(QdRMatrixSeq_var V, int nestsize);

QdRMatrix_var CrossVec(QdRMatrix_var V, QdRMatrix_var R);

QdRMatrix_var LeviCivita();

void ToSpreadsheet(const char *filename, QdRMatrix_var Normal, QdRMatrix_var U, int row,  bool nl);

QdRMatrix_var V_HofTel(QdRMatrixSeq_var B, QdRMatrixSeq_var V, int nestsize);

double DotVec(QdRMatrix_var V, QdRMatrix_var R);

QdRMatrix_var ScaleVec(double d, QdRMatrix_var V);

double vecsum(QdRMatrix_var Vector);

double vecsum2(QdRMatrix_var Vector);

double unit2SI(QdObject_var AnyQdObject);

double slope(QdRMatrixSeq_var X_in,QdRMatrixSeq_var Y_in,int n);

double correlationCoeff(QdRMatrixSeq_var X_in,QdRMatrixSeq_var Y_in,int n);

double weight_k_calc(QdRMatrix_var eval);

double sigma(QdRMatrixSeq_var V,QdRMatrix_var normal,int n);

double err_824(QdRMatrix_var B_ave, QdRMatrix_var Eigenvectors, QdRMatrix_var Eigenvalues, 
	       QdRMatrix_var dYij, QdRMatrix_var Q_MVA, double sigma, int nestsize, int modify);

double err_917(QdRMatrixSeq_var V, QdRMatrixSeq_var B, QdRMatrix_var Normal, int nestsize);

double minimizeD(int M, QdRMatrix_var V, QdRMatrixSeq_var vm, QdRMatrixSeq_var Bm);

double jumpvalue(QdRMatrixSeq_var V, const int n=5);

double jumpvalue(QdRScalarSeq_var S, const int n=5);

QdRMatrix_var Q_GEN_calc(QdRMatrixSeq_var TS);
