#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include <algorithm>

#include "qplug_if.h"
#include "qdutil.h"

#include "Moments_utils.h"

int Facet3D::currentStatus=1;
int Facet4D::currentStatus=1;

using namespace QSAS;

extern "C" QplugReturnStatus Moments(QplugArgList *);
extern "C" QplugReturnStatus Slices(QplugArgList *);
extern "C" QplugReturnStatus DensitySpectrogram(QplugArgList *);

QplugReturnStatus Moments(QplugArgList * call_list)
{
    QdObject_var data=(*call_list)[0];
    QdObject_var potential=(*call_list)[1];
    QdObject_var threshold=(*call_list)[2];
    QdObject_var gse_conversion=(*call_list)[3];
    QdObject_var direction=(*call_list)[3];
 
    QdRScalar_var method=QdRScalar_var::narrow((*call_list)[5]);
    
    if(method.is_nil()) method=new QdRScalar(0.);
            
    MomentsParameters para(data, potential, threshold, gse_conversion, direction);
    if(!para.init_ok)
    {
        return QPLUG_FAILURE;
    }
    
    QdRScalarSeq_var density;
    QdRMatrixSeq_var velocity;
    QdRMatrixSeq_var pressure;
    QdRMatrixSeq_var heat;
    QdRScalarSeq_var temperature;
//     QdRMatrixSeq_var pad;
    
    if(*method!=1.)
    {
        para.getMomentsSummed(MOM_DENSITY|MOM_VELOCITY|MOM_PRESSURE|MOM_HEATFLUX, density, velocity, pressure, heat, temperature);
    }
    else
    {
        para.getMomentsIntegrated(MOM_DENSITY|MOM_VELOCITY|MOM_PRESSURE|MOM_HEATFLUX, density, velocity, pressure, heat, temperature/*, pad*/);
    }
    
    call_list->push_back(density);
    call_list->push_back(velocity);
    call_list->push_back(pressure);
    call_list->push_back(heat);
    call_list->push_back(temperature);
//     call_list->push_back(pad);
    
    return QPLUG_SUCCESS;

}


QplugReturnStatus Slices(QplugArgList * call_list)
{
    QdObject_var data=(*call_list)[0];
    QdObject_var potential=new QdRScalar(0.);
    QdObject_var threshold=new QdRScalar(0.);
    QdObject_var gse_conversion=(*call_list)[1];
    
    QdRMatrixSeq_var centre_seq=QdRMatrixSeq_var::narrow((*call_list)[2]);
    QdRMatrix_var centre_const=QdRMatrix_var::narrow((*call_list)[2]);
    
    QdRMatrixSeq_var dir_seq=QdRMatrixSeq_var::narrow((*call_list)[3]);
    QdRMatrix_var dir_const=QdRMatrix_var::narrow((*call_list)[3]);
    
    QdObject_var data_direction=(*call_list)[4];
    
    QdRScalar_var resolution=QdRScalar_var::narrow((*call_list)[5]);
    
    MomentsParameters para(data, potential, threshold, gse_conversion, data_direction);
    
    
    if(!para.init_ok)
    {
        return QPLUG_FAILURE;
    }
    
    if(!centre_seq.is_nil())
    {
        int qar_status=QarTestJoined(centre_seq, data);
        if(qar_status==QAR_ERR)
        {
            QplugAppendTextDisplay("Centre and Data not joined.");
            return QPLUG_FAILURE;
        }
    }
    
    if(!dir_seq.is_nil())
    {
        int qar_status=QarTestJoined(dir_seq, data);
        if(qar_status==QAR_ERR)
        {
            QplugAppendTextDisplay("Direction and Data not joined.");
            return QPLUG_FAILURE;
        }
    }
    
    if(resolution.is_nil())
    {
        resolution=new QdRScalar(50);
    }
    else
    {
        if((double)(*resolution)<=1.)
            resolution=new QdRScalar(50);
    }
    
    QdRMatrixSeq_var result=new QdRMatrixSeq(para.data->sequence_size());
    
    //TODO check matrices dimensions
    
    QdRMatrix_var centre, direction;
    QdRScalarSeq_var scalemin, scalemax;
    centre=new QdRMatrix(Dim(3), 0.);
    direction=new QdRMatrix(Dim(3));
    (*direction)(0)=0.;
    (*direction)(1)=0.;
    (*direction)(2)=1.;
    
    QdRMatrix_var temp;
    
    scalemin=new QdRScalarSeq(para.data->sequence_size());
    scalemax=new QdRScalarSeq(para.data->sequence_size());
    
    for(int time=0; time<para.data->sequence_size(); time++)
    {
        if(time%100==0)
        {
            cout << floor(time/(para.data->sequence_size()-1.)*100.) << "%" << flush << "\r";
        }
        if(!centre_seq.is_nil())
        {
            centre=new QdRMatrix(centre_seq->elt(time));
        }
        else
        {
            if(!centre_const.is_nil())
            {
                centre=centre_const;
            }
        }
        
        if(!dir_seq.is_nil())
        {
            direction=new QdRMatrix(dir_seq->elt(time));
        }
        else
        {
            if(!dir_const.is_nil())
            {
                direction=dir_const;
            }
        }
        
        temp=para.computeSlice((int)(*resolution), time, centre, direction, 0.);
	//temp=para.computePAD(time, centre, direction, 0., 10000., 10, 10);
        
	//Computation of scale min and scale max
	bool found_not_fill=false;
	double min(-1.), max(-1.);
	for(int i=0; i<(int)(*resolution); ++i)
	{
		for(int j=0; j<(int)(*resolution); ++j)
		{
			if((*temp)(i,j)!=-1)
			{
				if(found_not_fill==false)
				{
					found_not_fill=true;
					min=(*temp)[i][j];
					max=(*temp)[i][j];
				}
				else
				{
					if(min>(*temp)[i][j]) min=(*temp)[i][j];
					if(max<(*temp)[i][j]) max=(*temp)[i][j];
				}
			}
		}
	}
	scalemin->elt(time)=min;
	scalemax->elt(time)=max;
	
        result->elt(time)=*temp;
	
        if(time==0)
        {
            result->copy_xrefs_from(temp);
        }
    }
    
    set_timetags(result, get_timetags(para.data));
    result->set_xref("DEPEND_0", get_timetags(para.data));
    result->set_xref("SCALEMIN", scalemin);
    result->set_xref("SCALEMAX", scalemax);
    
    cout << endl << "Done" << endl;
    
    call_list->push_back(result);
    
    return QPLUG_SUCCESS;
}


QplugReturnStatus DensitySpectrogram(QplugArgList * call_list)
{
    QdObject_var data=(*call_list)[0];
    QdObject_var potential=new QdRScalar(0.);
    QdObject_var threshold=new QdRScalar(0.);
    QdObject_var gse_conversion=new QdRScalar(0.);
    QdObject_var direction=new QdRScalar(0.);
    MomentsParameters para(data, potential, threshold, gse_conversion, direction);
    
    if(!para.init_ok)
    {
        return QPLUG_FAILURE;
    }
    
    int seq_size=para.data->sequence_size();
    
    QdRMatrixSeq_var result=new QdRMatrixSeq(seq_size/*, Dim(para.maxE+1)*/);
    QdRMatrixSeq_var e_axis=new QdRMatrixSeq(seq_size);
    QdRMatrixSeq_var e_axis_plus=new QdRMatrixSeq(seq_size);
    QdRMatrixSeq_var e_axis_minus=new QdRMatrixSeq(seq_size);
    
    double tot_density, density_ok;
    double val, vol;
    
    for(int time=0; time<seq_size; time++)
    {
        int count_no_fill=0;
        for(int e=0; e<=para.maxE; e++)
        {
            if(para.getE(time, e)>=0.) count_no_fill++;
        }
        
        int index=0;
        result->elt(time)=QdRMatrix(Dim(count_no_fill));
        e_axis->elt(time)=QdRMatrix(Dim(count_no_fill));
        e_axis_plus->elt(time)=QdRMatrix(Dim(count_no_fill));
        e_axis_minus->elt(time)=QdRMatrix(Dim(count_no_fill));
        
        for(int e=0; e<=para.maxE; e++)
        {
            if(para.getE(time, e)>=0.)
            {
                tot_density=0.;
                density_ok=false;
    
                for(int phi=0; phi<=para.maxPhi; phi++)
                {
                    for(int theta=0; theta<=para.maxTheta; theta++)
                    {
                        val=para.getData(time, e, phi, theta);
                        vol=para.getElementVolume(time, e, phi, theta);
    
                        if(val>=0. && vol>0.)
                        {
                            tot_density+=val*vol;
                            density_ok=true;
                        }
                    }
                }
                
                (*e_axis)[time][index]=para.getE(time, e)*6.24150974e18;
                (*e_axis_plus)[time][index]=(para.getEPlus(time, e)-para.getE(time, e))*6.24150974e18;
                (*e_axis_minus)[time][index]=(para.getE(time, e)-para.getEMinus(time, e))*6.24150974e18;
                
                if(density_ok)
                {
                    (*result)[time][index]=tot_density/1000000.;
                }
                else
                {
                    (*result)[time][index]=-1.;
                }
                index++;
            }
        }
    }
    
    set_timetags(result, get_timetags(data));
    set_timetags(e_axis, get_timetags(data));
    set_timetags(e_axis_plus, get_timetags(data));
    set_timetags(e_axis_minus, get_timetags(data));
    
    QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) result);
    QuSetTxtAttr("UNITS", "cm^-3", (QdObject_var) result);
    QuSetTxtAttr("FIELDNAM", "Density per energy range", (QdObject_var) result);
    QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) result);
    QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) result);
    QuSetTxtAttr("SI_CONVERSION", "1.0E6>m^-3", (QdObject_var) result);
    QuSetTxtAttr("DISPLAYTYPE", "Spectrogram", (QdObject_var) result);
    
    QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) e_axis);
    QuSetTxtAttr("UNITS", "eV", (QdObject_var) e_axis);
    QuSetTxtAttr("FIELDNAM", "Energy range", (QdObject_var) e_axis);
    QuSetTxtAttr("SCALETYP", "Log", (QdObject_var) e_axis);

    QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) e_axis_plus);
    QuSetTxtAttr("UNITS", "eV", (QdObject_var) e_axis_plus);
    QuSetTxtAttr("FIELDNAM", "Energy range", (QdObject_var) e_axis_plus);
    
    QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) e_axis_minus);
    QuSetTxtAttr("UNITS", "eV", (QdObject_var) e_axis_minus);
    QuSetTxtAttr("FIELDNAM", "Energy range", (QdObject_var) e_axis_minus);
    
    e_axis->set_xref("DELTA_PLUS", (QdObject_var)e_axis_plus);
    e_axis->set_xref("DELTA_MINUS", (QdObject_var)e_axis_minus);
    result->set_xref("DEPEND_1", (QdObject_var)e_axis);
    
//     if(!para.e_const.is_nil())
//     {
//         result->set_xref("DEPEND_1", (QdObject_var)para.e_const);
//     }
//     else
//     {
//         QdRMatrix_var temp=new QdRMatrix(para.e_seq[0]);
//         result->set_xref("DEPEND_1", (QdObject_var)temp);
//     }
    
    call_list->push_back(result);
    
    return QPLUG_SUCCESS;

}
