#include "Moments_utils.h"

void MomentsParameters::triangulation2(int time, list<Element3D>& tetra) // TODO GSE
{
    double valA, valB, valC, valD, valE, valF, valG, valH;
    Point4D *A, *B, *C, *D, *E, *F, *G, *H;
    double theta1, theta2, phi1, phi2, v1, v2;
    
    for(int e=0; e<=maxE; e++)
    {
        for(int phi=0; phi<=maxPhi; phi++)
        {
            for(int theta=0; theta<=maxTheta; theta++)
            {
                bool ok=true;
                valA=getData(time, e, phi, theta);
                if(valA==dataFillVal || getElementVolume(time, e, phi, theta)<0.) ok=false;
                valB=getData(time, e, phi, (theta+1)%(maxTheta+1));
                if(valB==dataFillVal || getElementVolume(time, e, phi, (theta+1)%maxTheta)<0.) ok=false;
                valC=getData(time, e, (phi+1)%(maxPhi+1), theta);
                if(valC==dataFillVal || getElementVolume(time, e, (phi+1)%maxPhi, theta)<0.) ok=false;
                valD=getData(time, e, (phi+1)%(maxPhi+1), (theta+1)%(maxTheta+1));
                if(valD==dataFillVal || getElementVolume(time, e, (phi+1)%maxPhi, (theta+1)%maxTheta)<0.) ok=false;
                valE=getData(time, (e+1)%(maxE+1), phi, theta);
                if(valE==dataFillVal || getElementVolume(time, (e+1)%maxE, phi, theta)<0.) ok=false;
                valF=getData(time, (e+1)%(maxE+1), phi, (theta+1)%(maxTheta+1));
                if(valF==dataFillVal || getElementVolume(time, (e+1)%maxE, phi, (theta+1)%maxTheta)<0.) ok=false;
                valG=getData(time, (e+1)%(maxE+1), (phi+1)%(maxPhi+1), theta);
                if(valG==dataFillVal || getElementVolume(time, (e+1)%maxE, (phi+1)%maxPhi, theta)<0.) ok=false;
                valH=getData(time, (e+1)%(maxE+1), (phi+1)%(maxPhi+1), (theta+1)%(maxTheta+1));
                if(valH==dataFillVal || getElementVolume(time, (e+1)%maxE, (phi+1)%maxPhi, (theta+1)%maxTheta)<0.) ok=false;
                
                if(ok)
                {
                    theta1=getTheta(time, theta);
                    theta2=getTheta(time, (theta+1)%(maxTheta+1));
                    phi1=getPhiMinus(time, phi);
                    phi2=getPhiPlus(time, (phi+1)%(maxPhi+1));
                    v1=getVMinus(time, e);
                    v2=getVPlus(time, (e+1)%(maxE+1));
                    
                    if(v1>0. && v2>0.)
                    {
                    
                        A=new Point4D(v1*v_sign*sin(theta1*theta2rad)*cos(phi1*phi2rad), v1*v_sign*sin(theta1*theta2rad)*sin(phi1*phi2rad), v1*v_sign*cos(theta1*theta2rad), 0., valA);
                        
                        B=new Point4D(v1*v_sign*sin(theta2*theta2rad)*cos(phi1*phi2rad), v1*v_sign*sin(theta2*theta2rad)*sin(phi1*phi2rad), v1*v_sign*cos(theta2*theta2rad), 0., valB);
                        
                        C=new Point4D(v1*v_sign*sin(theta1*theta2rad)*cos(phi2*phi2rad), v1*v_sign*sin(theta1*theta2rad)*sin(phi2*phi2rad), v1*v_sign*cos(theta1*theta2rad), 0., valC);
                        
                        D=new Point4D(v1*v_sign*sin(theta2*theta2rad)*cos(phi2*phi2rad), v1*v_sign*sin(theta2*theta2rad)*sin(phi2*phi2rad), v1*v_sign*cos(theta2*theta2rad), 0., valD);
                        
                        E=new Point4D(v2*v_sign*sin(theta1*theta2rad)*cos(phi1*phi2rad), v2*v_sign*sin(theta1*theta2rad)*sin(phi1*phi2rad), v2*v_sign*cos(theta1*theta2rad), 0., valE);
                        
                        F=new Point4D(v2*v_sign*sin(theta2*theta2rad)*cos(phi1*phi2rad), v2*v_sign*sin(theta2*theta2rad)*sin(phi1*phi2rad), v2*v_sign*cos(theta2*theta2rad), 0., valF);
                        
                        G=new Point4D(v2*v_sign*sin(theta1*theta2rad)*cos(phi2*phi2rad), v2*v_sign*sin(theta1*theta2rad)*sin(phi2*phi2rad), v2*v_sign*cos(theta1*theta2rad), 0., valG);
                        
                        H=new Point4D(v2*v_sign*sin(theta2*theta2rad)*cos(phi2*phi2rad), v2*v_sign*sin(theta2*theta2rad)*sin(phi2*phi2rad), v2*v_sign*cos(theta2*theta2rad), 0., valH);
                        
                        tetra.push_front(Element3D(A, B, C, E));
                        tetra.push_front(Element3D(B, E, F, H));
                        tetra.push_front(Element3D(B, C, D, H));
                        tetra.push_front(Element3D(C, E, G, H));
                        tetra.push_front(Element3D(B, C, E, H));
                        
//                         cout << time << "/" << e << "/" << phi << "/" << theta << "  -  " << Element3D(A, B, C, E).volume()+Element3D(B, E, F, H).volume()+Element3D(B, C, D, H).volume()+Element3D(C, E, G, H).volume()+Element3D(B, C, E, H).volume() << " - " << getElementVolume(time, e, phi, theta) << endl;
//                         cout << v1 << " " << v2 << "  -  " << theta1 << " " << theta2 << "  -  " << phi1 << " " << phi2 << endl << endl;
                        
                        delete A, B, C, D, E, F, G, H;
                    }
                }
        }
    }
    }
}

bool MomentsParameters::init_data(QdObject_var& obj)
{
    data=QdRMatrixSeq_var::narrow(obj);
    
    if(data.is_nil())
    {
        QplugAppendTextDisplay("Wrong data type for distribution.");
        return false;
    }
    
    QdObject_var temp;
    temp=QarConvertUnitsTo(data, "1>s^3 m^-6");
    data=QdRMatrixSeq_var::narrow(temp);
    if(data.is_nil())
    {
        QplugAppendTextDisplay("Failed to convert distribution to SI units.");
        return false;
    }
    
    dataFillVal=QString(QuGetAttrText("FILLVAL", data)).toDouble();

    charge=1.60217646e-19; //C
    mass=9.11e-31; // kg
    QString entity=QString(QuGetAttrText("ENTITY", data));
    if(entity.lower()=="proton")
    {
        mass=1.67262158e-27; //kg
    }
    else
    {
        if(entity.lower()=="electron")
        {
        }
        else
        {
            if(entity.lower()=="helium+")
            {
                mass=4.002602*1.660538782e-27;
            }
            else
            {
                if(entity.lower()=="alpha")
                {
                    charge=2.*1.60217646e-19; //C
                    mass=4.002602*1.660538782e-27;
                }
                else
                {
                    cout << "Unknown particle type - assuming electron" << endl;
                }
            }
        }
    }
    
    // Associating the xrefs to the coordinates
    
    bool initE(false), initPhi(false), initTheta(false);
    char strDepend[30];
    char strError[200];
    
    for(int i=1; i<=3; i++)
    {
        sprintf(strDepend, "DEPEND_%d", i);
        QdObject_var o;
        try
        {
            o=obj->get_xref(strDepend);
        }
        catch(Exception& e)
        {
            sprintf(strError, "Failed looking for %s", strDepend);
            QplugAppendTextDisplay(strError);
            return false;
        }
        
        try
        {
            QString units=QuGetAttrText("UNITS", o);
            if(units.contains("deg", false) || units.contains("rad", false))
            {
                double conv2rad=1.;
                if(units.contains("deg", false)) conv2rad=M_PI/180.;

                double fillVal=QString(QuGetAttrText("FILLVAL", o)).toDouble();

                bool isPhi=false;
                // Theta <=> every value is between 0 and Pi
                //TODO use REPRESENTATION_x as additional parameter to determine nature of angle
                QdRMatrixSeq_var o_seq=QdRMatrixSeq_var::narrow(o);
                double minVal, maxVal;
                if(!o_seq.is_nil())
                {
                    int n_el=o_seq->sequence_size();
                    int n_var=o_seq[0].dimdata()[0];
                    minVal=o_seq[0][0]*conv2rad;
                    maxVal=o_seq[0][0]*conv2rad;
                    for(int j=0; j<n_el; j++)
                    {
                        for(int k=0; k<n_var; k++)
                        {
                            if((double)o_seq[j][k]!=fillVal)
                            {
                                if(o_seq[j][k]*conv2rad<minVal) minVal=o_seq[j][k]*conv2rad;
                                if(o_seq[j][k]*conv2rad>maxVal) maxVal=o_seq[j][k]*conv2rad;
                            }
                        }
                    }
                }
                else
                {
                    QdRMatrix_var o_const=QdRMatrix_var::narrow(o);
                    if(!o.is_nil())
                    {
                        minVal=(*o_const)[0]*conv2rad;
                        maxVal=(*o_const)[0]*conv2rad;
                        int n_var=o_const->dimdata()[0];
                        for(int j=0; j<n_var; j++)
                        {
                            if((double)(*o_const)[j]!=fillVal)
                            {
                                if((*o_const)[j]*conv2rad<minVal) minVal=(*o_const)[j]*conv2rad;
                                if((*o_const)[j]*conv2rad>maxVal) maxVal=(*o_const)[j]*conv2rad;
                            }
                        }
                    }
                    else
                    {
                        sprintf(strError, "Wrong data type for %s", strDepend);
                        QplugAppendTextDisplay(strError);
                        return false;
                    }
                }
                
                double offsetTheta=0.;
                
                if(fabs(maxVal-minVal)>M_PI)
                {
                    isPhi=true;
                }
                else
                {
                    if(minVal<0. && minVal>=-M_PI/2.)
                    {
                        offsetTheta=M_PI/2.;
                    }
                }
                
                if(!isPhi) // Let's see if the variable name hints us...
                {
                    QString name=QuGetAttrText("FIELDNAM", o);
                    isPhi=(name.contains("phi") || name.contains("azimut"));
                }
                
                if(isPhi)
                {
                    phi_const=QdRMatrix_var::narrow(o);
                    phi_seq=QdRMatrixSeq_var::narrow(o);
                    
                    QdObject_var phiPlus=o->get_xref("DELTA_PLUS");
                    phi_plus_matrix_seq=QdRMatrixSeq_var::narrow(phiPlus);
                    phi_plus_matrix=QdRMatrix_var::narrow(phiPlus);
                    phi_plus_scalar_seq=QdRScalarSeq_var::narrow(phiPlus);
                    phi_plus_scalar=QdRScalar_var::narrow(phiPlus);
                    
                    if(phi_plus_matrix_seq.is_nil() && phi_plus_matrix.is_nil() && phi_plus_scalar_seq.is_nil() && phi_plus_scalar.is_nil())
                    {
                        QplugAppendTextDisplay("Invalid DELTA_PLUS for phi");
                        return false;
                    }
                    else
                    {
                        if(!phi_plus_matrix_seq.is_nil()) type_phi_plus=MATRIX_SEQ;
                        if(!phi_plus_matrix.is_nil()) type_phi_plus=MATRIX;
                        if(!phi_plus_scalar_seq.is_nil()) type_phi_plus=SCALAR_SEQ;
                        if(!phi_plus_scalar.is_nil()) type_phi_plus=SCALAR;
                    }
                    
                    QdObject_var phiMinus=o->get_xref("DELTA_MINUS");
                    phi_minus_matrix_seq=QdRMatrixSeq_var::narrow(phiMinus);
                    phi_minus_matrix=QdRMatrix_var::narrow(phiMinus);
                    phi_minus_scalar_seq=QdRScalarSeq_var::narrow(phiMinus);
                    phi_minus_scalar=QdRScalar_var::narrow(phiMinus);
                    
                    if(phi_minus_matrix_seq.is_nil() && phi_minus_matrix.is_nil() && phi_minus_scalar_seq.is_nil() && phi_minus_scalar.is_nil())
                    {
                        QplugAppendTextDisplay("Invalid DELTA_MINUS for phi");
                        return false;
                    }
                    else
                    {
                        if(!phi_minus_matrix_seq.is_nil()) type_phi_minus=MATRIX_SEQ;
                        if(!phi_minus_matrix.is_nil()) type_phi_minus=MATRIX;
                        if(!phi_minus_scalar_seq.is_nil()) type_phi_minus=SCALAR_SEQ;
                        if(!phi_minus_scalar.is_nil()) type_phi_minus=SCALAR;
                    }
                    
                    phiFillVal=fillVal;
                    phi2rad=conv2rad;
                    initPhi=true;
                    indexPhi=i-1;
                    
                    sprintf(strError, "Using \"%s\" as phi coordinate", QuGetAttrText("FIELDNAM", o));
                    QplugAppendTextDisplay(strError);
                }
                else
                {
                    theta_const=QdRMatrix_var::narrow(o);
                    theta_seq=QdRMatrixSeq_var::narrow(o);
                    
                    if(!theta_const.is_nil()){
                        (*theta_const)+=offsetTheta/conv2rad;
                        if(offsetTheta==M_PI/2)
                        {
                            (*theta_const)=-(*theta_const);
                        }
                    }
                    if(!theta_seq.is_nil())
                    {
                        (*theta_seq)+=offsetTheta/conv2rad;
                        if(offsetTheta==M_PI/2)
                        {
                            (*theta_seq)=-(*theta_seq);
                        }
                    }
                    
                    QdObject_var thetaPlus=o->get_xref("DELTA_PLUS");
                    theta_plus_matrix_seq=QdRMatrixSeq_var::narrow(thetaPlus);
                    theta_plus_matrix=QdRMatrix_var::narrow(thetaPlus);
                    theta_plus_scalar_seq=QdRScalarSeq_var::narrow(thetaPlus);
                    theta_plus_scalar=QdRScalar_var::narrow(thetaPlus);
                    
                    if(theta_plus_matrix_seq.is_nil() && theta_plus_matrix.is_nil() && theta_plus_scalar_seq.is_nil() && theta_plus_scalar.is_nil())
                    {
                        QplugAppendTextDisplay("Invalid DELTA_PLUS for theta");
                        return false;
                    }
                    else
                    {
                        if(!theta_plus_matrix_seq.is_nil()) type_theta_plus=MATRIX_SEQ;
                        if(!theta_plus_matrix.is_nil()) type_theta_plus=MATRIX;
                        if(!theta_plus_scalar_seq.is_nil()) type_theta_plus=SCALAR_SEQ;
                        if(!theta_plus_scalar.is_nil()) type_theta_plus=SCALAR;
                    }
                    
                    QdObject_var thetaMinus=o->get_xref("DELTA_MINUS");
                    theta_minus_matrix_seq=QdRMatrixSeq_var::narrow(thetaMinus);
                    theta_minus_matrix=QdRMatrix_var::narrow(thetaMinus);
                    theta_minus_scalar_seq=QdRScalarSeq_var::narrow(thetaMinus);
                    theta_minus_scalar=QdRScalar_var::narrow(thetaMinus);
                    
                    if(theta_minus_matrix_seq.is_nil() && theta_minus_matrix.is_nil() && theta_minus_scalar_seq.is_nil() && theta_minus_scalar.is_nil())
                    {
                        QplugAppendTextDisplay("Invalid DELTA_MINUS for theta");
                        return false;
                    }
                    else
                    {
                        if(!theta_minus_matrix_seq.is_nil()) type_theta_minus=MATRIX_SEQ;
                        if(!theta_minus_matrix.is_nil()) type_theta_minus=MATRIX;
                        if(!theta_minus_scalar_seq.is_nil()) type_theta_minus=SCALAR_SEQ;
                        if(!theta_minus_scalar.is_nil()) type_theta_minus=SCALAR;
                    }
                    
                    thetaFillVal=fillVal;
                    theta2rad=conv2rad;
                    initTheta=true;
                    indexTheta=i-1;
                    
                    sprintf(strError, "Using \"%s\" as theta coordinate", QuGetAttrText("FIELDNAM", o));
                    QplugAppendTextDisplay(strError);
                }
            }
            else
            {
                // We suppose that it is E
                e_const=QdRMatrix_var::narrow(o);
                e_seq=QdRMatrixSeq_var::narrow(o);
                
                //TODO check conversion success
                if(!e_seq.is_nil())
                {
                    temp=QarConvertUnitsTo(e_seq, "1>J");
                    e_seq=QdRMatrixSeq_var::narrow(temp);
                    eFillVal=QString(QuGetAttrText("FILLVAL", e_seq)).toDouble();
                }
                if(!e_const.is_nil())
                {
                    temp=QarConvertUnitsTo(e_const, "1>J");
                    e_const=QdRMatrix_var::narrow(temp);
                    eFillVal=QString(QuGetAttrText("FILLVAL", e_const)).toDouble();
                }

                QdObject_var ePlus=(e_seq.is_nil() ? e_const->get_xref("DELTA_PLUS") : e_seq->get_xref("DELTA_PLUS"));
                e_plus_seq=QdRMatrixSeq_var::narrow(ePlus);
                e_plus_const=QdRMatrix_var::narrow(ePlus);

                if(e_plus_seq.is_nil() && e_plus_const.is_nil())
                {
                    QplugAppendTextDisplay("Invalid DELTA_PLUS for energy");
                    return false;
                }
                
                QdObject_var rMinus=(e_seq.is_nil() ? e_const->get_xref("DELTA_MINUS") : e_seq->get_xref("DELTA_MINUS"));
                e_minus_seq=QdRMatrixSeq_var::narrow(rMinus);
                e_minus_const=QdRMatrix_var::narrow(rMinus);

                if(e_plus_seq.is_nil() && e_plus_const.is_nil())
                {
                    QplugAppendTextDisplay("Invalid DELTA_MINUS for energy");
                    return false;
                }

                initE=true;
                indexE=i-1;
                
                sprintf(strError, "Using \"%s\" as energy coordinate", QuGetAttrText("FIELDNAM", o));
                QplugAppendTextDisplay(strError);
            }
            
        }
        catch(Exception& e)
        {
            cout << e.msg() << endl;
            return false;
        }
    }
    
    if(!initE || !initPhi || !initTheta)
    {
        QplugAppendTextDisplay("Failed finding the 3 coordinates");
        return false;
    }
    
    Dim mDim=data[0].dimdata();
    if(!mDim[indexE]==(e_const.is_nil() ? e_seq[0].dimdata()[0] : e_const->dimdata()[0]))
    { 
        QplugAppendTextDisplay("Dimensions for data and energy coordinates do not match");
        return false;
    }
    if(!mDim[indexTheta]==(theta_const.is_nil() ? theta_seq[0].dimdata()[0] :
        theta_const->dimdata()[0]))
    {
        QplugAppendTextDisplay("Dimensions for data and theta coordinates do not match");
        return false;
    }
    if(!mDim[indexPhi]==(phi_const.is_nil() ? phi_seq[0].dimdata()[0] : phi_const->dimdata()[0]))
    {
        QplugAppendTextDisplay("Dimensions for data and Phi coordinates do not match");
        return false;
    }
    
    maxTime=data->sequence_size()-1;
    maxE=mDim[indexE]-1;
    maxTheta=mDim[indexTheta]-1;
    maxPhi=mDim[indexPhi]-1;
    
    return true;
}

bool MomentsParameters::init_potential(QdObject_var & obj)
{
    potential=QdRScalarSeq_var::narrow(obj);
    if(!potential.is_nil())
    {
        QdObject_var temp=QarConvertUnitsTo(potential, "1 > V");
        if(temp.is_nil())
        {
            QplugAppendTextDisplay("Potential conversion to Volts failed - Assuming values are in Volts");
        }
        else potential=QdRScalarSeq_var::narrow(temp);
        
        int qar_status=QarTestJoined(potential, data);
        if(qar_status==QAR_ERR)
        {
            QplugAppendTextDisplay("Potential and Data not joined.");
            return false;
        }
        //TODO potFillVal
    }
    else
    {
        QdRScalar_var pot_scalar=QdRScalar_var::narrow(obj);
        if(!pot_scalar.is_nil())
        {
            potential=new QdRScalarSeq(data->sequence_size(), *pot_scalar);
        }
        else
        {
            potential=new QdRScalarSeq(data->sequence_size(), 0.);
        }
    }
    
    return true;
}

bool MomentsParameters::init_threshold(QdObject_var & obj)
{
    threshold=QdRScalarSeq_var::narrow(obj);
    if(!threshold.is_nil())
    {
        QdObject_var temp=QarConvertUnitsTo(threshold, "1 > J");
        if(temp.is_nil())
        {
            QplugAppendTextDisplay("Threshold conversion to Joules failed");
            return false;
        }
        else threshold=QdRScalarSeq_var::narrow(temp);
        
        int qar_status=QarTestJoined(threshold, data);
        if(qar_status==QAR_ERR)
        {
            QplugAppendTextDisplay("Threshold and Data not joined.");
            return false;
        }
    }
    else
    {
        QdRScalar_var thr_scalar=QdRScalar_var::narrow(obj);
        if(!thr_scalar.is_nil())
        {
            threshold=new QdRScalarSeq(data->sequence_size(), (*thr_scalar)*1.60217646e-19);
        }
        else
        {
            threshold=new QdRScalarSeq(data->sequence_size(), 0.);
        }
    }
    
    return true;
}

bool MomentsParameters::init_gse(QdObject_var& obj)
{
    gse_seq=QdRMatrixSeq_var::narrow(obj);
    if(!gse_seq.is_nil())
    {
        int qar_status=QarTestJoined(gse_seq, data);
        if(qar_status==QAR_ERR)
        {
            QplugAppendTextDisplay("GSE sequence and Data not joined.");
            return false;
        }
        
        if(gse_seq[0].rank()!=2)
        {
            QplugAppendTextDisplay("Wrong dimensions for GSE.");
            return false;
        }
        
        if(gse_seq[0].dimdata()[0]!=3 || gse_seq[0].dimdata()[1]!=3)
        {
            QplugAppendTextDisplay("Wrong dimensions for GSE.");
            return false;
        }
        
        type_gse=MATRIX_SEQ;
        return true;
    }
    
    gse_const=QdRMatrix_var::narrow(obj);
    if(!gse_const.is_nil())
    {
        if(gse_const->rank()!=2)
        {
            QplugAppendTextDisplay("Wrong dimensions for GSE.");
            return false;
        }
        
        if(gse_const->dimdata()[0]!=3 || gse_const->dimdata()[1]!=3)
        {
            QplugAppendTextDisplay("Wrong dimensions for GSE.");
            return false;
        }
        
        type_gse=MATRIX;
        return true;
    }
    
    type_gse=SCALAR;
    
    return true;
}

//TODO take into account SCALEMIN SCALEMAX SCALETYP
MomentsParameters::MomentsParameters(QdObject_var& obj, QdObject_var& potential_in, QdObject_var& threshold_in, QdObject_var gse_in, QdObject_var direction_in, bool _logData, bool _logE) : init_ok(false), logData(_logData), logE(_logE), phi2rad(1.), theta2rad(1.), time_tetra_cache(-1)
{
    tetra.clear();
    
    if(!init_data(obj)) return;
    if(!init_potential(potential_in)) return;
    if(!init_threshold(threshold_in)) return;
    if(!init_gse(gse_in)) return;
    
    QdRScalar_var dir_scal=QdRScalar_var::narrow(direction_in);
    
    if(dir_scal.is_nil()) v_sign=-1.;
    else
    {
        if(*dir_scal==1.) v_sign=1.;
        else v_sign=-1.;
    }
    
    init_ok=true;
}

// double MomentsParameters::getIntegratedValue(int time)
// {
//     double res(0.);
//     bool data_ok=false;
//     double vol;
//     double val;
//     
//     for(int r=0; r<=maxE; r++)
//     {
//         for(int phi=0; phi<=maxPhi; phi++)
//         {
//             for(int theta=0; theta<=maxTheta; theta++)
//             {
//                 vol=getElementVolume(time, r, phi, theta);
//                 val=getData(time, r, phi, theta);
//                 if(vol>=0. && val!=dataFillVal)
//                 {
//                     res+=vol*val;
//                     data_ok=true;
//                 }
//                 else if(vol>=0. && val==dataFillVal) cout << "data gap" << endl;
//             }
//         }
//     }
//     
//     if(!data_ok) return -1.;
//     
//     return res;
// }

double MomentsParameters::getData(int time, int r, int phi, int theta)
{
    int index[3];
    index[indexE]=r;
    index[indexPhi]=phi;
    index[indexTheta]=theta;

    return data[time][index[0]][index[1]][index[2]];
}

QdRMatrix_var MomentsParameters::computeSlice(int resolution, int time, QdRMatrix_var& CoM, QdRMatrix_var& direction, double offset)
{
    if(CoM.is_nil() || direction.is_nil())
    {
        cout << "MomentsParameters::computeSlice: nil centerOfMass and/or direction" << endl;
        return QdRMatrix_var();
    }
    
    if(CoM->dimdata()[0]!=3 || direction->dimdata()[0]!=3)
    {
        cout << "MomentsParameters::computeSlice: wrong dimension for centerOfMass and/or direction" << endl;
        return QdRMatrix_var();
    }

    double val;
    double max_v=0.;
    for(int e=0; e<=maxE; e++)
    {
        double v=getVPlus(time, e);
        if(v>max_v) max_v=v;
    }
    
    if(time_tetra_cache!=time)
    {
        time_tetra_cache=time;
        buildTetra(time);
    }

    QdRMatrix_var res=new QdRMatrix(Dim(resolution, resolution));
    double element_size=2.*max_v/resolution;
    
    // Find rotation matrix
    
    QdRMatrix_var rotation=new QdRMatrix(Dim(3,3));
    getRotationMatrix(direction, rotation);

    if(!tetra.empty())
    {
        list<Point3D> points;
        for(int i=0; i<resolution; i++)
        {
            for(int j=0; j<resolution; j++)
            {
                double _x=(i+0.5)*element_size-max_v;
                double _y=(j+0.5)*element_size-max_v;
                double _z=offset;
               
                double x=(*rotation)[0][0]*_x+(*rotation)[0][1]*_y+(*rotation)[0][2]*_z+(*CoM)[0];
                double y=(*rotation)[1][0]*_x+(*rotation)[1][1]*_y+(*rotation)[1][2]*_z+(*CoM)[1];
                double z=(*rotation)[2][0]*_x+(*rotation)[2][1]*_y+(*rotation)[2][2]*_z+(*CoM)[2];
                
                Point3D temp(x, y, z, -1.);
                temp.orig_x=i;
                temp.orig_y=j;
                points.push_front(temp);
                
            }
        }
        list<Element3D> tetra_temp(tetra);
        interpolation(points, tetra_temp);
        
        for(list<Point3D>::iterator iter=points.begin(); iter!=points.end(); iter++)
        {
            (*res)[(int)(iter->orig_x)][(int)(iter->orig_y)]=iter->val;
        }
        
        // Creation of the metadata
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) res);
        QuSetTxtAttr("UNITS", "s^3/m^6", (QdObject_var) res);
        char fieldnam[100];
        sprintf(fieldnam, "%s - Slice", QuGetAttrText(FIELDNAM, data));
        QuSetTxtAttr("FIELDNAM", fieldnam, (QdObject_var) res);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) res);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) res);
        QuSetTxtAttr("SI_CONVERSION", "1>s^3 m^-6", (QdObject_var) res);
        QuSetTxtAttr("DISPLAYTYPE", "Spectrogram", (QdObject_var) res);
        
        QdRMatrix_var axis=new QdRMatrix(Dim(resolution));
        for(int i=0; i<resolution; i++)
        {
            (*axis)[i]=(i+0.5)*element_size-max_v;
        }
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) axis);
        QuSetTxtAttr("UNITS", "m s-1", (QdObject_var) axis);
        sprintf(fieldnam, "%s - Slice - Axis", QuGetAttrText(FIELDNAM, data));
        QuSetTxtAttr("FIELDNAM", fieldnam, (QdObject_var) axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) axis);
        QuSetTxtAttr("SI_CONVERSION", "1>m s^-1", (QdObject_var) axis);
        
        QdRScalar_var delta_axis=new QdRScalar(element_size/2.);
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) delta_axis);
        QuSetTxtAttr("UNITS", "m s-1", (QdObject_var) delta_axis);
        sprintf(fieldnam, "%s - Slice - Axis Delta", QuGetAttrText(FIELDNAM, data));
        QuSetTxtAttr("FIELDNAM", fieldnam, (QdObject_var) delta_axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) delta_axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) delta_axis);
        QuSetTxtAttr("SI_CONVERSION", "1>m s^-1", (QdObject_var) delta_axis);
        
        axis->set_xref("DELTA_PLUS", (QdObject_var)delta_axis);
        axis->set_xref("DELTA_MINUS", (QdObject_var)delta_axis);
        res->set_xref("DEPEND_1", (QdObject_var)axis);
        res->set_xref("DEPEND_2", (QdObject_var)axis);

     }
    else
    {
        return QdRMatrix_var();
    }
    return res;
}

QdRMatrix_var MomentsParameters::computeValuesAlongVector(int time, QdRMatrix_var& CoM, QdRMatrix_var& direction, double v_min, double v_max, int nb_values)
{
    if(CoM.is_nil() || direction.is_nil())
    {
        cout << "MomentsParameters::computeValuesAlongVector: nil centerOfMass and/or direction" << endl;
        return QdRMatrix_var();
    }
    
    if(CoM->dimdata()[0]!=3 || direction->dimdata()[0]!=3)
    {
        cout << "MomentsParameters::computeValuesAlongVector: wrong dimension for centerOfMass and/or direction" << endl;
        return QdRMatrix_var();
    }
    
    double val;
    
    if(time_tetra_cache!=time)
    {
        time_tetra_cache=time;
        buildTetra(time);
    }
    
    QdRMatrix_var res=new QdRMatrix(Dim(nb_values));
    double element_size=(v_max-v_min)/nb_values;
    
    QdRMatrix_var rotation=new QdRMatrix(Dim(3,3));
    getRotationMatrix(direction, rotation);
    
    if(!tetra.empty())
    {
        list<Point3D> points;
        for(int i=0; i<nb_values; i++)
        {
            double _z=v_min+(i+0.5)*element_size;
            double x=(*rotation)[0][2]*_z+(*CoM)[0];
            double y=(*rotation)[1][2]*_z+(*CoM)[1];
            double z=(*rotation)[2][2]*_z+(*CoM)[2];
            
            Point3D temp(x, y, z, -1.);
            temp.orig_x=i;
            points.push_front(temp);
        }
        
        list<Element3D> tetra_temp(tetra);
        interpolation(points, tetra_temp);
        
        for(list<Point3D>::iterator iter=points.begin(); iter!=points.end(); iter++)
        {
            (*res)[(int)(iter->orig_x)]=iter->val;
        }
        
        // Creation of the metadata
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) res);
        QuSetTxtAttr("UNITS", "s^3/m^6", (QdObject_var) res);
        char fieldnam[100];
        sprintf(fieldnam, "%s - Values along vector", QuGetAttrText(FIELDNAM, data));
        QuSetTxtAttr("FIELDNAM", fieldnam, (QdObject_var) res);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) res);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) res);
        QuSetTxtAttr("SI_CONVERSION", "1>s^3 m^-6", (QdObject_var) res);
//         QuSetTxtAttr("DISPLAYTYPE", "Spectrogram", (QdObject_var) res);
        
        QdRMatrix_var axis=new QdRMatrix(Dim(nb_values));
        for(int i=0; i<nb_values; i++)
        {
            (*axis)[i]=(i+0.5)*element_size+v_min;
        }
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) axis);
        QuSetTxtAttr("UNITS", "m s-1", (QdObject_var) axis);
        sprintf(fieldnam, "%s - Values along vector - Axis", QuGetAttrText(FIELDNAM, data));
        QuSetTxtAttr("FIELDNAM", fieldnam, (QdObject_var) axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) axis);
        QuSetTxtAttr("SI_CONVERSION", "1>m s^-1", (QdObject_var) axis);
        
        QdRScalar_var delta_axis=new QdRScalar(element_size/2.);
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) delta_axis);
        QuSetTxtAttr("UNITS", "m s-1", (QdObject_var) delta_axis);
        sprintf(fieldnam, "%s - Values along vector - Axis Delta", QuGetAttrText(FIELDNAM, data));
        QuSetTxtAttr("FIELDNAM", fieldnam, (QdObject_var) delta_axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) delta_axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) delta_axis);
        QuSetTxtAttr("SI_CONVERSION", "1>m s^-1", (QdObject_var) delta_axis);
        
        axis->set_xref("DELTA_PLUS", (QdObject_var)delta_axis);
        axis->set_xref("DELTA_MINUS", (QdObject_var)delta_axis);
        res->set_xref("DEPEND_1", (QdObject_var)axis);
    }
    else
    {
        return QdRMatrix_var();
    }

    return res;
}

/*void MomentsParameters::plot(char* filename, int time)
{
    ofstream output(filename, ios::out | ios::trunc);
    
    double max_val=0.;
    double min_val=0.;
    double val;
    
    for(int e=0; e<=maxE; e++)
    {
        for(int phi=0; phi<=maxPhi; phi++)
        {
            for(int theta=0; theta<=maxTheta; theta++)
            {
                val=getData(time, e, phi, theta);
                if(val!=dataFillVal)
                {
                    if(val>max_val) max_val=val;
                    if(min_val<=0. || (min_val>0. && val<min_val)) min_val=val;
                }
            }
        }
    }

    for(int e=0; e<=maxE; e++)
    {
        for(int phi=0; phi<=maxPhi; phi++)
        {
            for(int theta=0; theta<=maxTheta; theta++)
            {
                val=getData(time, e, phi, theta);
                double norm_v=getV(time, e);
                double _theta(getTheta(time, theta)), _phi(getPhi(time, phi));
                if(norm_v<=0. && _theta!=thetaFillVal && _phi!=phiFillVal)
                {
                    double x=norm_v*sin(_theta*theta2rad)*cos(_phi*phi2rad);
                    double y=norm_v*sin(_theta*theta2rad)*sin(_phi*phi2rad);
                    double z=norm_v*cos(_theta*theta2rad);
                    if(val!=dataFillVal)
                    {
                        output << "vtk_plot3(["<< x << "], [" << y << "], [" << z << "], 'MarkerSize', "<< 5.e6 <<", 'MarkerColor', [" << (log(val)-log(min_val))/(log(max_val)-log(min_val)) << ", 0, 0])" << endl;
                    }
                }
            }
        }
    }
    
    output.close();
}*/

QdRMatrix_var MomentsParameters::computePAD(int time, QdRMatrix_var& CoM, QdRMatrix_var& direction, double v_min, double v_max, int nb_v, int nb_theta)
{
    if(CoM.is_nil() || direction.is_nil())
    {
        cout << "MomentsParameters::computePAD: nil centerOfMass and/or direction" << endl;
        return QdRMatrix_var();
    }
    
    if(CoM->dimdata()[0]!=3 || direction->dimdata()[0]!=3)
    {
        cout << "MomentsParameters::computePAD: wrong dimension for centerOfMass and/or direction" << endl;
        return QdRMatrix_var();
    }
    
    if(time_tetra_cache!=time)
    {
        time_tetra_cache=time;
        buildTetra(time);
    }
    
    QdRMatrix_var res=new QdRMatrix(Dim(nb_v, nb_theta));
    *res=-1.;
    double element_size_v=(v_max-v_min)/nb_v;
    
    // Find rotation matrix
    
    QdRMatrix_var rotation=new QdRMatrix(Dim(3,3));
    getRotationMatrix(direction, rotation);
    
    // End find rotation matrix
    
    if(!tetra.empty())
    {
        list<Point3D> points;
        for(int i=0; i<nb_v; i++)
        {
            double v=v_min+(i+0.5)*element_size_v;
            for(int j=0; j<nb_theta; j++)
            {
                double theta=(M_PI/nb_theta)*(j+0.5);
                double total=0.;
                bool foundOnce=false;
                
                for(int k=0; k<360; k+=10)
                {
                    double phi=M_PI/180.*(k+5.);
                    double x=v*sin(theta)*cos(phi);
                    double y=v*sin(theta)*sin(phi);
                    double z=v*cos(theta);
                    
                    Point3D temp(x, y, z, -1.);
                    temp.orig_x=i;
                    temp.orig_y=j;
                    points.push_front(temp);
                }
               
            }
        }
        
        list<Element3D> tetra_temp(tetra);
        interpolation(points, tetra_temp);
                
        for(list<Point3D>::iterator iter=points.begin(); iter!=points.end(); iter++)
        {
            if(iter->val>=0.)
            {
                if((*res)[(int)(iter->orig_x)][(int)(iter->orig_y)]>=0.)
                    (*res)[(int)(iter->orig_x)][(int)(iter->orig_y)]+=iter->val*10.;
                else
                    (*res)[(int)(iter->orig_x)][(int)(iter->orig_y)]=iter->val*10.;
            }
        }
        
        // Creation of the metadata
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) res);
        QuSetTxtAttr("UNITS", "s^3 rad/m^6", (QdObject_var) res);
        char fieldnam[100];
        sprintf(fieldnam, "%s - PAD", QuGetAttrText(FIELDNAM, data));
        QuSetTxtAttr("FIELDNAM", fieldnam, (QdObject_var) res);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) res);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) res);
        QuSetTxtAttr("SI_CONVERSION", "1>s^3 rad m^-6", (QdObject_var) res);
        QuSetTxtAttr("DISPLAYTYPE", "Spectrogram", (QdObject_var) res);
        
        QdRMatrix_var v_axis=new QdRMatrix(Dim(nb_v));
        for(int i=0; i<nb_v; i++)
        {
            (*v_axis)[i]=v_min+(i+0.5)*element_size_v;
        }
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) v_axis);
        QuSetTxtAttr("UNITS", "m s-1", (QdObject_var) v_axis);
        sprintf(fieldnam, "%s - PAD - V Axis", QuGetAttrText(FIELDNAM, data));
        QuSetTxtAttr("FIELDNAM", fieldnam, (QdObject_var) v_axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) v_axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) v_axis);
        QuSetTxtAttr("SI_CONVERSION", "1>m s^-1", (QdObject_var) v_axis);
        
        QdRScalar_var delta_v_axis=new QdRScalar(element_size_v/2.);
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) delta_v_axis);
        QuSetTxtAttr("UNITS", "m s-1", (QdObject_var) delta_v_axis);
        sprintf(fieldnam, "%s - PAD - V Axis Delta", QuGetAttrText(FIELDNAM, data));
        QuSetTxtAttr("FIELDNAM", fieldnam, (QdObject_var) delta_v_axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) delta_v_axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) delta_v_axis);
        QuSetTxtAttr("SI_CONVERSION", "1>m s^-1", (QdObject_var) delta_v_axis);
        
        QdRMatrix_var t_axis=new QdRMatrix(Dim(nb_theta));
        for(int i=0; i<nb_theta; i++)
        {
            (*t_axis)[i]=(M_PI/nb_theta)*(i+0.5);
        }
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) t_axis);
        QuSetTxtAttr("UNITS", "rad", (QdObject_var) t_axis);
        sprintf(fieldnam, "%s - PAD - Theta Axis", QuGetAttrText(FIELDNAM, data));
        QuSetTxtAttr("FIELDNAM", fieldnam, (QdObject_var) t_axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) t_axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) t_axis);
        QuSetTxtAttr("SI_CONVERSION", "1>rad", (QdObject_var) t_axis);
        
        QdRScalar_var delta_t_axis=new QdRScalar(5./180.*M_PI);
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) delta_t_axis);
        QuSetTxtAttr("UNITS", "rad", (QdObject_var) delta_t_axis);
        sprintf(fieldnam, "%s - PAD - Theta Axis Delta", QuGetAttrText(FIELDNAM, data));
        QuSetTxtAttr("FIELDNAM", fieldnam, (QdObject_var) delta_t_axis);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) delta_t_axis);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) delta_t_axis);
        QuSetTxtAttr("SI_CONVERSION", "1>rad", (QdObject_var) delta_t_axis);
        
        v_axis->set_xref("DELTA_PLUS", (QdObject_var)delta_v_axis);
        v_axis->set_xref("DELTA_MINUS", (QdObject_var)delta_v_axis);
        t_axis->set_xref("DELTA_PLUS", (QdObject_var)delta_t_axis);
        t_axis->set_xref("DELTA_MINUS", (QdObject_var)delta_t_axis);
        res->set_xref("DEPEND_1", (QdObject_var)v_axis);
        res->set_xref("DEPEND_2", (QdObject_var)t_axis);
    }
    else
    {
        return QdRMatrix_var();
    }
    
    return res;
}

void MomentsParameters::buildTetra(int time)
{
    time_tetra_cache=time;
    tetra.clear();
    list<Point3D> points;
    double val, vol;
    
    for(int phi=0; phi<=maxPhi; phi++)
    {
        for(int theta=0; theta<=maxTheta; theta++)
        {
            double max_v(0.), val_max_v(0.), vol_max_v(0.);
            double min_v(1.), val_min_v(0.), vol_min_v(0.);
            double _theta(getTheta(time, theta)), _phi(getPhi(time, phi));
            
            for(int e=0; e<=maxE; e++)
            {
                val=getData(time, e, phi, theta);
                vol=getElementVolume(time, e, phi, theta);
                if(val!=dataFillVal && vol>0.)
                {
                    double _v(getV(time, e));
                    if(_v>=0.)
                    {
                        double x=_v*v_sign*sin(_theta*theta2rad)*cos(_phi*phi2rad);
                        double y=_v*v_sign*sin(_theta*theta2rad)*sin(_phi*phi2rad);
                        double z=_v*v_sign*cos(_theta*theta2rad);
                        
                        convertToGSE(time, x, y, z);
    
                        Point3D temp(x, y, z, val, vol);
                        points.push_front(temp);
                        
                        if(_v>max_v)
                        {
                            max_v=getVPlus(time, e);
                            val_max_v=val;
                            vol_max_v=vol;
                        }
                        if(_v<min_v || min_v>0.)
                        {
                            min_v=getVMinus(time, e);
                            val_min_v=val;
                            vol_min_v=vol;
                        }
                    }
                }
            }
            
            double x=max_v*v_sign*sin(_theta*theta2rad)*cos(_phi*phi2rad);
            double y=max_v*v_sign*sin(_theta*theta2rad)*sin(_phi*phi2rad);
            double z=max_v*v_sign*cos(_theta*theta2rad);
            convertToGSE(time, x, y, z);
            Point3D temp_max(x, y, z, val_max_v, vol_max_v);
            points.push_front(temp_max);
            
            x=min_v*v_sign*sin(_theta*theta2rad)*cos(_phi*phi2rad);
            y=min_v*v_sign*sin(_theta*theta2rad)*sin(_phi*phi2rad);
            z=min_v*v_sign*cos(_theta*theta2rad);
            convertToGSE(time, x, y, z);
            Point3D temp_min(x, y, z, val_min_v, vol_min_v);
            points.push_front(temp_min);
                    
        }
    }
        
    if(points.size()>4)
    {
         triangulation(points, tetra);
    }
}

void MomentsParameters::buildTetra2(int time)
{
    time_tetra_cache=time;
    tetra.clear();
    triangulation2(time, tetra);
}

void getRotationMatrix(QdRMatrix_var& direction, QdRMatrix_var& rotation)
{
    double norm=sqrt((*direction)[0]*(*direction)[0]+(*direction)[1]*(*direction)[1]+(*direction)[2]*(*direction)[2]);
    (*direction)[0]/=norm;
    (*direction)[1]/=norm;
    (*direction)[2]/=norm;
    
    if(fabs((*direction)[0])<0.01 && fabs((*direction)[1])<0.01)
    {
        if((*direction)[2]>0.)
        {
            (*rotation)[0][0]=1.;
            (*rotation)[1][1]=1.;
            (*rotation)[2][2]=1.;
            
            (*rotation)[0][1]=0.;
            (*rotation)[0][2]=0.;
            (*rotation)[1][0]=0.;
            (*rotation)[1][2]=0.;
            (*rotation)[2][0]=0.;
            (*rotation)[2][1]=0.;
        }
        else
        {
            (*rotation)[0][0]=-1.;
            (*rotation)[1][1]=-1.;
            (*rotation)[2][2]=-1.;
            
            (*rotation)[0][1]=0.;
            (*rotation)[0][2]=0.;
            (*rotation)[1][0]=0.;
            (*rotation)[1][2]=0.;
            (*rotation)[2][0]=0.;
            (*rotation)[2][1]=0.;
        }
    }
    else
    {
        double normal[3]; // Normal to z axis and direction vector
        normal[0]=-(double)(*direction)[1];
        normal[1]=(*direction)[0];
        normal[2]=0.;
        double normal_norm=sqrt(normal[0]*normal[0]+normal[1]*normal[1]+normal[2]*normal[2]);
        normal[0]/=normal_norm;
        normal[1]/=normal_norm;
        normal[2]/=normal_norm;
        double angle=acos((*direction)[2]);// Rotation angle around the normal so that z matches the direction 
        // Change angle sign if necessary
        //if ((normal X z).direction >0, sin>0, no sign change)
        if(normal[1]*(*direction)[0]-normal[0]*(*direction)[1] <0.) angle=-angle;

        (*rotation)[0][0]=1. + (1.-cos(angle))*(normal[0]*normal[0]-1);
        (*rotation)[0][1]=-normal[2]*sin(angle)+(1.-cos(angle))*normal[0]*normal[1];
        (*rotation)[0][2]=normal[1]*sin(angle)+(1.-cos(angle))*normal[0]*normal[2];
        (*rotation)[1][0]=normal[2]*sin(angle)+(1.-cos(angle))*normal[0]*normal[1];
        (*rotation)[1][1]=1. + (1.-cos(angle))*(normal[1]*normal[1]-1);
        (*rotation)[1][2]=-normal[0]*sin(angle)+(1.-cos(angle))*normal[1]*normal[2];
        (*rotation)[2][0]=-normal[1]*sin(angle)+(1.-cos(angle))*normal[0]*normal[2];
        (*rotation)[2][1]=normal[0]*sin(angle)+(1.-cos(angle))*normal[1]*normal[2];
        (*rotation)[2][2]=1. + (1.-cos(angle))*(normal[2]*normal[2]-1);
    }
}

void MomentsParameters::getMomentsSummed(int flags, QdRScalarSeq_var& density, QdRMatrixSeq_var& velocity, QdRMatrixSeq_var& pressure, QdRMatrixSeq_var& heat, QdRScalarSeq_var& temperature)
{
    density=new QdRScalarSeq(data->sequence_size());
    velocity=new QdRMatrixSeq(data->sequence_size(), Dim(3));
    pressure=new QdRMatrixSeq(data->sequence_size(), Dim(3, 3));
    heat=new QdRMatrixSeq(data->sequence_size(), Dim(3));
    temperature=new QdRScalarSeq(data->sequence_size());
    
    double tot_density, tot_numberflux[3], tot_pressure[3][3], tot_heat[3];
    double v_centre[3], v_centre_norm;
    double val, vol, x, y, z, _v, _phi, _theta;
    bool density_ok, velocity_ok, pressure_ok, heat_ok;
    
    int seq_size=data->sequence_size();
    
    for(int time=0; time<seq_size; time++)
    {
        if(time%100==0)
        {
            cout << floor(time/(seq_size-1.)*100.) << "%" << flush << "\r";
        }
        
        tot_density=0.; density_ok=false;
        tot_numberflux[0]=tot_numberflux[1]=tot_numberflux[2]=0.; velocity_ok=false;
        // Computation of density and velocity
        for(int phi=0; phi<=maxPhi; phi++)
        {
            for(int theta=0; theta<=maxTheta; theta++)
            {
                for(int e=0; e<=maxE; e++)
                {
                    val=getData(time, e, phi, theta);
                    vol=getElementVolume(time, e, phi, theta);
                    
                    if(val>=0. && vol>0.)
                    {
                        tot_density+=val*vol;

                        density_ok=true;
                        
                        if(flags & MOM_VELOCITY)
                        {
                            _v=getV(time, e);
                            
                            if(_v>=0.)
                            {
                                _theta=getTheta(time, theta);
                                _phi=getPhi(time, phi);

                                x=_v*v_sign*sin(_theta*theta2rad)*cos(_phi*phi2rad);
                                y=_v*v_sign*sin(_theta*theta2rad)*sin(_phi*phi2rad);
                                z=_v*v_sign*cos(_theta*theta2rad);

                                convertToGSE(time, x, y, z);

                                tot_numberflux[0]+=val*x*vol;
                                tot_numberflux[1]+=val*y*vol;
                                tot_numberflux[2]+=val*z*vol;
                                velocity_ok=true;
                            }
                        }
                    }
                }
            }
        }
        
        if(density_ok) density[time]=tot_density;
        else density[time]=-1;
        
        if(flags & MOM_VELOCITY)
        {
            if(velocity_ok && density_ok)
            {
                velocity[time][0]=tot_numberflux[0]/(density[time]);
                velocity[time][1]=tot_numberflux[1]/(density[time]);
                velocity[time][2]=tot_numberflux[2]/(density[time]);
            }
            else
            {
                velocity[time]=-1e+31;
            }
        }
        
        if(((flags & MOM_PRESSURE) || (flags & MOM_HEATFLUX)) && (flags & MOM_DENSITY) && (flags & MOM_VELOCITY) && (double)velocity[time][0]!=-1e+31)
        {
            tot_pressure[0][0]=tot_pressure[1][0]=tot_pressure[2][0]=0.;
            tot_pressure[0][1]=tot_pressure[1][1]=tot_pressure[2][1]=0.;
            tot_pressure[0][2]=tot_pressure[1][2]=tot_pressure[2][2]=0.;
            pressure_ok=false;
            
            tot_heat[0]=tot_heat[1]=tot_heat[2]=0.;
            heat_ok=false;

            for(int phi=0; phi<=maxPhi; phi++)
            {
                for(int theta=0; theta<=maxTheta; theta++)
                {
                    for(int e=0; e<=maxE; e++)
                    {
                        val=getData(time, e, phi, theta);
                        vol=getElementVolume(time, e, phi, theta);

                        if(val>=0. && vol>0.)
                        {
                            _v=getV(time, e);
                            if(_v>=0.)
                            {
                                _theta=getTheta(time, theta);
                                _phi=getPhi(time, phi);
                                
                                x=_v*v_sign*sin(_theta*theta2rad)*cos(_phi*phi2rad);
                                y=_v*v_sign*sin(_theta*theta2rad)*sin(_phi*phi2rad);
                                z=_v*v_sign*cos(_theta*theta2rad);
                                
                                convertToGSE(time, x, y, z);
                                
                                v_centre[0]=x-(double)velocity[time][0];
                                v_centre[1]=y-(double)velocity[time][1];
                                v_centre[2]=z-(double)velocity[time][2];
                                
                                v_centre_norm=v_centre[0]*v_centre[0]+v_centre[1]*v_centre[1]+v_centre[2]*v_centre[2];
    
                                if(flags & MOM_PRESSURE)
                                {
                                    for(int i=0; i<3; i++)
                                    {
                                        for(int j=0; j<3; j++)
                                        {
                                            tot_pressure[i][j]+=val*v_centre[i]*v_centre[j]*vol;
                                        }
                                    }
                                    pressure_ok=true;
                                }
                                
                                if(flags & MOM_HEATFLUX)
                                {
                                    for(int i=0; i<3; i++)
                                    {
                                        tot_heat[i]+=val*v_centre_norm*v_centre[i]*vol;
                                    }
                                    heat_ok=true;
                                }
                            }
                        } // end if values ok

                    } // end for e
                } // end for theta
            } // end for phi
        } // end computation of pressure and heat

        if((flags & MOM_PRESSURE) && pressure_ok)
        {
            if(pressure_ok)
            {
                pressure[time][0][0]=tot_pressure[0][0]*mass;
                pressure[time][1][0]=tot_pressure[1][0]*mass;
                pressure[time][2][0]=tot_pressure[2][0]*mass;
                pressure[time][0][1]=tot_pressure[0][1]*mass;
                pressure[time][1][1]=tot_pressure[1][1]*mass;
                pressure[time][2][1]=tot_pressure[2][1]*mass;
                pressure[time][0][2]=tot_pressure[0][2]*mass;
                pressure[time][1][2]=tot_pressure[1][2]*mass;
                pressure[time][2][2]=tot_pressure[2][2]*mass;
                
                temperature[time]=(pressure[time][0][0]+pressure[time][1][1]+pressure[time][2][2])/(3.*density[time]*1.3806503e-23);
            }
            else
            {
                pressure[time]=-1e+31;
                temperature[time]=-1e+31;
            }
        }
        
        if((flags & MOM_HEATFLUX) && heat_ok)
        {
            if(heat_ok)
            {
                heat[time][0]=tot_heat[0]*mass/2.;
                heat[time][1]=tot_heat[1]*mass/2.;
                heat[time][2]=tot_heat[2]*mass/2.;
            }
            else
            {
                heat[time]=-1e+31;
            }
        }

    }
    
    // Metadata
    
    if(flags & MOM_DENSITY)
    {
        for(int i=0; i<seq_size; i++)
        {
            if((double)density[i]!=-1) density[i]=density[i]/1000000.;
        }
        set_timetags(density, get_timetags(data));
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) density);
        QuSetTxtAttr("UNITS", "cm^-3", (QdObject_var) density);
        QuSetTxtAttr("FIELDNAM", "Density - Sum", (QdObject_var) density);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) density);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) density);
        QuSetTxtAttr("SI_CONVERSION", "1.0E6>m^-3", (QdObject_var) density);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) density);
    }
    
    if(flags & MOM_VELOCITY)
    {
        for(int i=0; i<seq_size; i++)
        {
            if((double)velocity[i][0]!=-1e+31)
            {
                velocity[i][0]=velocity[i][0]/1000.;
                velocity[i][1]=velocity[i][1]/1000.;
                velocity[i][2]=velocity[i][2]/1000.;
            }
        }
        set_timetags(velocity, get_timetags(density));
        QuSetTxtAttr("UNITS", "km s^-1", (QdObject_var) velocity);
        QuSetTxtAttr("FIELDNAM", "Bulk Velocity", (QdObject_var) velocity);
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) velocity);
        QuSetTxtAttr("SI_CONVERSION", "1.0E3>m s^-1", (QdObject_var) velocity);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) velocity);
        QuSetTxtAttr("FRAME", "vector>gse_xyz", (QdObject_var) velocity);
    }
    
    if(flags & MOM_PRESSURE)
    {
        for(int i=0; i<seq_size; i++)
        {
            if((double)pressure[i][0][0]!=-1e+31)
            {
                for(int j=0; j<3; j++)
                {
                    for(int k=0; k<3; k++)
                    {
                        pressure[i][j][k]=pressure[i][j][k]*1000000000.;
                    }
                }
            }
        }
        
        set_timetags(pressure, get_timetags(data));
	QdSMatrix_var rep=new QdSMatrix(Dim(3));
	(*rep)[0]="x";
	(*rep)[1]="y";
	(*rep)[2]="z";
	pressure->set_xref("REPRESENTATION_1", new QdSMatrix(*rep));
	pressure->set_xref("REPRESENTATION_2", new QdSMatrix(*rep));
	pressure->set_xref("LABEL_1", new QdSMatrix(*rep));
	pressure->set_xref("LABEL_2", new QdSMatrix(*rep));
        QuSetTxtAttr("UNITS", "nPa", (QdObject_var) pressure);
        QuSetTxtAttr("FIELDNAM", "Density Tensor", (QdObject_var) pressure);
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) pressure);
        QuSetTxtAttr("SI_CONVERSION", "1.0E-9>Pa", (QdObject_var) pressure);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) pressure);
        QuSetTxtAttr("FRAME", "vector>gse_xyz", (QdObject_var) pressure);
	QuSetTxtAttr("TENSOR_ORDER", "2", (QdObject_var) pressure);
    }
    
    if(flags & MOM_HEATFLUX)
    {
        for(int i=0; i<seq_size; i++)
        {
            if((double)heat[i][0]!=-1e+31)
            {
                heat[i][0]=heat[i][0]*1000000.;
                heat[i][1]=heat[i][1]*1000000.;
                heat[i][2]=heat[i][2]*1000000.;
            }
        }
        set_timetags(heat, get_timetags(data));
        QuSetTxtAttr("UNITS", "microW m^-2", (QdObject_var) heat);
        QuSetTxtAttr("FIELDNAM", "Heat Flux", (QdObject_var) heat);
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) heat);
        QuSetTxtAttr("SI_CONVERSION", "1.0E-6>W m^-2", (QdObject_var) heat);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) heat);
        QuSetTxtAttr("FRAME", "vector>gse_xyz", (QdObject_var) heat);
        
        for(int i=0; i<seq_size; i++)
        {
            if((double)temperature[i]!=-1e+31) temperature[i]=temperature[i]/1000000.;
        }
        set_timetags(temperature, get_timetags(data));
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) temperature);
        QuSetTxtAttr("UNITS", "MK", (QdObject_var) temperature);
        QuSetTxtAttr("FIELDNAM", "Temperature", (QdObject_var) temperature);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) temperature);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) temperature);
        QuSetTxtAttr("SI_CONVERSION", "1e+6>K", (QdObject_var) temperature);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) temperature);
    }
    cout << endl << "Done" << endl;
}

void MomentsParameters::getMomentsIntegrated(int flags, QdRScalarSeq_var& density, QdRMatrixSeq_var& velocity, QdRMatrixSeq_var& pressure, QdRMatrixSeq_var& heat, QdRScalarSeq_var& temperature/*, QdRMatrixSeq_var& pad*/)
{
    density=new QdRScalarSeq(data->sequence_size());
    velocity=new QdRMatrixSeq(data->sequence_size(), Dim(3));
    pressure=new QdRMatrixSeq(data->sequence_size(), Dim(3, 3));
    heat=new QdRMatrixSeq(data->sequence_size(), Dim(3));
    temperature=new QdRScalarSeq(data->sequence_size());
    
    double tot_density, tot_numberflux[3], tot_pressure[3][3], tot_heat[3];
    double v_centre[3], v_centre_norm;
    double val, vol, x, y, z;
    
    int seq_size=data->sequence_size();
    
    for(int time=0; time<seq_size; time++)
    {
        if(time%100==0)
        {
            cout << floor(time/(seq_size-1.)*100.) << "%" << flush << "\r";
        }
        
        tot_density=0.;
        tot_numberflux[0]=tot_numberflux[1]=tot_numberflux[2]=0.; 
        
        // Computation of density and velocity
        
        if(time_tetra_cache!=time)
        {
            time_tetra_cache=time;
            buildTetra(time);
        }
        
        // Minimum velocity, to avoid integrating in the (undefined) centre of the space
        double min_v=-1.;
        
        for(int i=0; i<=maxE; i++)
        {
            if(getVMinus(time, i)>0.)
            {
                if(min_v<0.) min_v=getV(time, i);
                else
                {
                    if(min_v>getV(time, i)) min_v=getV(time, i);
                }
            }
        }
        
        min_v*=1.1;
        min_v=min_v*min_v;
        // end minimum velocity
        
        if(!tetra.empty())
        {
            for(list<Element3D>::iterator iter=tetra.begin(); iter!=tetra.end(); iter++)
            {
                int nb=0;
                if(iter->dist1 < min_v) nb++;
                if(iter->dist2 < min_v) nb++;
                if(iter->dist3 < min_v) nb++;
                if(iter->dist4 < min_v) nb++;
                if(nb<4/*iter->distToOrigin < min_v*/)
                {
                    val=(iter->val[0]+iter->val[1]+iter->val[2]+iter->val[3])/4.;
                    vol=iter->volume();
                    tot_density+=val*vol;

                    x=(iter->points[0][0]+iter->points[1][0]+iter->points[2][0]+iter->points[3][0])/4.;

                    y=(iter->points[0][1]+iter->points[1][1]+iter->points[2][1]+iter->points[3][1])/4.;

                    z=(iter->points[0][2]+iter->points[1][2]+iter->points[2][2]+iter->points[3][2])/4.;

                    tot_numberflux[0]+=val*x*vol;
                    tot_numberflux[1]+=val*y*vol;
                    tot_numberflux[2]+=val*z*vol;
                }
            }
            
            density[time]=tot_density;
            
            if(flags & MOM_VELOCITY)
            {
                velocity[time][0]=tot_numberflux[0]/(density[time]);
                velocity[time][1]=tot_numberflux[1]/(density[time]);
                velocity[time][2]=tot_numberflux[2]/(density[time]);
            }
        }
        else
        {
            density[time]=-1;
            velocity[time]=-1e+31;
        }
        
        
        if(((flags & MOM_PRESSURE) || (flags & MOM_HEATFLUX)) && (flags & MOM_DENSITY) && (flags & MOM_VELOCITY) && (double)velocity[time][0]!=-1e+31)
        {
            tot_pressure[0][0]=tot_pressure[1][0]=tot_pressure[2][0]=0.;
            tot_pressure[0][1]=tot_pressure[1][1]=tot_pressure[2][1]=0.;
            tot_pressure[0][2]=tot_pressure[1][2]=tot_pressure[2][2]=0.;
            
            tot_heat[0]=tot_heat[1]=tot_heat[2]=0.;
            
            for(list<Element3D>::iterator iter=tetra.begin(); iter!=tetra.end(); iter++)
            {
                int nb=0;
                if(iter->dist1 < min_v) nb++;
                if(iter->dist2 < min_v) nb++;
                if(iter->dist3 < min_v) nb++;
                if(iter->dist4 < min_v) nb++;
                if(nb<4/*iter->distToOrigin < min_v*/)
                {
                    val=(iter->val[0]+iter->val[1]+iter->val[2]+iter->val[3])/4.;
                    vol=iter->volume();
                    tot_density+=val*vol;

                    x=(iter->points[0][0]+iter->points[1][0]+iter->points[2][0]+iter->points[3][0])/4.;

                    y=(iter->points[0][1]+iter->points[1][1]+iter->points[2][1]+iter->points[3][1])/4.;

                    z=(iter->points[0][2]+iter->points[1][2]+iter->points[2][2]+iter->points[3][2])/4.;
                    
                    v_centre[0]=x-(double)velocity[time][0];
                    v_centre[1]=y-(double)velocity[time][1];
                    v_centre[2]=z-(double)velocity[time][2];
                    
                    v_centre_norm=v_centre[0]*v_centre[0]+v_centre[1]*v_centre[1]+v_centre[2]*v_centre[2];
                    
                    if(flags & MOM_PRESSURE)
                    {
                        for(int i=0; i<3; i++)
                        {
                            for(int j=0; j<3; j++)
                            {
                                tot_pressure[i][j]+=val*v_centre[i]*v_centre[j]*vol;
                            }
                        }
                    }
                            
                    if(flags & MOM_HEATFLUX)
                    {
                        for(int i=0; i<3; i++)
                        {
                            tot_heat[i]+=val*v_centre_norm*v_centre[i]*vol;
                        }
                    }
                }
            }
        } // end computation of pressure and heat
        
        if(flags & MOM_PRESSURE)
        {
            pressure[time][0][0]=tot_pressure[0][0]*mass;
            pressure[time][1][0]=tot_pressure[1][0]*mass;
            pressure[time][2][0]=tot_pressure[2][0]*mass;
            pressure[time][0][1]=tot_pressure[0][1]*mass;
            pressure[time][1][1]=tot_pressure[1][1]*mass;
            pressure[time][2][1]=tot_pressure[2][1]*mass;
            pressure[time][0][2]=tot_pressure[0][2]*mass;
            pressure[time][1][2]=tot_pressure[1][2]*mass;
            pressure[time][2][2]=tot_pressure[2][2]*mass;
            
            temperature[time]=(pressure[time][0][0]+pressure[time][1][1]+pressure[time][2][2])/(3.*density[time]*1.3806503e-23);
        }
        else
        {
            pressure[time]=-1e+31;
            temperature[time]=-1e+31;
        }
        
        if(flags & MOM_HEATFLUX)
        {
            heat[time][0]=tot_heat[0]*mass/2.;
            heat[time][1]=tot_heat[1]*mass/2.;
            heat[time][2]=tot_heat[2]*mass/2.;
        }
        else
        {
            heat[time]=-1e+31;
        }
    }

    // Metadata
    
    if(flags & MOM_DENSITY)
    {
        for(int i=0; i<seq_size; i++)
        {
            if((double)density[i]!=-1) density[i]=density[i]/1000000.;
        }
        set_timetags(density, get_timetags(data));
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) density);
        QuSetTxtAttr("UNITS", "cm^-3", (QdObject_var) density);
        QuSetTxtAttr("FIELDNAM", "Density - Sum", (QdObject_var) density);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) density);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) density);
        QuSetTxtAttr("SI_CONVERSION", "1.0E6>m^-3", (QdObject_var) density);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) density);
    }
    
    if(flags & MOM_VELOCITY)
    {
        for(int i=0; i<seq_size; i++)
        {
            if((double)velocity[i][0]!=-1e+31)
            {
                velocity[i][0]=velocity[i][0]/1000.;
                velocity[i][1]=velocity[i][1]/1000.;
                velocity[i][2]=velocity[i][2]/1000.;
            }
        }
        set_timetags(velocity, get_timetags(density));
        QuSetTxtAttr("UNITS", "km s^-1", (QdObject_var) velocity);
        QuSetTxtAttr("FIELDNAM", "Bulk Velocity", (QdObject_var) velocity);
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) velocity);
        QuSetTxtAttr("SI_CONVERSION", "1.0E3>m s^-1", (QdObject_var) velocity);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) velocity);
        QuSetTxtAttr("FRAME", "vector>gse_xyz", (QdObject_var) velocity);
    }
    
    if(flags & MOM_PRESSURE)
    {
        for(int i=0; i<seq_size; i++)
        {
            if((double)pressure[i][0][0]!=-1e+31)
            {
                for(int j=0; j<3; j++)
                {
                    for(int k=0; k<3; k++)
                    {
                        pressure[i][j][k]=pressure[i][j][k]*1000000000.;
                    }
                }
            }
        }
        
        set_timetags(pressure, get_timetags(data));
	QdSMatrix_var rep=new QdSMatrix(Dim(3));
	(*rep)[0]="x";
	(*rep)[1]="y";
	(*rep)[2]="z";
	pressure->set_xref("REPRESENTATION_1", new QdSMatrix(*rep));
	pressure->set_xref("REPRESENTATION_2", new QdSMatrix(*rep));
	pressure->set_xref("LABEL_1", new QdSMatrix(*rep));
	pressure->set_xref("LABEL_2", new QdSMatrix(*rep));
	QuSetTxtAttr("UNITS", "nPa", (QdObject_var) pressure);
	QuSetTxtAttr("FIELDNAM", "Density Tensor", (QdObject_var) pressure);
	QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) pressure);
	QuSetTxtAttr("SI_CONVERSION", "1.0E-9>Pa", (QdObject_var) pressure);
	QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) pressure);
	QuSetTxtAttr("FRAME", "vector>gse_xyz", (QdObject_var) pressure);
	QuSetTxtAttr("TENSOR_ORDER", "2", (QdObject_var) pressure);
    }
    
    if(flags & MOM_HEATFLUX)
    {
        for(int i=0; i<seq_size; i++)
        {
            if((double)heat[i][0]!=-1e+31)
            {
                heat[i][0]=heat[i][0]*1000000.;
                heat[i][1]=heat[i][1]*1000000.;
                heat[i][2]=heat[i][2]*1000000.;
            }
        }
        set_timetags(heat, get_timetags(data));
        QuSetTxtAttr("UNITS", "microW m^-2", (QdObject_var) heat);
        QuSetTxtAttr("FIELDNAM", "Heat Flux", (QdObject_var) heat);
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) heat);
        QuSetTxtAttr("SI_CONVERSION", "1.0E-6>W m^-2", (QdObject_var) heat);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) heat);
        QuSetTxtAttr("FRAME", "vector>gse_xyz", (QdObject_var) heat);
        
        for(int i=0; i<seq_size; i++)
        {
            if((double)temperature[i]!=-1e+31) temperature[i]=temperature[i]/1000000.;
        }
        set_timetags(temperature, get_timetags(data));
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) temperature);
        QuSetTxtAttr("UNITS", "MK", (QdObject_var) temperature);
        QuSetTxtAttr("FIELDNAM", "Temperature", (QdObject_var) temperature);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) temperature);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) temperature);
        QuSetTxtAttr("SI_CONVERSION", "1e+6>K", (QdObject_var) temperature);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) temperature);
    }
    cout << endl << "Done" << endl;
}

inline double tetra_volume(double xa, double ya, double za, double xb, double yb, double zb, double xc, double yc, double zc, double xd, double yd, double zd)
{
    double ax(xa-xd), ay(ya-yd), az(za-zd),
    bx(xb-xd), by(yb-yd), bz(zb-zd),
    cx(xc-xd), cy(yc-yd), cz(zc-zd);

    return fabs(ax*(by*cz-bz*cy)+ay*(bz*cx-bx*cz)+az*(bx*cy-by*cx))/6.;
}

void MomentsParameters::getMomentsIntegratedO(int flags, QdRScalarSeq_var& density, QdRMatrixSeq_var& velocity, QdRMatrixSeq_var& pressure, QdRMatrixSeq_var& heat, QdRScalarSeq_var& temperature)
{
    density=new QdRScalarSeq(data->sequence_size());
    velocity=new QdRMatrixSeq(data->sequence_size(), Dim(3));
    pressure=new QdRMatrixSeq(data->sequence_size(), Dim(3, 3));
    heat=new QdRMatrixSeq(data->sequence_size(), Dim(3));
    
    double tot_density, tot_numberflux[3], tot_pressure[3][3], tot_heat[3];
    double v1_centre[3], v2_centre[3], v3_centre[3], v4_centre[3], v1_centre_norm, v2_centre_norm, v3_centre_norm, v4_centre_norm;
    double val, val1, val2, val3, val4, vol, vol1, vol2, vol3, vol4, x1, x2, x3, x4, y1, y2, y3, y4, z1, z2, z3, z4, x, y, z;
    
    int seq_size=data->sequence_size();
    
    for(int time=0; time<seq_size; time++)
    {
        tot_density=0.;
        tot_numberflux[0]=tot_numberflux[1]=tot_numberflux[2]=0.; 
        
        // Computation of density and velocity
        
        if(time_tetra_cache!=time)
        {
            time_tetra_cache=time;
            buildTetra(time);
        }
        
        // Minimum velocity, to avoid integrating in the (undefined) centre of the space
        double min_v=-1.;
        
        for(int i=0; i<=maxE; i++)
        {
            if(getVMinus(time, i)>0.)
            {
                if(min_v<0.) min_v=getV(time, i);
                else
                {
                    if(min_v>getV(time, i)) min_v=getV(time, i);
                }
            }
        }
        
        min_v*=1.1;
        min_v=min_v*min_v;
        // end minimum velocity
        
        if(!tetra.empty())
        {
            for(list<Element3D>::iterator iter=tetra.begin(); iter!=tetra.end(); iter++)
            {
                int nb=0;
                if(iter->dist1 < min_v) nb++;
                if(iter->dist2 < min_v) nb++;
                if(iter->dist3 < min_v) nb++;
                if(iter->dist4 < min_v) nb++;
                if(nb<4/*iter->distToOrigin < min_v*/)
                {
                    val=(iter->val[0]+iter->val[1]+iter->val[2]+iter->val[3])/4.;
                    val1=(val+iter->val[1]+iter->val[2]+iter->val[3])/4.;
                    val2=(iter->val[0]+val+iter->val[2]+iter->val[3])/4.;
                    val3=(iter->val[0]+iter->val[1]+val+iter->val[3])/4.;
                    val4=(iter->val[0]+iter->val[1]+iter->val[2]+val)/4.;

                    x=(iter->points[0][0]+iter->points[1][0]+iter->points[2][0]+iter->points[3][0])/4.;

                    y=(iter->points[0][1]+iter->points[1][1]+iter->points[2][1]+iter->points[3][1])/4.;

                    z=(iter->points[0][2]+iter->points[1][2]+iter->points[2][2]+iter->points[3][2])/4.;
                    
                    x1=(x+iter->points[1][0]+iter->points[2][0]+iter->points[3][0])/4.;
                    x2=(iter->points[0][0]+x+iter->points[2][0]+iter->points[3][0])/4.;
                    x3=(iter->points[0][0]+iter->points[1][0]+x+iter->points[3][0])/4.;
                    x4=(iter->points[0][0]+iter->points[1][0]+iter->points[2][0]+x)/4.;
                    
                    y1=(y+iter->points[1][1]+iter->points[2][1]+iter->points[3][1])/4.;
                    y2=(iter->points[0][1]+y+iter->points[2][1]+iter->points[3][1])/4.;
                    y3=(iter->points[0][1]+iter->points[1][1]+y+iter->points[3][1])/4.;
                    y4=(iter->points[0][1]+iter->points[1][1]+iter->points[2][1]+y)/4.;
                    
                    z1=(z+iter->points[1][2]+iter->points[2][2]+iter->points[3][2])/4.;
                    z2=(iter->points[0][2]+z+iter->points[2][2]+iter->points[3][2])/4.;
                    z3=(iter->points[0][2]+iter->points[1][2]+z+iter->points[3][2])/4.;
                    z4=(iter->points[0][2]+iter->points[1][2]+iter->points[2][2]+z)/4.;
                    
                    vol1=tetra_volume(x, y, z,
                                      iter->points[1][0], iter->points[1][1], iter->points[1][2],
                                      iter->points[2][0], iter->points[2][1], iter->points[2][2],
                                      iter->points[3][0], iter->points[3][1], iter->points[3][2]);
                    
                    vol2=tetra_volume(iter->points[0][0], iter->points[0][1], iter->points[0][2],
                                      x, y, z,
                                      iter->points[2][0], iter->points[2][1], iter->points[2][2],
                                      iter->points[3][0], iter->points[3][1], iter->points[3][2]);
                    
                    vol3=tetra_volume(iter->points[0][0], iter->points[0][1], iter->points[0][2],
                                      iter->points[1][0], iter->points[1][1], iter->points[1][2],
                                      x, y, z,
                                      iter->points[3][0], iter->points[3][1], iter->points[3][2]);
                    
                    vol4=tetra_volume(iter->points[0][0], iter->points[0][1], iter->points[0][2],
                                      iter->points[1][0], iter->points[1][1], iter->points[1][2],
                                      iter->points[2][0], iter->points[2][1], iter->points[2][2],
                                      x, y, z);
                    
                    tot_density+=val1*vol1+val2*vol2+val3*vol3+val4*vol4;
                    
                    tot_numberflux[0]+=val1*x1*vol1 + val2*x2*vol2 + val3*x3*vol3 + val4*x4*vol4;
                    tot_numberflux[1]+=val1*y1*vol1 + val2*y2*vol2 + val3*y3*vol3 + val4*y4*vol4;
                    tot_numberflux[2]+=val1*z1*vol1 + val2*z2*vol2 + val3*z3*vol3 + val4*z4*vol4;
                }
            }
            
            density[time]=tot_density;
            
            if(flags & MOM_VELOCITY)
            {
                velocity[time][0]=tot_numberflux[0]/(density[time]);
                velocity[time][1]=tot_numberflux[1]/(density[time]);
                velocity[time][2]=tot_numberflux[2]/(density[time]);
            }
        }
        else
        {
            density[time]=-1;
            velocity[time]=-1e+31;
        }
        
        
        if(((flags & MOM_PRESSURE) || (flags & MOM_HEATFLUX)) && (flags & MOM_DENSITY) && (flags & MOM_VELOCITY) && (double)velocity[time][0]!=-1e+31)
        {
            tot_pressure[0][0]=tot_pressure[1][0]=tot_pressure[2][0]=0.;
            tot_pressure[0][1]=tot_pressure[1][1]=tot_pressure[2][1]=0.;
            tot_pressure[0][2]=tot_pressure[1][2]=tot_pressure[2][2]=0.;
            
            tot_heat[0]=tot_heat[1]=tot_heat[2]=0.;
            
            for(list<Element3D>::iterator iter=tetra.begin(); iter!=tetra.end(); iter++)
            {
                int nb=0;
                if(iter->dist1 < min_v) nb++;
                if(iter->dist2 < min_v) nb++;
                if(iter->dist3 < min_v) nb++;
                if(iter->dist4 < min_v) nb++;
                if(nb<4/*iter->distToOrigin < min_v*/)
                {
                    val=(iter->val[0]+iter->val[1]+iter->val[2]+iter->val[3])/4.;
                    val1=(val+iter->val[1]+iter->val[2]+iter->val[3])/4.;
                    val2=(iter->val[0]+val+iter->val[2]+iter->val[3])/4.;
                    val3=(iter->val[0]+iter->val[1]+val+iter->val[3])/4.;
                    val4=(iter->val[0]+iter->val[1]+iter->val[2]+val)/4.;

                    x=(iter->points[0][0]+iter->points[1][0]+iter->points[2][0]+iter->points[3][0])/4.;

                    y=(iter->points[0][1]+iter->points[1][1]+iter->points[2][1]+iter->points[3][1])/4.;

                    z=(iter->points[0][2]+iter->points[1][2]+iter->points[2][2]+iter->points[3][2])/4.;
                    
                    x1=(x+iter->points[1][0]+iter->points[2][0]+iter->points[3][0])/4.;
                    x2=(iter->points[0][0]+x+iter->points[2][0]+iter->points[3][0])/4.;
                    x3=(iter->points[0][0]+iter->points[1][0]+x+iter->points[3][0])/4.;
                    x4=(iter->points[0][0]+iter->points[1][0]+iter->points[2][0]+x)/4.;
                    
                    y1=(y+iter->points[1][1]+iter->points[2][1]+iter->points[3][1])/4.;
                    y2=(iter->points[0][1]+y+iter->points[2][1]+iter->points[3][1])/4.;
                    y3=(iter->points[0][1]+iter->points[1][1]+y+iter->points[3][1])/4.;
                    y4=(iter->points[0][1]+iter->points[1][1]+iter->points[2][1]+y)/4.;
                    
                    z1=(z+iter->points[1][2]+iter->points[2][2]+iter->points[3][2])/4.;
                    z2=(iter->points[0][2]+z+iter->points[2][2]+iter->points[3][2])/4.;
                    z3=(iter->points[0][2]+iter->points[1][2]+z+iter->points[3][2])/4.;
                    z4=(iter->points[0][2]+iter->points[1][2]+iter->points[2][2]+z)/4.;
                    
                    vol1=tetra_volume(x, y, z,
                                      iter->points[1][0], iter->points[1][1], iter->points[1][2],
                                      iter->points[2][0], iter->points[2][1], iter->points[2][2],
                                      iter->points[3][0], iter->points[3][1], iter->points[3][2]);
                    
                    vol2=tetra_volume(iter->points[0][0], iter->points[0][1], iter->points[0][2],
                                      x, y, z,
                                      iter->points[2][0], iter->points[2][1], iter->points[2][2],
                                      iter->points[3][0], iter->points[3][1], iter->points[3][2]);
                    
                    vol3=tetra_volume(iter->points[0][0], iter->points[0][1], iter->points[0][2],
                                      iter->points[1][0], iter->points[1][1], iter->points[1][2],
                                      x, y, z,
                                      iter->points[3][0], iter->points[3][1], iter->points[3][2]);
                    
                    vol4=tetra_volume(iter->points[0][0], iter->points[0][1], iter->points[0][2],
                                      iter->points[1][0], iter->points[1][1], iter->points[1][2],
                                      iter->points[2][0], iter->points[2][1], iter->points[2][2],
                                      x, y, z);
                    
                    v1_centre[0]=x1-(double)velocity[time][0];
                    v1_centre[1]=y1-(double)velocity[time][1];
                    v1_centre[2]=z1-(double)velocity[time][2];
                    
                    v1_centre_norm=v1_centre[0]*v1_centre[0]+v1_centre[1]*v1_centre[1]+v1_centre[2]*v1_centre[2];
                    
                    v2_centre[0]=x2-(double)velocity[time][0];
                    v2_centre[1]=y2-(double)velocity[time][1];
                    v2_centre[2]=z2-(double)velocity[time][2];
                    
                    v2_centre_norm=v2_centre[0]*v2_centre[0]+v2_centre[1]*v2_centre[1]+v2_centre[2]*v2_centre[2];
                    
                    v3_centre[0]=x3-(double)velocity[time][0];
                    v3_centre[1]=y3-(double)velocity[time][1];
                    v3_centre[2]=z3-(double)velocity[time][2];
                    
                    v3_centre_norm=v3_centre[0]*v3_centre[0]+v3_centre[1]*v3_centre[1]+v3_centre[2]*v3_centre[2];
                    
                    if(flags & MOM_PRESSURE)
                    {
                        for(int i=0; i<3; i++)
                        {
                            for(int j=0; j<3; j++)
                            {
                                tot_pressure[i][j]+=val1*v1_centre[i]*v1_centre[j]*vol1+
                                        val2*v2_centre[i]*v2_centre[j]*vol2+
                                        val3*v3_centre[i]*v3_centre[j]*vol3+
                                        val4*v4_centre[i]*v4_centre[j]*vol4;
                            }
                        }
                    }
                            
                    if(flags & MOM_HEATFLUX)
                    {
                        for(int i=0; i<3; i++)
                        {
                            tot_heat[i]+=val1*v1_centre_norm*v1_centre[i]*vol1+
                                    val2*v2_centre_norm*v2_centre[i]*vol2+
                                    val3*v3_centre_norm*v3_centre[i]*vol3+
                                    val4*v4_centre_norm*v4_centre[i]*vol4;
                        }
                    }
                }
            }
        } // end computation of pressure and heat
        
        if(flags & MOM_PRESSURE)
        {
            pressure[time][0][0]=tot_pressure[0][0]*mass;
            pressure[time][1][0]=tot_pressure[1][0]*mass;
            pressure[time][2][0]=tot_pressure[2][0]*mass;
            pressure[time][0][1]=tot_pressure[0][1]*mass;
            pressure[time][1][1]=tot_pressure[1][1]*mass;
            pressure[time][2][1]=tot_pressure[2][1]*mass;
            pressure[time][0][2]=tot_pressure[0][2]*mass;
            pressure[time][1][2]=tot_pressure[1][2]*mass;
            pressure[time][2][2]=tot_pressure[2][2]*mass;
        }
        else
        {
            pressure[time]=-1e+31;
        }
        
        if(flags & MOM_HEATFLUX)
        {
            heat[time][0]=tot_heat[0]*mass/2.;
            heat[time][1]=tot_heat[1]*mass/2.;
            heat[time][2]=tot_heat[2]*mass/2.;
        }
        else
        {
            heat[time]=-1e+31;
        }
    }

    // Metadata
    
    if(flags & MOM_DENSITY)
    {
        for(int i=0; i<seq_size; i++)
        {
            if((double)density[i]!=-1) density[i]=density[i]/1000000.;
        }
        set_timetags(density, get_timetags(data));
        QuSetTxtAttr("FILLVAL", "-1", (QdObject_var) density);
        QuSetTxtAttr("UNITS", "cm^-3", (QdObject_var) density);
        QuSetTxtAttr("FIELDNAM", "Density - Sum", (QdObject_var) density);
        QuSetTxtAttr("FRAME", "scalar>na", (QdObject_var) density);
        QuSetTxtAttr("PARAMETER_TYPE", "Data", (QdObject_var) density);
        QuSetTxtAttr("SI_CONVERSION", "1.0E6>m^-3", (QdObject_var) density);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) density);
    }
    
    if(flags & MOM_VELOCITY)
    {
        for(int i=0; i<seq_size; i++)
        {
            if((double)velocity[i][0]!=-1e+31)
            {
                velocity[i][0]=velocity[i][0]/1000.;
                velocity[i][1]=velocity[i][1]/1000.;
                velocity[i][2]=velocity[i][2]/1000.;
            }
        }
        set_timetags(velocity, get_timetags(density));
        QuSetTxtAttr("UNITS", "km s^-1", (QdObject_var) velocity);
        QuSetTxtAttr("FIELDNAM", "Bulk Velocity", (QdObject_var) velocity);
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) velocity);
        QuSetTxtAttr("SI_CONVERSION", "1.0E3>m s^-1", (QdObject_var) velocity);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) velocity);
        QuSetTxtAttr("FRAME", "vector>gse_xyz", (QdObject_var) velocity);
    }
    
    if(flags & MOM_PRESSURE)
    {
        for(int i=0; i<seq_size; i++)
        {
            if((double)pressure[i][0][0]!=-1e+31)
            {
                for(int j=0; j<3; j++)
                {
                    for(int k=0; k<3; k++)
                    {
                        pressure[i][j][k]=pressure[i][j][k]*1000000000.;
                    }
                }
            }
        }
        
        set_timetags(pressure, get_timetags(data));
        QuSetTxtAttr("UNITS", "nPa", (QdObject_var) pressure);
        QuSetTxtAttr("FIELDNAM", "Density Tensor", (QdObject_var) pressure);
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) pressure);
        QuSetTxtAttr("SI_CONVERSION", "1.0E-9>Pa", (QdObject_var) pressure);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) pressure);
        QuSetTxtAttr("FRAME", "vector>gse_xyz", (QdObject_var) pressure);
    }
    
    if(flags & MOM_HEATFLUX)
    {
        for(int i=0; i<seq_size; i++)
        {
            if((double)heat[i][0]!=-1e+31)
            {
                heat[i][0]=heat[i][0]*1000000.;
                heat[i][1]=heat[i][1]*1000000.;
                heat[i][2]=heat[i][2]*1000000.;
            }
        }
        set_timetags(heat, get_timetags(data));
        QuSetTxtAttr("UNITS", "microW m^-2", (QdObject_var) heat);
        QuSetTxtAttr("FIELDNAM", "Heat Flux", (QdObject_var) heat);
        QuSetTxtAttr("FILLVAL", "-1e+31", (QdObject_var) heat);
        QuSetTxtAttr("SI_CONVERSION", "1.0E-6>W m^-2", (QdObject_var) heat);
        QuSetTxtAttr("DISPLAYTYPE", "Time_Series", (QdObject_var) heat);
        QuSetTxtAttr("FRAME", "vector>gse_xyz", (QdObject_var) heat);
    }
}
