#include <stdio.h>
#include <math.h>

#ifdef HAVE_IEEEFP_H
 #include <ieeefp.h>
#endif

#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"


using namespace QSAS;

// local prototypes
extern "C" QplugReturnStatus stats (QplugArgList *);


// -------------------------------------------------

QplugReturnStatus stats( QplugArgList *call_list )
{
	 
	QplugAppendTextDisplay("\n\n");

	//  Unpack the input data objects

	QdObject_var dobj = QdObject_var::narrow((*call_list)[0]);
   
	if(dobj.is_nil()  ){
		QplugAppendTextDisplay( "Missing input object in stats filter\n" );
		return QPLUG_FAILURE; 
	}	

	// is sequence a matrix sequence
	QdRMatrixSeq_var ms_in = QdRMatrixSeq_var::narrow(dobj);
	if(!ms_in.is_nil())
	{
		if(ms_in->empty()) {
			QplugAppendTextDisplay( "Empty array series object in stats filter\n" );
			return QPLUG_FAILURE; 
		}
		
		double fillval = -10.e30;
		QdRMatrix_var 	m_fill = new QdRMatrix(ms_in[0].dims(), fillval); // default if matrix not found

		if( ms_in->xref_exists("FILLVAL") ) 
		{
			QdRScalar_var fillObj = QdRScalar_var::narrow(ms_in->get_xref("FILLVAL"));
			if(!fillObj.is_nil()) 
			{
				fillval = *fillObj;
				m_fill = new QdRMatrix(ms_in[0].dims(), fillval);
			}
			else
			{
				QdRMatrix_var fillObj = QdRMatrix_var::narrow(ms_in->get_xref("FILLVAL"));
				if(!fillObj.is_nil()) 
				{
					if(fillObj->dims().size() == ms_in[0].dims().size())
					{
						bool dimsSame = true;
						for(int k=0; k<fillObj->dims().size(); k++)
						{
							if( (fillObj->dims())[k] != (ms_in[0].dims())[k] ) dimsSame = false;
						}
						if(dimsSame) m_fill = fillObj;
					}
				}
			}
		}
		
		// initialise matrix output sequence and reserve max space
		QdRMatrix_var m_mean = new QdRMatrix(ms_in[0].dims(), 0.0);
		QdRMatrix_var m_rms = new QdRMatrix(ms_in[0].dims(), 0.0);
		QdRMatrix_var m_min = new QdRMatrix(ms_in[0].dims(), DBL_MAX);
		QdRMatrix_var m_max = new QdRMatrix(ms_in[0].dims(), -DBL_MAX);
		QdRMatrix_var n_mseq = new QdRMatrix(ms_in[0].dims(), 0.0);

		int nseq = 0;
		QplugAppendTextDisplay("mat sequence");
		QString str_ans = QString("seq size: %1\n").arg(ms_in->sequence_size());
		QplugAppendTextDisplay(str_ans.latin1());
		
		for(register int i=0; i<ms_in->sequence_size(); i++)
		{
			// get underlying data for matrix element
			valarray<double> v_in(ms_in[i].data()[ms_in[i].getgslice()]);
			for(register int j=0; j<v_in.size(); j++)
			{
				if(QuIsBad(v_in[j], (m_fill->data())[j]) ) continue;

				(m_mean->data())[j] += v_in[j];
				if( (m_min->data())[j] > v_in[j]) (m_min->data())[j] = v_in[j];
				if( (m_max->data())[j] < v_in[j]) (m_max->data())[j] = v_in[j];
				(n_mseq->data())[j] += 1;

			}

		}

		for(register int j=0; j<m_mean->data().size(); j++)
		{
			if((n_mseq->data())[j] >= 1) 
			{
				(m_mean->data())[j] /= (n_mseq->data())[j];
			}
		}

		// now we have mean, can get rms
		for(register int i=0; i<ms_in->sequence_size(); i++)
		{
			// get underlying data for matrix element
			valarray<double> v_in(ms_in[i].data()[ms_in[i].getgslice()]);
			for(register int j=0; j<v_in.size(); j++)
			{
				if(QuIsBad(v_in[j], (m_fill->data())[j]) ) continue;

				double diff = (m_mean->data())[j] -  v_in[j];
				diff *= diff;
				(m_rms->data())[j] += diff;
			}

		}
		for(register int j=0; j<m_mean->data().size(); j++)
		{
			if((n_mseq->data())[j] >= 1) 
			{
				(m_rms->data())[j] /= (n_mseq->data())[j];
				(m_rms->data())[j] = sqrt((m_rms->data())[j]);
			}
		}
		
		ms_in->change_xref("Mean", m_mean);
		ms_in->change_xref("Standard Deviation", m_rms);
		ms_in->change_xref("Minumum", m_min);
		ms_in->change_xref("Maximum", m_max);
		ms_in->change_xref("NumberInMean", n_mseq);

		if(ms_in->xref_exists(SI_CONVERSION))
		{
			QdObject_var SI_xref = ms_in->get_xref(SI_CONVERSION);
			m_mean->change_xref(SI_CONVERSION, SI_xref);
			m_rms->change_xref(SI_CONVERSION, SI_xref);
			m_min->change_xref(SI_CONVERSION, SI_xref);
			m_max->change_xref(SI_CONVERSION, SI_xref);
		}
		if(ms_in->xref_exists(DEPEND_1))
		{
			QdObject_var dep1_xref = ms_in->get_xref(DEPEND_1);
			m_mean->change_xref(DEPEND_1, dep1_xref);
			m_rms->change_xref(DEPEND_1, dep1_xref);
			m_min->change_xref(DEPEND_1, dep1_xref);
			m_max->change_xref(DEPEND_1, dep1_xref);
		}
		if(ms_in->xref_exists(DEPEND_2))
		{
			QdObject_var dep2_xref = ms_in->get_xref(DEPEND_2);
			m_mean->change_xref(DEPEND_2, dep2_xref);
			m_rms->change_xref(DEPEND_2, dep2_xref);
			m_min->change_xref(DEPEND_2, dep2_xref);
			m_max->change_xref(DEPEND_2, dep2_xref);
		}
		if(ms_in->xref_exists(DEPEND_3))
		{
			QdObject_var dep3_xref = ms_in->get_xref(DEPEND_3);
			m_mean->change_xref(DEPEND_3, dep3_xref);
			m_rms->change_xref(DEPEND_3, dep3_xref);
			m_min->change_xref(DEPEND_3, dep3_xref);
			m_max->change_xref(DEPEND_3, dep3_xref);
		}
		if(QuIsThreeVector(ms_in))
		{
			if(ms_in->xref_exists(REPRESENTATION_1))
			{
				QdObject_var rep_xref = ms_in->get_xref(REPRESENTATION_1);
				m_mean->change_xref(REPRESENTATION_1, rep_xref);
				m_rms->change_xref(REPRESENTATION_1, rep_xref);
				m_min->change_xref(REPRESENTATION_1, rep_xref);
				m_max->change_xref(REPRESENTATION_1, rep_xref);
			}
			if(ms_in->xref_exists(COORDINATE_SYSTEM))
			{
				QdObject_var coord_xref = ms_in->get_xref(COORDINATE_SYSTEM);
				m_mean->change_xref(COORDINATE_SYSTEM, coord_xref);
				m_rms->change_xref(COORDINATE_SYSTEM, coord_xref);
				m_min->change_xref(COORDINATE_SYSTEM, coord_xref);
				m_max->change_xref(COORDINATE_SYSTEM, coord_xref);
			}
			if(ms_in->xref_exists(TENSOR_ORDER))
			{
				QdObject_var order_xref = ms_in->get_xref(TENSOR_ORDER);
				m_mean->change_xref(TENSOR_ORDER, order_xref);
				m_rms->change_xref(TENSOR_ORDER, order_xref);
				m_min->change_xref(TENSOR_ORDER, order_xref);
				m_max->change_xref(TENSOR_ORDER, order_xref);
			}
			if(ms_in->xref_exists(FRAME))
			{
				QdObject_var frame_xref = ms_in->get_xref(FRAME);
				m_mean->change_xref(FRAME, frame_xref);
				m_rms->change_xref(FRAME, frame_xref);
				m_min->change_xref(FRAME, frame_xref);
				m_max->change_xref(FRAME, frame_xref);
			}
			
		}
		
		str_ans = QString(" Min: \n ----\n");
		QplugAppendTextDisplay(str_ans.latin1());	
		StringRep string_rep;
		str_ans = QString(string_rep(m_min).c_str());
		QplugAppendTextDisplay(str_ans.latin1());	

		str_ans = QString("\n Max: \n ----\n");
		QplugAppendTextDisplay(str_ans.latin1());
		str_ans = QString(string_rep(m_max).c_str());
		QplugAppendTextDisplay(str_ans.latin1());

		str_ans = QString("\n Mean: \n ------\n");
		QplugAppendTextDisplay(str_ans.latin1());
		str_ans = QString(string_rep(m_mean).c_str());
		QplugAppendTextDisplay(str_ans.latin1());

		str_ans = QString("\n Standard Deviation: \n ---------\n");
		QplugAppendTextDisplay(str_ans.latin1());
		str_ans = QString(string_rep(m_rms).c_str());
		QplugAppendTextDisplay(str_ans.latin1());

		QplugAppendTextDisplay ("\nMin, Max, Mean & Standard Deviation added as Attributes to input object \n");
		QplugAppendTextDisplay ("(or cached component if applicable). \n");
		return QPLUG_SUCCESS;

	}

	// scalar sequence case

	QdRScalarSeq_var ss_in = QdRScalarSeq_var::narrow(dobj);
	if(!ss_in.is_nil())
	{
		if(ss_in->empty()) {
			QplugAppendTextDisplay( "Empty scalar series object in stats filter\n" );
			return QPLUG_FAILURE; 
		}
		
		QdRScalar_var s_mean = new QdRScalar(0.0);
		QdRScalar_var s_rms = new QdRScalar(0.0);
		QdRScalar_var s_min = new QdRScalar(DBL_MAX);
		QdRScalar_var s_max = new QdRScalar(-DBL_MAX);

		double fillval = -DBL_MAX;
		if( ss_in->xref_exists("FILLVAL") ) 
		{
			QdRScalar_var fillObj = QdRScalar_var::narrow(ss_in->get_xref("FILLVAL"));
			if(!fillObj.is_nil()) fillval = *fillObj;
		}
		
		int nseq = 0;
		for(register int i=0; i< ss_in->sequence_size(); i++)
		{
			if(QuIsBad(ss_in[i], fillval) ) continue;
			
			*s_mean += ss_in[i];
			if(*s_min > ss_in[i]) *s_min = (double)ss_in[i];
			if(*s_max < ss_in[i]) *s_max = (double)ss_in[i];
			nseq++;
		}
		if(nseq > 0) *s_mean /= nseq;
			
		// now do RMS
		for(register int i=0; i< ss_in->sequence_size(); i++)
		{
			if(QuIsBad(ss_in[i], fillval) ) continue;
			
			double rms = *s_mean - ss_in[i];
			*s_rms += rms * rms;
		}
		if(nseq > 0) 
		{
			double variance = *s_rms / nseq;
			*s_rms = sqrt(variance);
		}
		
		ss_in->change_xref("Mean", s_mean);
		ss_in->change_xref("Standard Deviation", s_rms);
		ss_in->change_xref("Minumum", s_min);
		ss_in->change_xref("Maximum", s_max);
		
		if(ss_in->xref_exists(SI_CONVERSION))
		{
			QdObject_var SI_xref = ss_in->get_xref(SI_CONVERSION);
			s_mean->change_xref(SI_CONVERSION, SI_xref);
			s_rms->change_xref(SI_CONVERSION, SI_xref);
			s_min->change_xref(SI_CONVERSION, SI_xref);
			s_max->change_xref(SI_CONVERSION, SI_xref);
		}
		
		QString str_ans = QString("Min: %1\n").arg(*s_min);
		QplugAppendTextDisplay(str_ans.latin1());			
				
		str_ans = QString("Max: %1\n").arg(*s_max);
		QplugAppendTextDisplay(str_ans.latin1());
		
		str_ans = QString("Mean: %1\n").arg(*s_mean);
		QplugAppendTextDisplay(str_ans.latin1());
		
		str_ans = QString("Standard Deviation: %1\n").arg(*s_rms);
		QplugAppendTextDisplay(str_ans.latin1());
		
		QplugAppendTextDisplay ("\nMin, Max, Mean & Standard Deviation added as Attributes to input object \n");
		QplugAppendTextDisplay ("(or cached component if applicable). \n");
		return QPLUG_SUCCESS;
	}


   // string and other input types not cleaned - return input object

   QplugAppendTextDisplay( " Stats filter does not handle this data type\n" );
   return QPLUG_FAILURE;

}

