#include "QbTimeEditors.h"


void QbEditor::updateForRec(){

	if(record > maxRec) return;

	if (editorType == "QbScalarSeqEditor") ((QbScalarSeqEditor *)this)->updateForRec();
	else if (editorType == "QbMatrixSeqEditor") ((QbMatrixSeqEditor *)this)->updateForRec();
	else if (editorType == "QbTextSeqEditor") ((QbTextSeqEditor *)this)->updateForRec();
	else if (editorType == "QbTtagEditor") ((QbTtagEditor *)this)->updateForRec();
	else if (editorType == "QbEventTableEditor") ((QbEventTableEditor *)this)->updateForRec();

}

QbTimeIntervalEditor::QbTimeIntervalEditor( QdObject_var dobj, QWidget *parent, const char *name )
	:QbEditor(dobj, parent, name)
{
	setTitle("Time Interval");
	
	interval_edit=new QbTimeIntervalWidget(this);
	
	setData(dobj);
	
#if QT_VERSION < 0x040000
// 	this->layout()->add(interval_edit);
#else
	this->layout()->addWidget(interval_edit);
#endif
	connect(interval_edit, SIGNAL(intervalUpdated()), this, SLOT(updateTI()));
}

void QbTimeIntervalEditor::updateTI()
{
	_time_interval_var->set(interval_edit->_start_time_field->time(), interval_edit->_end_time_field->time());
	emit(editorHasInput());
}

void QbTimeIntervalEditor::setData(QdObject_var dobj)
{
	showTimeField(false);
	showValueField(false);

	interval_edit->setTimeInterval(dobj);

	if( dobj.is_nil() )
	{ 
		QDOS::Time end_time;
		end_time.set_today();
		QDOS::Time start_time(end_time-QDOS::Hours(24));
		_time_interval_var=new QdTimeInterval(start_time, end_time);
		return;
	}
	
	QdTimeInterval_var tint=QdTimeInterval_var::narrow(dobj);
   
	if( !tint.is_nil() ){
   
		_time_interval_var = Factory::clone(tint);
		_time_interval_var->copy_xrefs_from(dobj);
		string objectName;
		if(dobj->xref_exists(OBJECT_NAME))
		{
			QdString_var objectName_var=\
					QdString_var::narrow(dobj->get_xref(OBJECT_NAME));
			if(!objectName_var.is_nil())
				objectName="Time Intervals/"+*objectName_var;
		}
	}
}

void QbTimeIntervalEditor::save(Vostream& v_os)
{
	QdTimeInterval_var TIobj=QdTimeInterval_var::narrow(Factory::clone(*_time_interval_var));
	v_os << TIobj;

}
void QbTimeIntervalEditor::restore(Vistream& v_is)
{
	QdTimeInterval_var TIobj;
	v_is >> TIobj;
	this->set_time_interval(*TIobj);
}

void QbTimeIntervalEditor::set_time_interval(const QDOS::TimeInterval& tintv)
{
	QdTimeInterval_var temp=new QdTimeInterval(tintv);
	interval_edit->setTimeInterval(temp);
}

bool QbTimeIntervalEditor::get_time_interval( QDOS::TimeInterval& tintv ) const
{
	tintv=*_time_interval_var;
	return true;
}

QDOS::Time QbTimeIntervalEditor::get_duration()
{
	return interval_edit->getDuration();
}


QbEventTableEditor::QbEventTableEditor( QdObject_var dobj, QWidget *parent, const char *name )
	:QbEditor(dobj, parent, name)
{
	setTitle("Event Table");
	
	interval_edit=new QbTimeIntervalWidget(this);
	
	setData(dobj);
	
#if QT_VERSION >= 0x040000
	this->layout()->addWidget(interval_edit);
#endif
	connect(interval_edit, SIGNAL(intervalUpdated()), this, SLOT(updateCurrentTI()));
	
	editorType="QbEventTableEditor";
}

void QbEventTableEditor::updateCurrentTI()
{
	_event_table_var[record].set(interval_edit->_start_time_field->time(), interval_edit->_end_time_field->time());
	emit(editorHasInput());
}

void QbEventTableEditor::setData(QdObject_var dobj)
{
	showTimeField(false);
	showValueField(false);
	if(dobj.is_nil()) return;

	QdTimeIntervalSeq_var seq=QdTimeIntervalSeq_var::narrow(dobj);
	
	if(!seq.is_nil() )
	{ 
		maxRec=seq->sequence_size()-1;
		record=0;
		
		_event_table_var=Factory::clone(seq);
		_event_table_var->copy_xrefs_from(dobj);
		
		set_event_table(*_event_table_var);
		return;
	}
	else
		QuMsgBar("Data object not recognised", true);
}

void QbEventTableEditor::set_event_table(const QdTimeIntervalSeq& evttbl)
{
	interval_edit->setTimeInterval(evttbl[record]);
	
	//update_time(); //?
}

void QbEventTableEditor::updateForRec()
{
	if(!_event_table_var.is_nil())
	{
		interval_edit->setTimeInterval(_event_table_var[record]);
	}
}

void QbuLineEdit::focusOutEvent(QFocusEvent *e)
{
	const QValidator *v=validator();
	int curpos=cursorPosition();
	QString str=text();
	bool changed=str!=_old_text;
	if(changed)
	{
		if(v)
		{
			if(v->validate(str,curpos)==QValidator::Acceptable)
			{
				if(curpos!=cursorPosition())
					setCursorPosition(curpos);
				if(str!=text())
					setText(str);
			}
			else 
			{
				if(curpos!=cursorPosition())
					setCursorPosition(curpos);
				v->fixup(str);
				if(str!=text())
					setText(str);
			}
		}
		else
		{
			if(curpos!=cursorPosition())
				setCursorPosition(curpos);
			if(str!=text())
				setText(str);
		}
	}
	QLineEdit::focusOutEvent(e);
	if(changed) 
		emit focusOut();
}


