
#include "qslot.h"
#include "qwui.h"
#include "qdutil.h"
#include "TimeUtils.h"

#if QT_VERSION < 0x040000
#include "qgu.h"
#else
#include "qgu_qt4.h"
#endif

 indexSlider::indexSlider(int atNum, QString title,  int maxN, QWidget *parent, char * name)
: QWidget(parent,name)
{

  if(QuSlotDebug())
  {
	printf(" indexSlider::indexSlider \n");
	fflush(stdout);
  }
   
   maxNum = maxN-1;
   num = atNum;
   
   QGridLayout * layout=new QGridLayout(this);
   
   layout->setSpacing(0);
   
   echoLabel=new QLabel(title, this);
#if QT_VERSION < 0x040000
   layout->addMultiCellWidget(echoLabel, 0, 0, 0, 2, Qt::AlignCenter);
#else
   layout->addWidget(echoLabel, 0, 0, 1, 3, Qt::AlignCenter);
#endif

   QToolButton * leftSlide = new QToolButton(Qt::LeftArrow, this, "");
   leftSlide->setFixedSize(20,21);
   layout->addWidget(leftSlide, 1, 0, Qt::AlignLeft);
   numField = new QguWLinput(this,"numField");
   numField->acceptInteger = true;
   numField->setEnabled(true);
   numField->setText(QString("%1").arg(atNum+1));
   numField->setFixedWidth(40);
   numField->setAlignment(Qt::AlignHCenter);
   layout->addWidget(numField, 1, 1);
   
   QToolButton * rightSlide = new QToolButton( Qt::RightArrow, this, "");
   rightSlide->setFixedSize(20,21);
   layout->addWidget(rightSlide, 1, 2, Qt::AlignLeft);
   
   connect( rightSlide, SIGNAL(clicked()), this, SLOT(incRec()) );
   connect( leftSlide,  SIGNAL(clicked()), this, SLOT(decRec()) );
   connect( numField,   SIGNAL(validated_OK()), this, SLOT(adjustRec()) );
   
}

indexSlider::~indexSlider(){
  delete numField;
 // delete sliderBox;
}

void indexSlider::incRec(){
  if(QuSlotDebug())
  {
	printf("indexSlider::incRec %ld\n", this);
	fflush(stdout);
  }

    num += 1;
    numField->setText(tr("%1").arg(num+1));
    adjustRec();  // done explicitly as textChanged() is not emitted if slot does not have focus
}

void indexSlider::decRec(){
  if(QuSlotDebug())
  {
	printf("indexSlider::decRec %ld\n", this);
	fflush(stdout);
  }

    num -= 1;
    if(num < 0) num += 1; // don't go to zero as this is not caught by adjustRec due to empty field option
    numField->setText(tr("%1").arg(num+1));
    adjustRec(); // done explicitly as textChanged() is not emitted if slot does not have focus
}

void indexSlider::adjustRec(){
  if(QuSlotDebug())
  {
   printf("indexSlider::adjustRec %ld\n", this);
   fflush(stdout);
  }

  bool OK;
  num = numField->text().toInt(&OK);

  if(num == 0){ // zero or empty field
    // set red
    numField->setColour("not_ok");
    return;
  }

  num -= 1; // as feedback goes 1 to N while code goes 0 to N-1

  if(OK){
    if(num < 0) num = 0;
    if(num > maxNum) num = maxNum;
    numField->setText(tr("%1").arg(num+1));
    emit indexChanged();
  }
  else
    numField->setText("1");
    emit indexChanged();
}

// class rangeSlider
//-------------------
#if QT_VERSION < 0x040000
 rangeSlider::rangeSlider(int idx,  int MaxN, QdObject_var dep_i, QWidget *parent, char * name)
   : QHGroupBox(parent,name)
{
	if(QuSlotDebug())
	{
		printf("rangeSlider::rangeSlider \n");
		fflush(stdout);
	}
   
	this->setMargin(QSAS_profiles->current()->GUImargins);

	binTextSeq = new QdStringSeq();
	QdObject_var index_label;
	QdObject_var index_units;

	if( !dep_i.is_nil()){

		if(dep_i->xref_exists("LABLAXIS") )
			index_label = dep_i->get_xref("LABLAXIS");
		if(index_label.is_nil() && dep_i->xref_exists("FIELDNAM") )
			index_label = dep_i->get_xref("FIELDNAM");

		if(dep_i->xref_exists("UNITS") )
			index_units = dep_i->get_xref("UNITS");

		// scalar 1-D array

		QdRMatrixSeq_var depS = QdRMatrixSeq_var::narrow(dep_i);
		QdRMatrix_var dep;
		// if time dependent take first matrix, else take matrix itself
		if ( !depS.is_nil() ) dep = new QdRMatrix(depS[0]);
		else dep = QdRMatrix_var::narrow(dep_i);

		if( !dep.is_nil())
		{
			for (int i=0; i<dep->size(); i++){
				QString nextBin = QString("%1").arg((*dep)[i]);
				binTextSeq->push_back( QdString(nextBin.latin1()) );
			}
		}
		else
		{
			QdSMatrix_var lab_i = QdSMatrix_var::narrow(dep_i);

			if( !lab_i.is_nil() )
			{
				for (int i=0; i<lab_i->size(); i++){
					binTextSeq->push_back( (*lab_i)[i] );
				}
			}
		}

   }

   QString title;

   if( !index_label.is_nil() ) {
      QdString_var indexString = QdString_var::narrow(index_label);
	 if( !indexString.is_nil() ) title = QString(indexString->c_str());
	 title += QString(",   ");
   }
   else{
     title = QString("index %1,   ").arg(idx+1);
   }

   if( !index_units.is_nil() ) {
      QdString_var unitsString = QdString_var::narrow(index_units);
	 if( !unitsString.is_nil() ) {
	    title += QString(unitsString->c_str());
	    title += QString(",   ");
	 }
   }

   title += QString(" %1 bins").arg(MaxN );
   this->setTitle(title);

   // Combo box
   actionChoice = new QComboBox( false, this, "actionChoice" );
   actionChoice->insertItem( "elements", 0);
   actionChoice->insertItem( "sum", 1 );
   actionChoice->insertItem( "slice", 2 );
   actionChoice->insertItem( "average", 3 );
   actionChoice->resize( actionChoice->sizeHint() );

   this->addSpace(0);

   fromLabel = new QLabel("   from ", this);
   fromLabel->resize( fromLabel->sizeHint() );

   QString fromBinTxt;
   QString toBinTxt;
   if( binTextSeq->sequence_size() > 0 ){
      fromBinTxt = QString( (binTextSeq[0]).c_str() );
      toBinTxt = QString( (binTextSeq[ binTextSeq->sequence_size()-1]).c_str() );
   }

   fromSlide = new indexSlider(0, fromBinTxt, MaxN, this, "fromSlider" );
   
   toLabel = new QLabel("   to ", this);
   toLabel->resize( toLabel->sizeHint() );
   
   toSlide = new indexSlider(MaxN-1, toBinTxt, MaxN, this, "toSlider" );
   
   connect( fromSlide, SIGNAL(indexChanged()), this, SLOT(updateOnIdx()) );
   connect( toSlide, SIGNAL(indexChanged()), this, SLOT(updateOnIdx()) );
   connect( actionChoice, SIGNAL(activated(int)), this,SLOT(updateRangeWidget(int)) );
   this->setEnabled(true);
}
#else
 rangeSlider::rangeSlider(int idx,  int MaxN, QdObject_var dep_i, QWidget *parent, char * name)
	: QGroupBox(parent,name)
{
	if(QuSlotDebug())
	{
		printf("rangeSlider::rangeSlider \n");
		fflush(stdout);
	}
	
	QHBoxLayout * mainLayout=new QHBoxLayout;
	
	binTextSeq = new QdStringSeq();
	QdObject_var index_label;
	QdObject_var index_units;
	
	if( !dep_i.is_nil())
	{
		if(dep_i->xref_exists("LABLAXIS") )
			index_label = dep_i->get_xref("LABLAXIS");
		if(index_label.is_nil() && dep_i->xref_exists("FIELDNAM") )
			index_label = dep_i->get_xref("FIELDNAM");

		if(dep_i->xref_exists("UNITS") )
			index_units = dep_i->get_xref("UNITS");

		// scalar 1-D array

		QdRMatrixSeq_var depS = QdRMatrixSeq_var::narrow(dep_i);
		QdRMatrix_var dep;
		// if time dependent take first matrix, else take matrix itself
		if ( !depS.is_nil() ) dep = new QdRMatrix(depS[0]);
		else dep = QdRMatrix_var::narrow(dep_i);

		if( !dep.is_nil())
		{
			for (int i=0; i<dep->size(); i++){
				QString nextBin = QString("%1").arg((*dep)[i]);
				binTextSeq->push_back( QdString(nextBin.latin1()) );
			}
		}
		else
		{
			QdSMatrix_var lab_i = QdSMatrix_var::narrow(dep_i);

			if( !lab_i.is_nil() )
			{
				for (int i=0; i<lab_i->size(); i++){
					binTextSeq->push_back( (*lab_i)[i] );
				}
			}
		}
	}
	
	QString title;

	if( !index_label.is_nil() ) {
		QdString_var indexString = QdString_var::narrow(index_label);
		if( !indexString.is_nil() ) title = QString(indexString->c_str());
		title += QString(",   ");
	}
	else{
		title = QString("index %1,   ").arg(idx+1);
	}
	
	if( !index_units.is_nil() ) {
		QdString_var unitsString = QdString_var::narrow(index_units);
		if( !unitsString.is_nil() ) {
			title += QString(unitsString->c_str());
			title += QString(",   ");
		}
	}
	
	title += QString(" %1 bins").arg(MaxN );
	this->setTitle(title);
	
	// Combo box
	actionChoice = new QComboBox( false, this, "actionChoice" );
	actionChoice->insertItem( "elements", 0);
	actionChoice->insertItem( "sum", 1 );
	actionChoice->insertItem( "slice", 2 );
	actionChoice->insertItem( "average", 3 );
	actionChoice->resize( actionChoice->sizeHint() );
	mainLayout->addWidget(actionChoice);

	fromLabel = new QLabel("   from ", this);
	fromLabel->resize( fromLabel->sizeHint() );
	mainLayout->addWidget(fromLabel);

	QString fromBinTxt;
	QString toBinTxt;
	if( binTextSeq->sequence_size() > 0 ){
		fromBinTxt = QString( (binTextSeq[0]).c_str() );
		toBinTxt = QString( (binTextSeq[ binTextSeq->sequence_size()-1]).c_str() );
	}

	fromSlide = new indexSlider(0, fromBinTxt, MaxN, this, "fromSlider" );
   	mainLayout->addWidget(fromSlide);
	
	toLabel = new QLabel("   to ", this);
	toLabel->resize( toLabel->sizeHint() );
	mainLayout->addWidget(toLabel);
   
	toSlide = new indexSlider(MaxN-1, toBinTxt, MaxN, this, "toSlider" );
   	mainLayout->addWidget(toSlide);
	
	connect( fromSlide, SIGNAL(indexChanged()), this, SLOT(updateOnIdx()) );
	connect( toSlide, SIGNAL(indexChanged()), this, SLOT(updateOnIdx()) );
	connect( actionChoice, SIGNAL(activated(int)), this,SLOT(updateRangeWidget(int)) );
	this->setEnabled(true);
	
	setLayout(mainLayout);
}
#endif



rangeSlider::~rangeSlider(){
}

void rangeSlider::updateRangeWidget(int idx)
{
  if(QuSlotDebug())
  {
	printf("rangeSlider::updateRangeWidget %ld\n", this);
	fflush(stdout);
  }
   
  if(idx == 2){
     fromLabel->setText("       at");
     toSlide->hide();
	toLabel->setText("                -------------     ");

  }
  else{
     fromLabel->setText("   from");
     toSlide->show();
	toLabel->setText("   to");
  }

}

void rangeSlider::updateOnIdx()
{
  if(QuSlotDebug())
  {
	printf("rangeSlider::updateOnIdx %ld\n", this);
	fflush(stdout);
  }

   int k;
   bool ok;
   // set label feedback
   if (binTextSeq.is_nil() || binTextSeq->sequence_size() == 0) return;

   // from Label
   k = fromSlide->numField->text().toLong(&ok);
   if(ok) {
     k--; // feedback is from 1 to N
	if (k < 0 || k >= binTextSeq->sequence_size() )
	{
	   fromSlide->setLabel( QString( "out of range" ) );
	}
	else
	{
	   fromSlide->setLabel( QString( (binTextSeq[k]).c_str() ) );
	}
   }
   else  fromSlide->setLabel( QString( "" ) );

   // to Label
   k = toSlide->numField->text().toLong(&ok);
   if(ok) {
     k--; // feedback is from 1 to N
	if (k < 0 || k >= binTextSeq->sequence_size() )
	{
	   toSlide->setLabel( QString( "out of range" ) );
	}
	else
	{
	   toSlide->setLabel( QString( (binTextSeq[k]).c_str() ) );
	}
   }
   else toSlide->setLabel( QString( "" ) );

}

ArrayReduceWindow::ArrayReduceWindow(QString WLcontext, Dim &arrayDim,  QWidget *parent, const char *name, 
                                     bool acceptTSubset, bool acceptTSlice, bool modal )
  : QDialog(parent,name,modal)
{
  if(QuSlotDebug())
  {
	printf("ArrayReduceWindow::ArrayReduceWindow \n");
	fflush(stdout);
  }


  isSequence = FALSE;

#if QT_VERSION < 0x040000
  QWidget::setWFlags(Qt::WDestructiveClose);
#else
  setAttribute(Qt::WA_DeleteOnClose);
#endif

  QVBoxLayout *array_window_top_layout=new QVBoxLayout(this);
//   array_window_top_layout->setSpacing(8);

  QMenuBar * menu_bar = new QMenuBar(this, "menuBar");
  array_window_top_layout->addWidget(menu_bar);

  Q_POPUP_MENU * help = new Q_POPUP_MENU( this );
  menu_bar->insertItem( " Help", help );
  help->insertItem("QSAS &Manual", this, SLOT( launchManual() ), Qt::CTRL+Qt::Key_M );
  help->insertSeparator();
  help->insertItem("...Array Reduction", this, SLOT( launchManArrayReduce() ));
  help->insertSeparator();
  help->insertItem("Array Module Version", this, SLOT( launchVersion() ));
  menu_bar->setFixedSize(menu_bar->sizeHint());

  this->setCaption(WLcontext);

//   QFrame * underline = new QFrame(this, "underline");
//   underline->setFrameStyle(QFrame::HLine | QFrame::Raised);
//   underline->setLineWidth(1);
//   array_window_top_layout->addWidget(underline);

   // set up sliders for each dimension
   QdObject_var dobj = QuGetObjByName(WLcontext);
   for (int i=0; i < arrayDim.rank(); i++){
       QdObject_var dep_i;
	   QString dep_i_name = QString("DEPEND_%1").arg(i+1);
	   QString lab_i_name = QString("LABEL_%1").arg(i+1);
       if( !dobj.is_nil() ){
	     if( dobj->xref_exists( dep_i_name.latin1()) ) dep_i = dobj->get_xref( dep_i_name.latin1() );
		 else if ( dobj->xref_exists( lab_i_name.latin1()) ) dep_i = dobj->get_xref( lab_i_name.latin1() ); // use label_i instead
		 
	     if(dobj->sequence_size() > 1 ) isSequence = TRUE;
	   }
	   // Note that rangeSlider class can cope with dep_i.is_nil()
       rangeSlider *slider_i = new rangeSlider(i,  arrayDim[i], dep_i, this);
	   sliderStack.push_back(slider_i);
	   connect(slider_i->actionChoice, SIGNAL(activated(int)), this, SLOT(updateFeedback()) );
	   array_window_top_layout->addWidget(slider_i);
	   slider_i->setMinimumSize(slider_i->sizeHint());
   }
   
   if( !dobj.is_nil() &&(acceptTSlice || acceptTSubset) ) {
      array_window_top_layout->addSpacing(10);
      slicer = new timeSlider(dobj, acceptTSlice, acceptTSubset, this);
      array_window_top_layout->addWidget(slicer);
   }
   else slicer = 0;
   
   
#if QT_VERSION < 0x040000
   buttons_layout=new QHGroupBox(this, "buttons");

   QPushButton *acceptBtn = new QPushButton( "Accept", buttons_layout, "acceptButton" );
   acceptBtn->setMinimumSize(acceptBtn->sizeHint());
   connect( acceptBtn, SIGNAL(clicked()), this, SLOT(acceptSelection()) );
   acceptBtn->setFocusPolicy(STRONG_FOCUS);
   acceptBtn->setDefault(true);

   QPushButton *setDefaultBtn = new QPushButton( "Reset", buttons_layout, "setDefaultBtn" );
   setDefaultBtn->setMinimumSize(setDefaultBtn->sizeHint());
   connect( setDefaultBtn, SIGNAL(clicked()), this, SLOT(setDefaults()) );

   dismissBtn = new QPushButton( "Cancel", buttons_layout, "dismissBtn" );
   dismissBtn->setMinimumSize(dismissBtn->sizeHint());
   connect( dismissBtn, SIGNAL(clicked()), this, SLOT(dismiss()) );

   QString feedbackMsg = QString("Input is %1-D array").arg(arrayDim.rank());
   buttons_layout->setAlignment(Qt::AlignHCenter);
   buttons_layout->setTitle(feedbackMsg);
   array_window_top_layout->addWidget(buttons_layout);
   buttons_layout->setMinimumSize(buttons_layout->sizeHint());
#else
   buttons_groupbox=new QGroupBox(this, "buttons");
   QHBoxLayout * buttons_layout=new QHBoxLayout;

   QPushButton *acceptBtn = new QPushButton( "Accept");
   acceptBtn->setMinimumSize(acceptBtn->sizeHint());
   connect( acceptBtn, SIGNAL(clicked()), this, SLOT(acceptSelection()) );
   acceptBtn->setFocusPolicy(STRONG_FOCUS);
   acceptBtn->setDefault(true);
   buttons_layout->addWidget(acceptBtn);

   QPushButton *setDefaultBtn = new QPushButton( "Reset");
   setDefaultBtn->setMinimumSize(setDefaultBtn->sizeHint());
   connect( setDefaultBtn, SIGNAL(clicked()), this, SLOT(setDefaults()) );
   buttons_layout->addWidget(setDefaultBtn);

   dismissBtn = new QPushButton( "Cancel", buttons_groupbox, "dismissBtn" );
   dismissBtn->setMinimumSize(dismissBtn->sizeHint());
   connect( dismissBtn, SIGNAL(clicked()), this, SLOT(dismiss()) );
   buttons_layout->addWidget(dismissBtn);

   QString feedbackMsg = QString("Input is %1-D array").arg(arrayDim.rank());
   buttons_groupbox->setAlignment(Qt::AlignHCenter);
   buttons_groupbox->setTitle(feedbackMsg);
   array_window_top_layout->addWidget(buttons_groupbox);
   
   buttons_groupbox->setLayout(buttons_layout);
#endif

   this->setMinimumSize(this->sizeHint());

   arrHelp_window = new QswHelpWindow(this, "arrHelpWindow");
}


void ArrayReduceWindow::updateFeedback(){

  if(QuSlotDebug())
	{
		printf("ArrayReduceWindow::updateFeedback %ld\n", this);
		fflush(stdout);
	}

   int nDims=0;
   for (int i=0; i<sliderStack.size(); i++){
      if( sliderStack[i]->actionChoice->currentItem() == 0) nDims++;
   }
   QString newTitle;
   
   if(isSequence){
      if( nDims > 0)
         newTitle = QString("  Output is %1-D array sequence").arg(nDims) ;
	 else
         newTitle =  QString("  Output is a scalar sequence") ;
   }
   else{
      if( nDims > 0)
         newTitle =  QString("  Output is %1-D array").arg(nDims) ;
	 else
         newTitle =  QString("  Output is a scalar") ;
   }
   
   if (slicer != 0){
     if(slicer->sliceChoice->currentItem() == slicer->Isubset){
	   int nrec = slicer->toSlider->record - slicer->fromSlider->record + 1;
	    newTitle += QString(", subset of %1 records  ").arg(nrec);
	 }
     if(slicer->sliceChoice->currentItem() == slicer->Iaverage){
	   int nrec = slicer->toSlider->record - slicer->fromSlider->record + 1;
	    newTitle += QString(", averaged over %1 records to single value  ").arg(nrec);
	 }
	 if(slicer->sliceChoice->currentItem() == slicer->Islice){
	    newTitle += QString(", value at record %1  ").arg(slicer->atSlider->record);
	 }
   }
   else  newTitle += "  ";
   
#if QT_VERSION < 0x040000
   buttons_layout->setTitle( newTitle);
#else
   buttons_groupbox->setTitle( newTitle);
#endif

}

void ArrayReduceWindow::dismiss(){
  if(QuSlotDebug())
  {
	printf("ArrayReduceWindow::dismiss %ld\n", this);
	fflush(stdout);
  }

  close();
}

void ArrayReduceWindow::setDefaults(){
  if(QuSlotDebug())
  {
	printf("ArrayReduceWindow::setDefaults %ld\n", this);
	fflush(stdout);
  }


     // set up sliders for each dimension
   for (int i=0; i<sliderStack.size(); i++){
      sliderStack[i]->actionChoice->setCurrentItem(0);
      sliderStack[i]->updateRangeWidget(0);
      sliderStack[i]->fromSlide->numField->setText("-1");      // force reset and cleanup
	  sliderStack[i]->toSlide->numField->setText(QString("%1").arg(sliderStack[i]->toSlide->maxNum+1) );
	  sliderStack[i]->fromSlide->adjustRec();
	  sliderStack[i]->toSlide->adjustRec();
   }
   
   if (slicer != 0){
         slicer->sliceChoice->setCurrentItem(slicer->Isubset); 
         slicer->fromSlider->adjustRec(0);
         slicer->toSlider->adjustRec(slicer->toSlider->maxRecord);
		 
         slicer->fromSlider->show();
	  	 slicer->toSlider->show();
         slicer->atSlider->hide();
   }

}

void ArrayReduceWindow::setToLastChoice(){

  if(QuSlotDebug())
  {
	printf("ArrayReduceWindow::setToLastChoice %ld\n", this);
	fflush(stdout);
  }

   dimTripleStack dim_stack = parentSlot->parseArrayReduce();
   dimTriple sliceTriple = parentSlot->parseTimeSlice();

   if(dim_stack.size() == sliderStack.size()) {
   
      // set up sliders for each dimension
      for (int i=0; i<dim_stack.size(); i++){
         sliderStack[i]->actionChoice->setCurrentItem(dim_stack[i][0]);
         sliderStack[i]->fromSlide->numField->setText(QString("%1").arg(dim_stack[i][1]) );      // force reset and cleanup
	     sliderStack[i]->fromSlide->adjustRec();
	     if(dim_stack[i][0] != 2)  {
	       sliderStack[i]->toSlide->numField->setText(QString("%1").arg(dim_stack[i][2]) );
	       sliderStack[i]->toSlide->adjustRec();
	     }
	     sliderStack[i]->updateRangeWidget(dim_stack[i][0]);
      }
   }

   if(sliceTriple.size() == 3 && slicer != 0){

      if(sliceTriple[0] == 0 || sliceTriple[0] == 3) {
	     slicer->sliceChoice->setCurrentItem(slicer->Isubset);
         slicer->fromSlider->recBar->setValue(sliceTriple[1]);
         slicer->toSlider->recBar->setValue(sliceTriple[2]);
		// slicer->fromSlider->adjustRec(sliceTriple[1]);
		// slicer->toSlider->adjustRec(sliceTriple[2]);
	  }
	  else if(sliceTriple[0] == 1) {
	     slicer->sliceChoice->setCurrentItem(slicer->Iaverage);
		 slicer->fromSlider->recBar->setValue(sliceTriple[1]);
		 slicer->toSlider->recBar->setValue(sliceTriple[2]);
	  }
	  else if(sliceTriple[0] == 2) {
	     slicer->sliceChoice->setCurrentItem(slicer->Islice);
		 slicer->atSlider->recBar->setValue(sliceTriple[1]);
		 slicer->updateTimeSlider(slicer->Islice);
	  }
   }
}

void ArrayReduceWindow::acceptSelection(){
  
  if(QuSlotDebug())
  {
	printf("ArrayReduceWindow::acceptSelection %ld\n", this);
	fflush(stdout);
  }

  dismissBtn->setEnabled(false); // prevent dismiss deleting window before signals complete
  
  QString newText = parentSlot->textObject();
  
  bool sliceTaken = false;
  
  for (int i=0; i<sliderStack.size(); i++){
       if( sliderStack[i]->actionChoice->currentItem() != 0) {
	      sliceTaken=true;
		  break;
	   }
	   if( sliderStack[i]->fromSlide->num != 0){
	      sliceTaken=true;
		  break;
	   }
	   if( sliderStack[i]->toSlide->num != sliderStack[i]->toSlide->maxNum){
	      sliceTaken=true;
		  break;
	   }
  }
  
  QString cmpText="";
  if( sliceTaken ){
    cmpText = " [ ";
    for (int i=0; i<sliderStack.size(); i++){
     
       int op = sliderStack[i]->actionChoice->currentItem();
       QString start = sliderStack[i]->fromSlide->numField->text();      
  	   if(op == 0){
  		  QString end = sliderStack[i]->toSlide->numField->text();
	 	  cmpText += "(";
	 	  cmpText += start;
	 	  cmpText += ",";
	 	  cmpText += end;
	 	  cmpText += ") ";
   	   }
	   else if (op == 1){
	 	  QString end = sliderStack[i]->toSlide->numField->text();
	 	  cmpText += "Sum(";
	 	  cmpText += start;
	 	  cmpText += ",";
	 	  cmpText += end;
	 	  cmpText += ") ";
	   }
	   else if (op == 3){
	 	  QString end = sliderStack[i]->toSlide->numField->text();
	 	  cmpText += "Ave(";
	 	  cmpText += start;
	 	  cmpText += ",";
	 	  cmpText += end;
	 	  cmpText += ") ";
	   }
	   else
	   {
	 	  cmpText += "(";
	 	  cmpText += start;
	 	  cmpText += ") ";
	   }
    }
    cmpText += "]";
  }
  
  if(slicer != 0 && !(slicer->isDefault()) && slicer->sliceChoice->currentItem() != slicer->Inothing){

     cmpText += " < ";
	 
     // Time slice/subset available
	 if(slicer->sliceChoice->currentItem() == slicer->Isubset){
		QString start = slicer->fromSlider->recField->text();      
		QString end   = slicer->toSlider->recField->text();
	 	cmpText += start;
	 	cmpText += ",";
	 	cmpText += end;
	 }
	 else if(slicer->sliceChoice->currentItem() == slicer->Iaverage){
		QString start = slicer->fromSlider->recField->text();      
		QString end   = slicer->toSlider->recField->text();
	 	cmpText += "Ave ";
	 	cmpText += start;
	 	cmpText += ",";
	 	cmpText += end;
	 }
	 else if(slicer->sliceChoice->currentItem() == slicer->Islice){
		QString atT = slicer->atSlider->recField->text();      
	 	cmpText += atT;
	 }
	 
     cmpText += " >";
  }
  
  newText += cmpText;
  parentSlot->setText(newText);
  parentSlot->reValidate();
  parentSlot->emitCmpSelected(cmpText.latin1(), "ARRAY");
  
  close();
}

void ArrayReduceWindow::launchManual(){ 

 if(QuSlotDebug())
 {
	printf("ArrayReduceWindow::launchManual %ld\n", this);
	fflush(stdout);
 }
 
   arrHelp_window->manEntry("index.html");
}

void ArrayReduceWindow::launchManArrayReduce(){ 

 if(QuSlotDebug())
 {
	printf("ArrayReduceWindow::launchManArrayReducer %ld\n", this);
	fflush(stdout);
 }
 
  
   arrHelp_window->manEntry("QSAS_array_reduction.html");
}


void ArrayReduceWindow::launchVersion(){

 if(QuSlotDebug())
 {
	printf("ArrayReduceWindow::launchVersion %ld\n", this);
	fflush(stdout);
 }
 
  
  QString text;
  text  = QString("Version Numbers...\n"); 
  text += QString(QSAS_VERSION_ID);
  text += "\n Array Reduction  ";
  text += QString(QSLOT_VERSION);
  QMessageBox::information( this, "QSAS information", text.latin1(), "OK" );

}

selectMetadataWindow::selectMetadataWindow(QdObject_var dobj,  QWidget *parent, const char *name )
  : QMainWindow(parent,name)
#if QT_VERSION < 0x040000
{
  
  if(QuSlotDebug())
  {
	printf("selectMetadataWindow::selectMetadataWindow \n");
	fflush(stdout);
  }

  QSizePolicy fixed(QSizePolicy::Fixed,QSizePolicy::Fixed);

  QWidget::setWFlags(Qt::WDestructiveClose);

   menuBar()->insertSeparator();
   QPopupMenu * help = new QPopupMenu( this );
   menuBar()->insertItem( " Help", help );
   help->insertItem("QSAS &Manual", this, SLOT( launchManual() ), Qt::CTRL+Qt::Key_M );
   help->insertSeparator();


   QVGroupBox *metaVB = new QVGroupBox( this, "metaVB" );
   setCentralWidget( metaVB );

   xrefList = new Q_LIST_VIEW(metaVB, "Attributes");
   xrefList->setAcceptDrops(false);
   xrefList->addColumn("            Attributes           ");
   xrefList->setMinimumHeight(150);
   xrefList->setMinimumWidth(160);
   xrefList->clear();

   // set Xrefs
   vector<string> xrefs;
   dobj->list_xref_names(xrefs);
   for(int i=0; i<xrefs.size(); i++){
      QdObject_var xref_obj=dobj->get_xref(xrefs[i]);
      Q_LIST_VIEW_ITEM * newitem = new Q_LIST_VIEW_ITEM( xrefList, QString(xrefs[i].c_str()));
   }

  QHGroupBox *buttons_layout=new QHGroupBox(metaVB, "buttons");

  QPushButton *selectBtn = new QPushButton( "Select", buttons_layout, "selectBtn" );
  selectBtn->setMinimumSize(selectBtn->sizeHint());
  connect( selectBtn, SIGNAL(clicked()), this, SLOT(selectAndClose()) );
  selectBtn->setFocusPolicy(STRONG_FOCUS);

  dismissBtn = new QPushButton( "Dismiss", buttons_layout, "dismissBtn" );
  dismissBtn->setMinimumSize(dismissBtn->sizeHint());
  connect( dismissBtn, SIGNAL(clicked()), this, SLOT(dismiss()) );

  this->setMinimumSize(this->sizeHint());

   metaHelp_window = new QswHelpWindow(this, "metaHelpWindow");
}
#else
{
	if(QuSlotDebug())
	{
		printf("selectMetadataWindow::selectMetadataWindow \n");
		fflush(stdout);
	}

	QSizePolicy fixed(QSizePolicy::Fixed,QSizePolicy::Fixed);

	setAttribute(Qt::WA_DeleteOnClose);

	menuBar()->insertSeparator();
	QMenu * help = new QMenu( this );
	menuBar()->insertItem( " Help", help );
	help->insertItem("QSAS &Manual", this, SLOT( launchManual() ), Qt::CTRL+Qt::Key_M );
	help->insertSeparator();


	//  Main widget
	QGroupBox *metaVB = new QGroupBox( this, "metaVB" );
	QVBoxLayout *metaVBLayout=new QVBoxLayout;
	
	// XRef list
	
	xrefList = new Q_LIST_VIEW(metaVB, "Attributes");
	xrefList->setAcceptDrops(false);
	xrefList->addColumn("            Attributes           ");
	xrefList->setMinimumHeight(150);
	xrefList->setMinimumWidth(160);
	xrefList->clear();
	metaVBLayout->addWidget(xrefList);

	vector<string> xrefs;
	dobj->list_xref_names(xrefs);
	for(int i=0; i<xrefs.size(); i++){
		QdObject_var xref_obj=dobj->get_xref(xrefs[i]);
		Q_LIST_VIEW_ITEM * newitem = new Q_LIST_VIEW_ITEM( xrefList, QString(xrefs[i].c_str()));
	}
	
	// Buttons
	
	QHBoxLayout * buttons_layout=new QHBoxLayout;
	
	QPushButton *selectBtn = new QPushButton( "Select");
	selectBtn->setMinimumSize(selectBtn->sizeHint());
	connect( selectBtn, SIGNAL(clicked()), this, SLOT(selectAndClose()) );
	selectBtn->setFocusPolicy(STRONG_FOCUS);
	buttons_layout->addWidget(selectBtn);

	dismissBtn = new QPushButton( "Dismiss");
	dismissBtn->setMinimumSize(dismissBtn->sizeHint());
	connect( dismissBtn, SIGNAL(clicked()), this, SLOT(dismiss()) );
	buttons_layout->addWidget(dismissBtn);
	
	metaVBLayout->addLayout(buttons_layout);
	
	//Done
	
	metaVB->setLayout(metaVBLayout);
	setCentralWidget(metaVB);
	
	this->setMinimumSize(this->sizeHint());

	metaHelp_window = new QswHelpWindow(this, "metaHelpWindow");
}
#endif

selectMetadataWindow::~selectMetadataWindow(){
 if(QuSlotDebug())
 {
	printf("selectMetadataWindow::~selectMetadataWindow %ld\n", this);
	fflush(stdout);
 }

 // delete arrHelp_window;
}

void selectMetadataWindow::dismiss(){
  
  if(QuSlotDebug())
  {
	printf("selectMetadataWindow::dismiss \n");
	fflush(stdout);
  }

  close();
}

void selectMetadataWindow::selectAndClose(){
  
  if(QuSlotDebug())
  {
	printf("selectMetadataWindow::selectAndClose \n");
	fflush(stdout);
  }
  dismissBtn->setEnabled(false); // prevent dismiss deleting window before signals complete
  
  Q_LIST_VIEW_ITEM *item = this->xrefList->selectedItem();
  if(item) {
     QString xrefStr=" [=> "+item->text(0)+" ]";
     QString txt = baseObjectText(parentSlot->text());
	txt += xrefStr;

     parentSlot->setText(txt);
     parentSlot->reValidate();
     parentSlot->lineBox->setFocus();
     if(parentSlot->canEmitCmpSel)  parentSlot->emitCmpSelected(xrefStr.latin1(), "XREF");
  }

  close();
}

void selectMetadataWindow::launchManual(){

 if(QuSlotDebug())
 {
	printf("selectMetadataWindow::launchManual %ld\n", this);
	fflush(stdout);
  }

   metaHelp_window->manEntry("index.html");
}

// Class RecordSlider

// used in QDI and QGU (Array)

#if QT_VERSION < 0x040000
RecordSlider::RecordSlider(const QString & title, QWidget * parent, const char * name )
  : QHGroupBox(title, parent, name)
{
  
  if(QuSlotDebug())
  {
	printf("RecordSlider::RecordSlider \n");
	fflush(stdout);
  }

   titleBase = QString(title.latin1()); // deep copy
   
   this->record = 0; // init to be safe, should be reset by parent
   this->maxRecord = 0; // init to be safe, should be reset by parent
   
   QToolButton * leftEnd = new QToolButton( this, "leftEnd" );
   leftEnd->setText("|");
   leftEnd->setFixedSize(15,21);
   QToolButton * leftSlide = new QToolButton( Qt::LeftArrow, this, "");
   leftSlide->setFixedSize(20,21);

   
   recBar = new QSlider(Qt::Horizontal, this, "recBar");
  
   QToolButton * rightSlide = new QToolButton( Qt::RightArrow, this, "");
   rightSlide->setFixedSize(20,21);
   QToolButton * rightEnd = new QToolButton( this, "rightEnd" );
   rightEnd->setText("|");
   rightEnd->setFixedSize(15,21);
   connect( rightSlide, SIGNAL(clicked()), this, SLOT(incRec()) );
   connect( leftSlide,  SIGNAL(clicked()), this, SLOT(decRec()) );
   connect( rightEnd, SIGNAL(clicked()), this, SLOT(lastRec()) );
   connect( leftEnd,  SIGNAL(clicked()), this, SLOT(firstRec()) );
   connect( recBar,   SIGNAL(valueChanged(int)), this, SLOT(adjustRec(int)) );
   
   recField = new QguWLinput(this,"recField");
   recField->acceptInteger = true;
   recField->setEnabled(true);
   recField->setFixedWidth(60);
   recField->setAlignment(Qt::AlignHCenter);
   recField->setText("0");
   connect( recField,   SIGNAL(textChanged(const QString &)), this, SLOT(setFromTyping()) );

   recBar->setValue(0);

}
#else
RecordSlider::RecordSlider(const QString & title, QWidget * parent, const char * name )
	: QGroupBox(title, parent, name)
{
  
	if(QuSlotDebug())
	{
		printf("RecordSlider::RecordSlider \n");
		fflush(stdout);
	}

	QHBoxLayout * mainLayout=new QHBoxLayout;
	
	titleBase = QString(title.latin1()); // deep copy
   
	this->record = 0; // init to be safe, should be reset by parent
	this->maxRecord = 0; // init to be safe, should be reset by parent
   
	QToolButton * leftEnd = new QToolButton( this, "leftEnd" );
	leftEnd->setText("|");
	leftEnd->setFixedSize(15,21);
	mainLayout->addWidget(leftEnd);
	
	QToolButton * leftSlide = new QToolButton( Qt::LeftArrow, this, "");
	leftSlide->setFixedSize(20,21);
	mainLayout->addWidget(leftSlide);
   
	recBar = new QSlider(Qt::Horizontal, this, "recBar");
  	mainLayout->addWidget(recBar);
	
	QToolButton * rightSlide = new QToolButton( Qt::RightArrow, this, "");
	rightSlide->setFixedSize(20,21);
	mainLayout->addWidget(rightSlide);
	
	QToolButton * rightEnd = new QToolButton( this, "rightEnd" );
	rightEnd->setText("|");
	rightEnd->setFixedSize(15,21);
	mainLayout->addWidget(rightEnd);
	
	connect( rightSlide, SIGNAL(clicked()), this, SLOT(incRec()) );
	connect( leftSlide,  SIGNAL(clicked()), this, SLOT(decRec()) );
	connect( rightEnd, SIGNAL(clicked()), this, SLOT(lastRec()) );
	connect( leftEnd,  SIGNAL(clicked()), this, SLOT(firstRec()) );
	connect( recBar,   SIGNAL(valueChanged(int)), this, SLOT(adjustRec(int)) );
   
	recField = new QguWLinput(this,"recField");
	recField->acceptInteger = true;
	recField->setEnabled(true);
	recField->setFixedWidth(60);
	recField->setAlignment(Qt::AlignHCenter);
	recField->setText("0");
	connect( recField,   SIGNAL(textChanged(const QString &)), this, SLOT(setFromTyping()) );
	mainLayout->addWidget(recField);

	recBar->setValue(0);
	
	setLayout(mainLayout);

}
#endif

void RecordSlider::updateTitle(){
  
  if(QuSlotDebug())
  {
	printf("RecordSlider::updateTitle \n");
	fflush(stdout);
  }

   if( !ttags.is_nil() && record < ttags->sequence_size()){
      StringRep string_rep;
      this->setTitle(titleBase + QString(" time %1").arg(ttags[record].iso().c_str()) );
   }
   else this->setTitle(titleBase + QString(" record %1").arg(record) );
}

void RecordSlider::setFromTyping(){

  if(QuSlotDebug())
  {
	printf("recordSlider::setFromTyping %ld\n", this);
	fflush(stdout);
  }
   
  bool OK;
  int newValue = recField->text().toInt(&OK);
  if (!OK) return;

  if(newValue > maxRecord){
    newValue = maxRecord;
	recField->setText(tr("%1").arg(newValue));
  }
  else if(newValue < 0) {
    newValue = 0;
	recField->setText("0");
  }
  
  recBar->setValue(newValue);
  this->updateTitle();
}

void RecordSlider::incRec(){

  if(QuSlotDebug())
  {
	printf("recordSlider::incRec %ld\n", this);
	fflush(stdout);
  }

  if(record < maxRecord){
    record +=1;
  }
  else{
    record = maxRecord;
  }
  recBar->setValue(record);
  this->updateTitle();
}

void RecordSlider::decRec(){

 if(QuSlotDebug())
 {
	printf("recordSlider::decRec %ld\n", this);
	fflush(stdout);
 }
 
  if(record > 0){
    record -= 1;
  }
  else{
    record = 0;
  }
  recBar->setValue(record);
  this->updateTitle();
}

void RecordSlider::lastRec(){

 if(QuSlotDebug())
 {
	printf("recordSlider::lastRec %ld\n", this);
	fflush(stdout);
 }
 
  record = maxRecord;
  recBar->setValue(record);
  this->updateTitle();
}

void RecordSlider::firstRec(){

 if(QuSlotDebug())
 {
	printf("recordSlider::firstRec %ld\n", this);
	fflush(stdout);
 }
 
  record = 0;
  recBar->setValue(record);
  this->updateTitle();
}

void RecordSlider::adjustRec(int rec){

 if(QuSlotDebug())
 {
	printf("recordSlider::adjustRec %ld %d\n", this, rec);
	fflush(stdout);
 }
 
  record = rec;
  recField->setText(tr("%1").arg(rec));
  recBar->setValue(rec);
  this->updateTitle();
  
  emit recordSliderChanged();
}

// Class TimeIntervalSlider

// used in QDI and QGU (Array)

#if QT_VERSION < 0x040000
TimeIntervalSlider::TimeIntervalSlider(const QString & title, QWidget * parent, const char * name )
    : QHGroupBox(title, parent, name)
{
  
    if(QuSlotDebug())
    {
		printf("TimeIntervalSlider::TimeIntervalSlider \n");
		fflush(stdout);
	}

    obj=0;
    
    titleBase = QString(title.latin1()); // deep copy
   
    this->record = 0; // init to be safe, should be reset by parent
    this->maxRecord = 0; // init to be safe, should be reset by parent
   
    QToolButton * leftEnd = new QToolButton( this, "leftEnd" );
    leftEnd->setText("|");
    leftEnd->setFixedSize(15,21);
    QToolButton * leftSlide = new QToolButton( Qt::LeftArrow, this, "");
    leftSlide->setFixedSize(20,21);

   
    recBar = new QSlider(Qt::Horizontal, this, "recBar");
  
    QToolButton * rightSlide = new QToolButton( Qt::RightArrow, this, "");
    rightSlide->setFixedSize(20,21);
    QToolButton * rightEnd = new QToolButton( this, "rightEnd" );
    rightEnd->setText("|");
    rightEnd->setFixedSize(15,21);
    connect( rightSlide, SIGNAL(clicked()), this, SLOT(incRec()) );
    connect( leftSlide,  SIGNAL(clicked()), this, SLOT(decRec()) );
    connect( rightEnd, SIGNAL(clicked()), this, SLOT(lastRec()) );
    connect( leftEnd,  SIGNAL(clicked()), this, SLOT(firstRec()) );
    connect( recBar,   SIGNAL(valueChanged(int)), this, SLOT(adjustRec(int)) );
   
    recField = new QguWLinput(this,"recField");
    recField->acceptInteger = true;
    recField->setEnabled(true);
    recField->setFixedWidth(60);
    recField->setAlignment(Qt::AlignHCenter);
    recField->setText("0");
    connect( recField,   SIGNAL(textChanged(const QString &)), this, SLOT(setFromTyping()) );

    recBar->setValue(0);
    
    setBaseSize(200, 30);

}
#else
TimeIntervalSlider::TimeIntervalSlider(const QString & title, QWidget * parent, const char * name )
	: QGroupBox(title, parent, name)
{
  
	if(QuSlotDebug())
	{
		printf("TimeIntervalSlider::TimeIntervalSlider \n");
		fflush(stdout);
	}

	obj=0;
    
	QHBoxLayout * mainLayout=new QHBoxLayout;
	
	titleBase = QString(title.latin1()); // deep copy
   
	this->record = 0; // init to be safe, should be reset by parent
	this->maxRecord = 0; // init to be safe, should be reset by parent
   
	QToolButton * leftEnd = new QToolButton( this, "leftEnd" );
	leftEnd->setText("|");
	leftEnd->setFixedSize(15,21);
	mainLayout->addWidget(leftEnd);
	
	QToolButton * leftSlide = new QToolButton( Qt::LeftArrow, this, "");
	leftSlide->setFixedSize(20,21);
	mainLayout->addWidget(leftSlide);
   
	recBar = new QSlider(Qt::Horizontal, this, "recBar");
	mainLayout->addWidget(recBar);
  
	QToolButton * rightSlide = new QToolButton( Qt::RightArrow, this, "");
	rightSlide->setFixedSize(20,21);
	mainLayout->addWidget(rightSlide);
	
	QToolButton * rightEnd = new QToolButton( this, "rightEnd" );
	rightEnd->setText("|");
	rightEnd->setFixedSize(15,21);
	mainLayout->addWidget(rightEnd);
	
	connect( rightSlide, SIGNAL(clicked()), this, SLOT(incRec()) );
	connect( leftSlide,  SIGNAL(clicked()), this, SLOT(decRec()) );
	connect( rightEnd, SIGNAL(clicked()), this, SLOT(lastRec()) );
	connect( leftEnd,  SIGNAL(clicked()), this, SLOT(firstRec()) );
	connect( recBar,   SIGNAL(valueChanged(int)), this, SLOT(adjustRec(int)) );
   
	recField = new QguWLinput(this,"recField");
	recField->acceptInteger = true;
	recField->setEnabled(true);
	recField->setFixedWidth(60);
	recField->setAlignment(Qt::AlignHCenter);
	recField->setText("0");
	connect( recField,   SIGNAL(textChanged(const QString &)), this, SLOT(setFromTyping()) );
	mainLayout->addWidget(recField);

	recBar->setValue(0);
    
	setBaseSize(200, 30);
	
	setLayout(mainLayout);

}
#endif

void TimeIntervalSlider::updateTitle(){
  
    if(QuSlotDebug())
    {
		printf("TimeIntervalSlider::updateTitle \n");
		fflush(stdout);
	}

    if(obj!=NULL)
    {
        QString text=titleBase + " ";
        text+=obj[record].start().iso().c_str();
        text+=" / ";
        text+=obj[record].end().iso().c_str();
     this->setTitle(text);
    }
}

void TimeIntervalSlider::setFromTyping(){

    if(QuSlotDebug())
    {
		printf("TimeIntervalSlider::setFromTyping %ld\n", this);
		fflush(stdout);
	}
   
    bool OK;
    int newValue = recField->text().toInt(&OK);
    if (!OK) return;

    if(newValue > maxRecord){
        newValue = maxRecord;
        recField->setText(tr("%1").arg(newValue));
    }
    else if(newValue < 0) {
        newValue = 0;
        recField->setText("0");
    }
  
    recBar->setValue(newValue);
    this->updateTitle();
}

void TimeIntervalSlider::incRec(){

    if(QuSlotDebug())
    {
		printf("TimeIntervalSlider::incRec %ld\n", this);
		fflush(stdout);
	}

    if(record < maxRecord){
        record +=1;
    }
    else{
        record = maxRecord;
    }
    recBar->setValue(record);
    this->updateTitle();
}

void TimeIntervalSlider::decRec(){

    if(QuSlotDebug())
    {
		printf("TimeIntervalSlider::decRec %ld\n", this);
		fflush(stdout);
	}
 
    if(record > 0){
        record -= 1;
    }
    else{
        record = 0;
    }
    recBar->setValue(record);
    this->updateTitle();
}

void TimeIntervalSlider::lastRec(){

    if(QuSlotDebug())
    {
		printf("TimeIntervalSlider::lastRec %ld\n", this);
		fflush(stdout);
	}
 
    record = maxRecord;
    recBar->setValue(record);
    this->updateTitle();
}

void TimeIntervalSlider::firstRec(){

    if(QuSlotDebug())
    {
		printf("TimeIntervalSlider::firstRec %ld\n", this);
		fflush(stdout);
	}
 
    record = 0;
    recBar->setValue(record);
    this->updateTitle();
}

void TimeIntervalSlider::adjustRec(int rec){

    if(QuSlotDebug())
	{
		printf("TimeIntervalSlider::adjustRec %ld %d\n", this, rec);
		fflush(stdout);
	}
 
    record = rec;
    recField->setText(tr("%1").arg(rec));
    this->updateTitle();
  
    emit sliderChanged();
}

// class timeSlider
//-------------------
#if QT_VERSION < 0x040000
 timeSlider::timeSlider(QdObject_var dobj, bool acceptTSlice, bool acceptTSubset, QWidget *parent, char * name)
   : Q_V_GROUP_BOX(parent,name)
{

 if(QuSlotDebug())
 {
	printf("timeSlider::timeSlider \n");
	fflush(stdout);
 }

 this->setMargin(QSAS_profiles->current()->GUImargins);
   this->setTitle(" Time Options ");
   int i=0;
   Isubset = -1;
   Iaverage = -1;
   Islice = -1;
   Inothing = -1;
   
    // Combo box
   sliceChoice = new QComboBox( false, this, "sliceChoice" );
   if(acceptTSubset)  {
     Isubset = i;
	    sliceChoice->insertItem( "Subset over time range", i++);
   }
   else{
     Inothing = i;
     sliceChoice->insertItem( "Keep as Time Series", i++);
   }
   
   if(acceptTSlice)  {     
     Iaverage = i;
     sliceChoice->insertItem( "Average over time range", i++);

     Islice = i;
     sliceChoice->insertItem( "Slice at time", i++ );
   }
   sliceChoice->resize( sliceChoice->sizeHint() );
   connect( sliceChoice, SIGNAL(activated(int)), this, SLOT(updateTimeSlider(int)) );
   
   this->addSpace(10);

   fromSlider = new  RecordSlider("From  ", this, "fromSlider");
   toSlider =   new  RecordSlider("To  ", this, "toSlider");
   atSlider =   new  RecordSlider("At  ", this, "atSlider");

   connect(fromSlider, SIGNAL(recordSliderChanged()), this, SLOT(resetEnd()) );
   connect(toSlider, SIGNAL(recordSliderChanged()), this, SLOT(resetStart()) );

   int seqLength = dobj->sequence_size();
   fromSlider->recBar->setRange(0, seqLength-1);
   toSlider->recBar->setRange(0, seqLength-1);
   atSlider->recBar->setRange(0, seqLength-1);
   
   fromSlider->maxRecord = seqLength-1;
   toSlider->maxRecord = seqLength-1;
   atSlider->maxRecord = seqLength-1;
   toSlider->recBar->setValue(toSlider->maxRecord);

   // get time line and keep globally for this file
   QdTimeSeq_var tt = get_timetags(dobj);
   if( !tt.is_nil() ){
     fromSlider->ttags = tt;
     toSlider->ttags = tt;
     atSlider->ttags = tt;
   }
  
   // now ttags are set, add time feedback to title
   fromSlider->updateTitle();
   toSlider->updateTitle();
   atSlider->updateTitle();

   if(acceptTSubset){
     fromSlider->show();
     toSlider->show();
     atSlider->hide();
   }
   else{
     fromSlider->hide();
     toSlider->hide();
     atSlider->hide();
   }
   
   this->setEnabled(true);
}
#else
 timeSlider::timeSlider(QdObject_var dobj, bool acceptTSlice, bool acceptTSubset, QWidget *parent, char * name)
	: QGroupBox(parent,name)
{
	if(QuSlotDebug())
	{
		printf("timeSlider::timeSlider \n");
		fflush(stdout);
	}
	
	mainLayout=new QVBoxLayout;
	
	mainLayout->setMargin(QSAS_profiles->current()->GUImargins);
	this->setTitle(" Time Options ");
	int i=0;
	Isubset = -1;
	Iaverage = -1;
	Islice = -1;
	Inothing = -1;
	
	 // Combo box
	sliceChoice = new QComboBox( false, this, "sliceChoice" );
	if(acceptTSubset)  {
		Isubset = i;
		sliceChoice->insertItem( "Subset over time range", i++);
	}
	else{
		Inothing = i;
		sliceChoice->insertItem( "Keep as Time Series", i++);
	}
	if(acceptTSlice)  {     
		Iaverage = i;
		sliceChoice->insertItem( "Average over time range", i++);

		Islice = i;
		sliceChoice->insertItem( "Slice at time", i++ );
	}
	sliceChoice->resize( sliceChoice->sizeHint() );
	connect( sliceChoice, SIGNAL(activated(int)), this, SLOT(updateTimeSlider(int)) );
	
	mainLayout->addWidget(sliceChoice);
	
	//
	
	mainLayout->addSpacing(10);
	
	// Sliders
	fromSlider = new  RecordSlider("From  ", this, "fromSlider");
	mainLayout->addWidget(fromSlider);
	toSlider =   new  RecordSlider("To  ", this, "toSlider");
	mainLayout->addWidget(toSlider);
	atSlider =   new  RecordSlider("At  ", this, "atSlider");
	mainLayout->addWidget(atSlider);
	fromSlider->resize(fromSlider->sizeHint());
	toSlider->resize(toSlider->sizeHint());
	atSlider->resize(atSlider->sizeHint());
	
	connect(fromSlider, SIGNAL(recordSliderChanged()), this, SLOT(resetEnd()) );
	connect(toSlider, SIGNAL(recordSliderChanged()), this, SLOT(resetStart()) );
	
	int seqLength = dobj->sequence_size();
	fromSlider->recBar->setRange(0, seqLength-1);
	toSlider->recBar->setRange(0, seqLength-1);
	atSlider->recBar->setRange(0, seqLength-1);
   
	fromSlider->maxRecord = seqLength-1;
	toSlider->maxRecord = seqLength-1;
	atSlider->maxRecord = seqLength-1;
	toSlider->recBar->setValue(toSlider->maxRecord);
	
	// get time line and keep globally for this file
	QdTimeSeq_var tt = get_timetags(dobj);
	if( !tt.is_nil() ){
		fromSlider->ttags = tt;
		toSlider->ttags = tt;
		atSlider->ttags = tt;
	}
  
   	// now ttags are set, add time feedback to title
	fromSlider->updateTitle();
	toSlider->updateTitle();
	atSlider->updateTitle();

	if(acceptTSubset){
		fromSlider->show();
		toSlider->show();
		atSlider->hide();
	}
	else{
		fromSlider->hide();
		toSlider->hide();
		atSlider->hide();
	}
   
	this->setMinimumHeight( sliceChoice->sizeHint().height() +  atSlider->sizeHint().height()*2.5);
	this->setEnabled(true);
	setLayout(mainLayout);
}
#endif

void timeSlider::resetStart(){
  if(fromSlider->record > toSlider->record) {
     fromSlider->recBar->setValue(toSlider->record);
  }
}

void timeSlider::resetEnd(){
  if(fromSlider->record > toSlider->record)  {
     toSlider->recBar->setValue(fromSlider->record);
  }
}


void timeSlider::updateTimeSlider(int opChoice){

  if (opChoice == Isubset){
    fromSlider->show();
    toSlider->show();
    atSlider->hide();
  }
  else if(opChoice == Iaverage){
    fromSlider->show();
    toSlider->show();
    atSlider->hide();
  }
  else if(opChoice == Islice){
    fromSlider->hide();
    toSlider->hide();
    atSlider->show();
  }
  else if(opChoice == Inothing){
    fromSlider->hide();
    toSlider->hide();
    atSlider->hide();
  }
  else {
    cout << " Slice option not available " << endl;
  }
}

bool timeSlider::isDefault(){

 if(QuSlotDebug())
 {
	printf("timeSlider::isDefault \n");
	fflush(stdout);
 }

  if (sliceChoice->currentItem() == Isubset){
    if(fromSlider->record == 0 && toSlider->record == toSlider->maxRecord) return true;
  }
  return false;
}

// Class MakeVectorWindow
//------------------------

MakeVectorWindow::MakeVectorWindow(QWidget *parent, const char *name ) 
  : QMainWindow(parent,name)
#if QT_VERSION < 0x040000
{

 if(QuSlotDebug())
 {
	printf("MakeVectorWindow::MakeVectorWindow \n");
	fflush(stdout);
  }
  
   this->setPalette(qApp->palette());
   this->setCaption("Add vector metadata to 3 element Data Array ");

   QFontMetrics qgu_font_mt(qApp->font());
   int char_width = qgu_font_mt.width("n");
   
   QWidget::setWFlags(Qt::WDestructiveClose);
  
   msgBar = statusBar();
   
   QVGroupBox *topVB = new QVGroupBox( this, "TopWidget" );
   setCentralWidget( topVB );

   QHGroupBox *slotPanel = new QHGroupBox(topVB, "slot");
   slotPanel->setTitle(" Data Array Object ");
   
   objField = new QguWLinput(slotPanel,"objField");
   objField->acceptObjectPlus = true; 
   objField->setEnabled(true);
   objField->setMinimumWidth(char_width*55);
   connect(objField, SIGNAL(validated_OK()), this, SLOT(testArray()));


   
   menuBar()->insertSeparator();   
   Q_POPUP_MENU * help = new Q_POPUP_MENU( this );
   menuBar()->insertItem( "  Help ", help );
//   help->insertItem("QSAS &Manual", this, SLOT( launchManual() ), CTRL+Key_M );
   help->insertSeparator();
 //help->insertItem("...Slice", this, SLOT( launchManSlice() ));
   help->insertSeparator();
//   help->insertItem("Module Version", this, SLOT( launchVersion() ));
 
   QHGroupBox *frame_panel=new QHGroupBox(topVB, "frame_panel");
   QLabel *frame_label = new QLabel("Frame    ", frame_panel);
   frame_field = new QguWLinput(frame_panel, "frame_field");
   frame_field->acceptText = true;
   frame_field->setEnabled(true);
   frame_field->setMinimumWidth(char_width*25);
   frame_field->setText("gse");
 
   QHGroupBox *rep_panel=new QHGroupBox(topVB, "rep_panel");
   QLabel *rep_label = new QLabel("Representation ", rep_panel);
   rep_field = new QComboBox(rep_panel, "rep_field");
   rep_field->setEnabled(true);
   rep_field->insertItem("xyz");
   rep_field->insertItem("rlp (rad)");
   rep_field->insertItem("rlp (deg)");
   rep_field->insertItem("rtp (rad)");
   rep_field->insertItem("rtp (deg)");
	 

  QHGroupBox *buttons_layout=new QHGroupBox(topVB, "buttons");
    
  doBtn = new QPushButton( "Convert into Vector", buttons_layout, "doBtn" );
  doBtn->setMinimumSize(doBtn->sizeHint());
  connect( doBtn, SIGNAL(clicked()), this, SLOT(makeVector()) );
  doBtn->setFocusPolicy(STRONG_FOCUS);

  dismissBtn = new QPushButton( "Dismiss", buttons_layout, "dismissBtn" );
  dismissBtn->setMinimumSize(dismissBtn->sizeHint());
  connect( dismissBtn, SIGNAL(clicked()), this, SLOT(close()) );
  dismissBtn->setDefault(true);
  
  this->setMinimumSize(this->sizeHint());
}
#else
{

	if(QuSlotDebug())
	{
		printf("MakeVectorWindow::MakeVectorWindow \n");
		fflush(stdout);
	}
  
	this->setPalette(qApp->palette());
	this->setCaption("Add vector metadata to 3 element Data Array ");

	QFontMetrics qgu_font_mt(qApp->font());
	int char_width = qgu_font_mt.width("n");
   
	setAttribute(Qt::WA_DeleteOnClose);
		
	// Status bar
	
	msgBar = statusBar();
	
	// Main widget
	
	QWidget * mainWidget=new QWidget(this);
	QVBoxLayout * mainLayout=new QVBoxLayout;
	
	// Slot widget
	
	QGroupBox * slotPanel=new QGroupBox(mainWidget);
	QHBoxLayout * slotLayout=new QHBoxLayout;
	
	slotPanel->setTitle(" Data Array Object ");
	
	objField = new QguWLinput(slotPanel,"objField");
	objField->acceptObjectPlus = true; 
	objField->setEnabled(true);
	objField->setMinimumWidth(char_width*55);
	connect(objField, SIGNAL(validated_OK()), this, SLOT(testArray()));
	slotLayout->addWidget(objField);
	
	slotPanel->setLayout(slotLayout);
	mainLayout->addWidget(slotPanel);
	
	// Frame panel
	
	QGroupBox * frame_panel=new QGroupBox(mainWidget);
	QHBoxLayout * frame_layout=new QHBoxLayout;
	
	QLabel *frame_label = new QLabel("Frame    ", frame_panel);
	frame_layout->addWidget(frame_label);
	
	frame_field = new QguWLinput(frame_panel, "frame_field");
	frame_field->acceptText = true;
	frame_field->setEnabled(true);
	frame_field->setMinimumWidth(char_width*25);
	frame_field->setText("gse");
	frame_layout->addWidget(frame_field);
	
	frame_panel->setLayout(frame_layout);
	mainLayout->addWidget(frame_panel);
	
	// Representation panel
	
	QGroupBox * rep_panel=new QGroupBox(mainWidget);
	QHBoxLayout * rep_layout=new QHBoxLayout;
	
	QLabel *rep_label = new QLabel("Representation ", rep_panel);
	rep_layout->addWidget(rep_label);
	
	rep_field = new QComboBox(rep_panel, "rep_field");
	rep_field->setEnabled(true);
	rep_field->insertItem("xyz");
	rep_field->insertItem("rlp (rad)");
	rep_field->insertItem("rlp (deg)");
	rep_field->insertItem("rtp (rad)");
	rep_field->insertItem("rtp (deg)");
	rep_layout->addWidget(rep_field);
	
	rep_panel->setLayout(rep_layout);
	mainLayout->addWidget(rep_panel);
	
	// Buttons layout
	
	QHBoxLayout * buttons_layout=new QHBoxLayout;
	
	doBtn = new QPushButton( "Convert into Vector");
	doBtn->setMinimumSize(doBtn->sizeHint());
	connect( doBtn, SIGNAL(clicked()), this, SLOT(makeVector()) );
	doBtn->setFocusPolicy(STRONG_FOCUS);
	buttons_layout->addWidget(doBtn);

	dismissBtn = new QPushButton( "Dismiss");
	dismissBtn->setMinimumSize(dismissBtn->sizeHint());
	connect( dismissBtn, SIGNAL(clicked()), this, SLOT(close()) );
	dismissBtn->setDefault(true);
	buttons_layout->addWidget(dismissBtn);
	
	mainLayout->addLayout(buttons_layout);
	
	
	
	mainWidget->setLayout(mainLayout);
	setCentralWidget(mainWidget);
	
	this->setMinimumSize(this->sizeHint());
}
#endif


void MakeVectorWindow::setObject(const char * WLctxt){
    
 if(QuSlotDebug())
 {
	printf("MakeVectorWindow::setObject \n");
	fflush(stdout);
 }
  
   objField->setText(WLctxt);
   objField->reValidate();
   this->testArray();

}


void MakeVectorWindow::testArray(){
    
 if(QuSlotDebug())
 {
	printf("MakeVectorWindow::setObject \n");
	fflush(stdout);
 }
  
   // if input not 3 comp array disable doBtn based on component subset
   QdObject_var dobj = QuGetObjByName(objField->text());
   QdRMatrix_var rmatrix=QdRMatrix_var::narrow(dobj);
   if(rmatrix.is_nil())
   {
     QdRMatrixSeq_var rmatrixseq=QdRMatrixSeq_var::narrow(dobj);
	 if(rmatrixseq.is_nil()){
	   doBtn->setEnabled(false);
	   return; // not a matrix object, cannot convert
	 }
   }
   doBtn->setEnabled(true);
}


void MakeVectorWindow::keyPressEvent(QKeyEvent * e )
{  
     
 if(QuSlotDebug())
 {
	printf("DataSliceWindow::keyPressEvent \n");
	fflush(stdout);
 }
  
	  // use escape key to toggle default buttons
  if(e->key() == Qt::Key_Escape){
	  if( doBtn->isDefault()  ) {
		doBtn->setDefault(false);
		   dismissBtn->setDefault(true);
	  }
	  else if(dismissBtn->isDefault()  )  {
		   dismissBtn->setDefault(false);
		   doBtn->setDefault(true);
	  }
	  return;
  }
  // use return or enter key to perform default
  if(e->key() == Qt::Key_Return ||  e->key() == Qt::Key_Enter) {
     if(doBtn->isDefault()) makeVector();
	 else if (dismissBtn->isDefault()) close();
     return;
  }

  // else pass on to list view 
  e->ignore(); // let selection change

}

void MakeVectorWindow::makeVector(){
      
 if(QuSlotDebug())
 {
	printf("MakeVectorWindow::makeVector \n");
	fflush(stdout);
 }

 
  QString objString = objField->text();
  
  QdObject_var newObject = QuGetObjByNameAndComp(objString);
  
  // attach xrefs here
  QString repTxt = QString(rep_field->currentText().latin1());

  if( repTxt.contains("rlp",false) > 0){
	   QString SI_conv = QString(QuGetAttrText("SI_CONVERSION", newObject));
       if( repTxt.contains("deg",false) > 0){
	     int ptr = SI_conv.find(",");
		 if(ptr > -1) SI_conv.truncate(ptr);
		 SI_conv += ", deg, deg";
       }
	   else{
	     int ptr = SI_conv.find(",");
		 if(ptr > -1) SI_conv.truncate(ptr);
		 SI_conv += ", rad, rad";	   
	   }
	   QuSetTxtAttr("SI_CONVERSION", SI_conv.latin1(),  newObject);
       QuSetFrame( newObject, "vector", frame_field->text().latin1(), "rlp");
  }
  else if(repTxt.contains("rtp",false) > 0){
	   QString SI_conv = QString(QuGetAttrText("SI_CONVERSION", newObject));
       if( repTxt.contains("deg",false) > 0){
	     int ptr = SI_conv.find(",");
		 if(ptr > -1) SI_conv.truncate(ptr);
		 SI_conv += ", deg, deg";
       }
	   else{
	     int ptr = SI_conv.find(",");
		 if(ptr > -1) SI_conv.truncate(ptr);
		 SI_conv += ", rad, rad";	   
	   }  
	   QuSetTxtAttr("SI_CONVERSION", SI_conv.latin1(),  newObject);
       QuSetFrame( newObject, "vector", frame_field->text().latin1(), "rtp");
  }

  else{
    QuSetFrame( newObject, "vector", frame_field->text().latin1(), "xyz");
  }

  if( !QuIsThreeVector(newObject) ) {
    QuMsgBar("VECTOR FAILED: Input array needs 3 components", true);
	return;
  }
  
  if( !(newObject.is_nil() ) ) QwuiPutOnWL(newObject, objString.latin1(), "Data", true);
  else QuMsgBar("NULL object after conversion to vector", true);

  this->close();
  return;
}

