#include "QCCWindow.h"

#if QT_VERSION < 0x040000
#define CASE_INSENSITIVE FALSE
#else
#define CASE_INSENSITIVE Qt::CaseInsensitive
#endif

#if QT_VERSION < 0x040000
QCCWindow::QCCWindow()
{	
	QPopupMenu * plotMenu = new QPopupMenu( this );
	menuBar()->insertItem("Plot", plotMenu);
	plotMenu->insertItem("Save Plot As PNG", this, SLOT(savePlotPNG()));
	plotMenu->insertItem("Save Plot As JPG", this, SLOT(savePlotJPG()));
#if defined(Q_OS_MACX) || QT_VERSION >= 0x040000
	plotMenu->insertItem("Save Plot As PDF", this, SLOT(savePlotPDF()));
#endif
#if !(defined(Q_OS_MACX) && QT_VERSION < 0x040000)
	plotMenu->insertItem("Save Plot As EPS", this, SLOT(savePlotPS()));
#endif
	plotMenu->insertItem("Print", this, SLOT(printPlot()));
	
	// Build interface
	
	QLabel* widget=new QLabel(this);
	QGridLayout * layout=new QGridLayout(widget);
	setCentralWidget(widget);
	
	plot=new QtPLWidget(600, 600, 600., 600., widget);
	layout->addWidget(plot, 0, 0);
	layout->setColStretch(0, 10);
	
	sliderAlt=new QSlider(Qt::Vertical, widget);
	sliderAlt->setRange(0, 90);
	sliderAlt->setValue(45);
	layout->addWidget(sliderAlt, 0, 1);
	connect(sliderAlt, SIGNAL(sliderMoved(int)), this, SLOT(altChanged(int)));
	
	sliderAz=new QSlider(Qt::Horizontal, widget);
	sliderAz->setRange(0, 360);
	sliderAz->setValue(45);
	layout->addWidget(sliderAz, 1, 0);
	connect(sliderAz, SIGNAL(sliderMoved(int)), this, SLOT(azChanged(int)));
	
	QGridLayout * displayParamLayout=new QGridLayout;
	layout->addLayout(displayParamLayout, 0, 2);
	layout->setColStretch(2, 3);
	
	displayParamLayout->addWidget(new QLabel("Title", widget), 0, 0);
	editTitle=new QLineEdit(widget); displayParamLayout->addWidget(editTitle, 1,0);
	connect(editTitle, SIGNAL(textChanged(const QString&)), this, SLOT(newTitle(const QString&)));
	displayParamLayout->addWidget(new QLabel("Axis Units", widget), 2, 0);
	editAxis=new QLineEdit(widget); displayParamLayout->addWidget(editAxis, 3,0);
	connect(editAxis, SIGNAL(textChanged(const QString&)), this, SLOT(newFrame(const QString&)));
	
	bgColour=Qt::cyan;
	buttonBgColour=new QPushButton("Box Colour", widget);
	QPalette pBg=buttonBgColour->palette(); pBg.setColor(QPalette::Active ,QColorGroup::ButtonText, bgColour); buttonBgColour->setPalette(pBg);
	displayParamLayout->addWidget(buttonBgColour, 4, 0);
	connect(buttonBgColour, SIGNAL(clicked()), this, SLOT(changeColourBg()));
	
	projX=new QCheckBox("No X component", widget);
	displayParamLayout->addWidget(projX, 5, 0);
	connect(projX, SIGNAL(stateChanged(int)), this, SLOT(projXChanged(int)));
	projY=new QCheckBox("No Y component", widget);
	displayParamLayout->addWidget(projY, 6, 0);
	connect(projY, SIGNAL(stateChanged(int)), this, SLOT(projYChanged(int)));
	projZ=new QCheckBox("No Z component", widget);
	displayParamLayout->addWidget(projZ, 7, 0);
	connect(projZ, SIGNAL(stateChanged(int)), this, SLOT(projZChanged(int)));
	
	displayParamLayout->addWidget(new QLabel(widget), 8, 0);
	displayParamLayout->setRowStretch(8, 10);
	
	QGridLayout * clusterParamLayout=new QGridLayout;
	layout->addMultiCell(clusterParamLayout, 2, 2, 0, 3, Qt::AlignCenter);
	clusterParamLayout->addWidget(new QLabel("C1", widget), 0, 1);
	clusterParamLayout->addWidget(new QLabel("C2", widget), 0, 2);
	clusterParamLayout->addWidget(new QLabel("C3", widget), 0, 3);
	clusterParamLayout->addWidget(new QLabel("C4", widget), 0, 4);
	
	clusterParamLayout->addWidget(new QLabel("X", widget), 1, 0);
	clusterParamLayout->addWidget(new QLabel("Y", widget), 2, 0);
	clusterParamLayout->addWidget(new QLabel("Z", widget), 3, 0);
	clusterParamLayout->addWidget(new QLabel("Colour", widget), 4, 0);
	
	QDoubleValidator * v=new QDoubleValidator(-1.e7, 1.e7, 2, 0);
	
	editX1=new QLineEdit(widget); editX1->setValidator(v);
	clusterParamLayout->addWidget(editX1, 1, 1);
	connect(editX1, SIGNAL(textChanged(const QString&)), this, SLOT(newX1(const QString&)));
	
	editY1=new QLineEdit(widget); editY1->setValidator(v);
	clusterParamLayout->addWidget(editY1, 2, 1);
	connect(editY1, SIGNAL(textChanged(const QString&)), this, SLOT(newY1(const QString&)));
	
	editZ1=new QLineEdit(widget); editZ1->setValidator(v);
	clusterParamLayout->addWidget(editZ1, 3, 1);
	connect(editZ1, SIGNAL(textChanged(const QString&)), this, SLOT(newZ1(const QString&)));
	
	editX2=new QLineEdit(widget); editX2->setValidator(v);
	clusterParamLayout->addWidget(editX2, 1, 2);
	connect(editX2, SIGNAL(textChanged(const QString&)), this, SLOT(newX2(const QString&)));
	
	editY2=new QLineEdit(widget); editY2->setValidator(v);
	clusterParamLayout->addWidget(editY2, 2, 2);
	connect(editY2, SIGNAL(textChanged(const QString&)), this, SLOT(newY2(const QString&)));
	
	editZ2=new QLineEdit(widget); editZ2->setValidator(v);
	clusterParamLayout->addWidget(editZ2, 3, 2);
	connect(editZ2, SIGNAL(textChanged(const QString&)), this, SLOT(newZ2(const QString&)));
	
	editX3=new QLineEdit(widget); editX3->setValidator(v);
	clusterParamLayout->addWidget(editX3, 1, 3);
	connect(editX3, SIGNAL(textChanged(const QString&)), this, SLOT(newX3(const QString&)));
	
	editY3=new QLineEdit(widget); editY3->setValidator(v);
	clusterParamLayout->addWidget(editY3, 2, 3);
	connect(editY3, SIGNAL(textChanged(const QString&)), this, SLOT(newY3(const QString&)));
	
	editZ3=new QLineEdit(widget); editZ3->setValidator(v);
	clusterParamLayout->addWidget(editZ3, 3, 3);
	connect(editZ3, SIGNAL(textChanged(const QString&)), this, SLOT(newZ3(const QString&)));
	
	editX4=new QLineEdit(widget); editX4->setValidator(v);
	clusterParamLayout->addWidget(editX4, 1, 4);
	connect(editX4, SIGNAL(textChanged(const QString&)), this, SLOT(newX4(const QString&)));
	
	editY4=new QLineEdit(widget); editY4->setValidator(v);
	clusterParamLayout->addWidget(editY4, 2, 4);
	connect(editY4, SIGNAL(textChanged(const QString&)), this, SLOT(newY4(const QString&)));
	
	editZ4=new QLineEdit(widget); editZ4->setValidator(v);
	clusterParamLayout->addWidget(editZ4, 3, 4);
	connect(editZ4, SIGNAL(textChanged(const QString&)), this, SLOT(newZ4(const QString&)));
	
	buttonCol1=new QPushButton("Change", widget);
	QPalette p1=buttonCol1->palette(); p1.setColor(QPalette::Active ,QColorGroup::ButtonText, Qt::black); buttonCol1->setPalette(p1);
	clusterParamLayout->addWidget(buttonCol1, 4, 1);
	connect(buttonCol1, SIGNAL(clicked()), this, SLOT(changeColour1()));
	
	buttonCol2=new QPushButton("Change", widget);
	QPalette p2=buttonCol2->palette(); p2.setColor(QPalette::Active ,QColorGroup::ButtonText, Qt::red); buttonCol2->setPalette(p2);
	clusterParamLayout->addWidget(buttonCol2, 4, 2);
	connect(buttonCol2, SIGNAL(clicked()), this, SLOT(changeColour2()));
	
	buttonCol3=new QPushButton("Change", widget);
	QPalette p3=buttonCol3->palette(); p3.setColor(QPalette::Active ,QColorGroup::ButtonText, Qt::green); buttonCol3->setPalette(p3);
	clusterParamLayout->addWidget(buttonCol3, 4, 3);
	connect(buttonCol3, SIGNAL(clicked()), this, SLOT(changeColour3()));
	
	buttonCol4=new QPushButton("Change", widget);
	QPalette p4=buttonCol4->palette(); p4.setColor(QPalette::Active ,QColorGroup::ButtonText, Qt::magenta); buttonCol4->setPalette(p4);
	clusterParamLayout->addWidget(buttonCol4, 4, 4);
	connect(buttonCol4, SIGNAL(clicked()), this, SLOT(changeColour4()));
	
	// Init plplot
	
// 	setCentralWidget(plot);
	plmkstrm(&strm);
	plsetqtdev((void*)plot);
	plsdev("qt");
	plinit();
	
	resetColourIndex();
	
	resize(600, 600);
	
	// Init variables
	
	for(int i=0; i<4; ++i)
	{
		for(int j=0; j<3; ++j)
			clusterParameters[i].position[j]=0.;
	}

	clusterParameters[0].colour=Qt::black;
	clusterParameters[1].colour=Qt::red;
	clusterParameters[2].colour=Qt::green;
	clusterParameters[3].colour=Qt::magenta;
	
	alt=45;
	az=45;
	projection[0]=projection[1]=projection[2]=false;
}
#else
QCCWindow::QCCWindow()
{
	QMenu * plotMenu = new QMenu( this );
	menuBar()->insertItem("Plot", plotMenu);
	plotMenu->insertItem("Save Plot As PNG", this, SLOT(savePlotPNG()));
	plotMenu->insertItem("Save Plot As JPG", this, SLOT(savePlotJPG()));
#if QT_VERSION >= 0x040300
	plotMenu->insertItem("Save Plot As SVG", this, SLOT(savePlotSVG()));
#endif
	plotMenu->insertItem("Save Plot As PDF", this, SLOT(savePlotPDF()));
#if !defined(Q_OS_MACX)
	plotMenu->insertItem("Save Plot As EPS", this, SLOT(savePlotPS()));
#endif
	plotMenu->insertItem("Print", this, SLOT(printPlot()));

	
	// Build interface
	
	QLabel* widget=new QLabel(this);
	QGridLayout * layout=new QGridLayout(widget);
	setCentralWidget(widget);
	
	plot=new QtPLWidget(600, 600, 600., 600., 0);
	layout->addWidget(plot, 0, 0);
	layout->setColumnStretch(0, 10);
	
	sliderAlt=new QSlider(Qt::Vertical, this);
	sliderAlt->setRange(-90, 90);
	sliderAlt->setSingleStep(5);
	sliderAlt->setValue(45);
	layout->addWidget(sliderAlt, 0, 1);
	connect(sliderAlt, SIGNAL(sliderMoved(int)), this, SLOT(altChanged(int)));
	
	sliderAz=new QSlider(Qt::Horizontal, this);
	sliderAz->setRange(0, 360);
	sliderAz->setSingleStep(5);
	sliderAz->setValue(45);
	layout->addWidget(sliderAz, 1, 0);
	connect(sliderAz, SIGNAL(sliderMoved(int)), this, SLOT(azChanged(int)));
	
	QGridLayout * displayParamLayout=new QGridLayout;
	layout->addLayout(displayParamLayout, 0, 2, Qt::AlignCenter);
	layout->setColumnStretch(2, 3);
	
	displayParamLayout->addWidget(new QLabel("Title"), 0, 0);
	editTitle=new QLineEdit; displayParamLayout->addWidget(editTitle, 1,0);
	connect(editTitle, SIGNAL(textEdited(const QString&)), this, SLOT(newTitle(const QString&)));
	displayParamLayout->addWidget(new QLabel("Axis Units"), 2, 0);
	editAxis=new QLineEdit; displayParamLayout->addWidget(editAxis, 3,0);
	connect(editAxis, SIGNAL(textEdited(const QString&)), this, SLOT(newFrame(const QString&)));
	
	bgColour=Qt::cyan;
	buttonBgColour=new QPushButton("Box Colour");
	QPalette pBg; pBg.setColor(QPalette::Active ,QColorGroup::ButtonText, bgColour); buttonBgColour->setPalette(pBg);
	displayParamLayout->addWidget(buttonBgColour, 4, 0);
	connect(buttonBgColour, SIGNAL(clicked()), this, SLOT(changeColourBg()));
	
	projX=new QCheckBox("No X component");
	displayParamLayout->addWidget(projX, 5, 0);
	connect(projX, SIGNAL(stateChanged(int)), this, SLOT(projXChanged(int)));
	projY=new QCheckBox("No Y component");
	displayParamLayout->addWidget(projY, 6, 0);
	connect(projY, SIGNAL(stateChanged(int)), this, SLOT(projYChanged(int)));
	projZ=new QCheckBox("No Z component");
	displayParamLayout->addWidget(projZ, 7, 0);
	connect(projZ, SIGNAL(stateChanged(int)), this, SLOT(projZChanged(int)));
	
	QGridLayout * clusterParamLayout=new QGridLayout;
	layout->addLayout(clusterParamLayout, 2, 0, 1, 3, Qt::AlignCenter);
	clusterParamLayout->addWidget(new QLabel("C1"), 0, 1);
	clusterParamLayout->addWidget(new QLabel("C2"), 0, 2);
	clusterParamLayout->addWidget(new QLabel("C3"), 0, 3);
	clusterParamLayout->addWidget(new QLabel("C4"), 0, 4);
	
	clusterParamLayout->addWidget(new QLabel("X"), 1, 0);
	clusterParamLayout->addWidget(new QLabel("Y"), 2, 0);
	clusterParamLayout->addWidget(new QLabel("Z"), 3, 0);
	clusterParamLayout->addWidget(new QLabel("Colour"), 4, 0);
	
	QDoubleValidator * v=new QDoubleValidator(-1.e7, 1.e7, 2, 0);
	
	editX1=new QLineEdit; editX1->setValidator(v);
	clusterParamLayout->addWidget(editX1, 1, 1);
	connect(editX1, SIGNAL(textEdited(const QString&)), this, SLOT(newX1(const QString&)));
	
	editY1=new QLineEdit; editY1->setValidator(v);
	clusterParamLayout->addWidget(editY1, 2, 1);
	connect(editY1, SIGNAL(textEdited(const QString&)), this, SLOT(newY1(const QString&)));
	
	editZ1=new QLineEdit; editZ1->setValidator(v);
	clusterParamLayout->addWidget(editZ1, 3, 1);
	connect(editZ1, SIGNAL(textEdited(const QString&)), this, SLOT(newZ1(const QString&)));
	
	editX2=new QLineEdit; editX2->setValidator(v);
	clusterParamLayout->addWidget(editX2, 1, 2);
	connect(editX2, SIGNAL(textEdited(const QString&)), this, SLOT(newX2(const QString&)));
	
	editY2=new QLineEdit; editY2->setValidator(v);
	clusterParamLayout->addWidget(editY2, 2, 2);
	connect(editY2, SIGNAL(textEdited(const QString&)), this, SLOT(newY2(const QString&)));
	
	editZ2=new QLineEdit; editZ2->setValidator(v);
	clusterParamLayout->addWidget(editZ2, 3, 2);
	connect(editZ2, SIGNAL(textEdited(const QString&)), this, SLOT(newZ2(const QString&)));
	
	editX3=new QLineEdit; editX3->setValidator(v);
	clusterParamLayout->addWidget(editX3, 1, 3);
	connect(editX3, SIGNAL(textEdited(const QString&)), this, SLOT(newX3(const QString&)));
	
	editY3=new QLineEdit; editY3->setValidator(v);
	clusterParamLayout->addWidget(editY3, 2, 3);
	connect(editY3, SIGNAL(textEdited(const QString&)), this, SLOT(newY3(const QString&)));
	
	editZ3=new QLineEdit; editZ3->setValidator(v);
	clusterParamLayout->addWidget(editZ3, 3, 3);
	connect(editZ3, SIGNAL(textEdited(const QString&)), this, SLOT(newZ3(const QString&)));
	
	editX4=new QLineEdit; editX4->setValidator(v);
	clusterParamLayout->addWidget(editX4, 1, 4);
	connect(editX4, SIGNAL(textEdited(const QString&)), this, SLOT(newX4(const QString&)));
	
	editY4=new QLineEdit; editY4->setValidator(v);
	clusterParamLayout->addWidget(editY4, 2, 4);
	connect(editY4, SIGNAL(textEdited(const QString&)), this, SLOT(newY4(const QString&)));
	
	editZ4=new QLineEdit; editZ4->setValidator(v);
	clusterParamLayout->addWidget(editZ4, 3, 4);
	connect(editZ4, SIGNAL(textEdited(const QString&)), this, SLOT(newZ4(const QString&)));
	
	buttonCol1=new QPushButton("Change");
	QPalette p1; p1.setColor(QPalette::Active ,QColorGroup::ButtonText, Qt::black); buttonCol1->setPalette(p1);
	clusterParamLayout->addWidget(buttonCol1, 4, 1);
	connect(buttonCol1, SIGNAL(clicked()), this, SLOT(changeColour1()));
	
	buttonCol2=new QPushButton("Change");
	QPalette p2; p2.setColor(QPalette::Active ,QColorGroup::ButtonText, Qt::red); buttonCol2->setPalette(p2);
	clusterParamLayout->addWidget(buttonCol2, 4, 2);
	connect(buttonCol2, SIGNAL(clicked()), this, SLOT(changeColour2()));
	
	buttonCol3=new QPushButton("Change");
	QPalette p3; p3.setColor(QPalette::Active ,QColorGroup::ButtonText, Qt::green); buttonCol3->setPalette(p3);
	clusterParamLayout->addWidget(buttonCol3, 4, 3);
	connect(buttonCol3, SIGNAL(clicked()), this, SLOT(changeColour3()));
	
	buttonCol4=new QPushButton("Change");
	QPalette p4; p4.setColor(QPalette::Active ,QColorGroup::ButtonText, Qt::magenta); buttonCol4->setPalette(p4);
	clusterParamLayout->addWidget(buttonCol4, 4, 4);
	connect(buttonCol4, SIGNAL(clicked()), this, SLOT(changeColour4()));
	
	// Init plplot
	
// 	setCentralWidget(plot);
	plmkstrm(&strm);
	plsetqtdev((void*)plot);
	plsdev("qt");
	plinit();
	
	resetColourIndex();
	
	resize(600, 600);
	
	// Init variables
	
	for(int i=0; i<4; ++i)
	{
		for(int j=0; j<3; ++j)
			clusterParameters[i].position[j]=0.;
	}

	clusterParameters[0].colour=Qt::black;
	clusterParameters[1].colour=Qt::red;
	clusterParameters[2].colour=Qt::green;
	clusterParameters[3].colour=Qt::magenta;
	
	alt=45;
	az=45;
	projection[0]=projection[1]=projection[2]=false;
}
#endif

void QCCWindow::setParameters(QrVector& v1, QrVector& v2, QrVector& v3, QrVector& v4, const QString& title, const QString& frame)
{
	this->title=title;
	this->frame=frame;
	
	editTitle->setText(title);
	editAxis->setText(frame);
	
	for(int i=0; i<3; i++)
	{
		clusterParameters[0].position[i]=v1.ptr->elt(0)[i];
		clusterParameters[1].position[i]=v2.ptr->elt(0)[i];
		clusterParameters[2].position[i]=v3.ptr->elt(0)[i];
		clusterParameters[3].position[i]=v4.ptr->elt(0)[i];
	}
	
	char text[15];
	sprintf(text, "%.2f", clusterParameters[0].position[0]);
	editX1->setText(QString(text));
	sprintf(text, "%.2f", clusterParameters[0].position[1]);
	editY1->setText(QString(text));
	sprintf(text, "%.2f", clusterParameters[0].position[2]);
	editZ1->setText(QString(text));
	
	sprintf(text, "%.2f", clusterParameters[1].position[0]);
	editX2->setText(QString(text));
	sprintf(text, "%.2f", clusterParameters[1].position[1]);
	editY2->setText(QString(text));
	sprintf(text, "%.2f", clusterParameters[1].position[2]);
	editZ2->setText(QString(text));
	
	sprintf(text, "%.2f", clusterParameters[2].position[0]);
	editX3->setText(QString(text));
	sprintf(text, "%.2f", clusterParameters[2].position[1]);
	editY3->setText(QString(text));
	sprintf(text, "%.2f", clusterParameters[2].position[2]);
	editZ3->setText(QString(text));
	
	sprintf(text, "%.2f", clusterParameters[3].position[0]);
	editX4->setText(QString(text));
	sprintf(text, "%.2f", clusterParameters[3].position[1]);
	editY4->setText(QString(text));
	sprintf(text, "%.2f", clusterParameters[3].position[2]);
	editZ4->setText(QString(text));

	replot();
	
	plot->update();
	
}

void QCCWindow::altChanged(int newAlt)
{
	this->alt=(double)newAlt;
	replot();
	plot->update();
}

void QCCWindow::azChanged(int newAz)
{
	this->az=(double)newAz;
	replot();
	plot->update();
}

void QCCWindow::projXChanged(int state)
{
	projection[0]=(state!=0);
	replot();
	plot->update();
}

void QCCWindow::projYChanged(int state)
{
	projection[1]=(state!=0);
	replot();
	plot->update();
}

void QCCWindow::projZChanged(int state)
{
	projection[2]=(state!=0);
	replot();
	plot->update();
}

void QCCWindow::newX1(const QString& s)
{
	clusterParameters[0].position[0]=s.toDouble();
	replot();
	plot->update();
}

void QCCWindow::newY1(const QString& s)
{
	clusterParameters[0].position[1]=s.toDouble();
	replot();
	plot->update();
}

void QCCWindow::newZ1(const QString& s)
{
	clusterParameters[0].position[2]=s.toDouble();
	replot();
	plot->update();
}

void QCCWindow::newX2(const QString& s)
{
	clusterParameters[1].position[0]=s.toDouble();
	replot();
	plot->update();
}

void QCCWindow::newY2(const QString& s)
{
	clusterParameters[1].position[1]=s.toDouble();
	replot();
	plot->update();
}

void QCCWindow::newZ2(const QString& s)
{
	clusterParameters[1].position[2]=s.toDouble();
	replot();
	plot->update();
}

void QCCWindow::newX3(const QString& s)
{
	clusterParameters[2].position[0]=s.toDouble();
	replot();
	plot->update();
}

void QCCWindow::newY3(const QString& s)
{
	clusterParameters[2].position[1]=s.toDouble();
	replot();
	plot->update();
}

void QCCWindow::newZ3(const QString& s)
{
	clusterParameters[2].position[2]=s.toDouble();
	replot();
	plot->update();
}

void QCCWindow::newX4(const QString& s)
{
	clusterParameters[3].position[0]=s.toDouble();
	replot();
	plot->update();
}

void QCCWindow::newY4(const QString& s)
{
	clusterParameters[3].position[1]=s.toDouble();
	replot();
	plot->update();
}

void QCCWindow::newZ4(const QString& s)
{
	clusterParameters[3].position[2]=s.toDouble();
	replot();
	plot->update();
}

void QCCWindow::newTitle(const QString& s)
{
	title=s;
	replot();
	plot->update();
}

void QCCWindow::newFrame(const QString& s)
{
	frame=s;
	replot();
	plot->update();
}

void QCCWindow::changeColour1()
{
	QPalette pal=buttonCol1->palette();
	clusterParameters[0].colour=QColorDialog::getColor(clusterParameters[0].colour, this);
	pal.setColor(QPalette::Active ,QColorGroup::ButtonText, clusterParameters[0].colour);
	buttonCol1->setPalette(pal);
	replot();
	plot->update();
}

void QCCWindow::changeColour2()
{
	QPalette pal=buttonCol2->palette();
	clusterParameters[1].colour=QColorDialog::getColor(clusterParameters[1].colour, this);
	pal.setColor(QPalette::Active ,QColorGroup::ButtonText, clusterParameters[1].colour);
	buttonCol2->setPalette(pal);
	replot();
	plot->update();
}

void QCCWindow::changeColour3()
{
	QPalette pal=buttonCol3->palette();
	clusterParameters[2].colour=QColorDialog::getColor(clusterParameters[2].colour, this);
	pal.setColor(QPalette::Active ,QColorGroup::ButtonText, clusterParameters[2].colour);
	buttonCol3->setPalette(pal);
	replot();
	plot->update();
}

void QCCWindow::changeColour4()
{
	QPalette pal=buttonCol4->palette();
	clusterParameters[3].colour=QColorDialog::getColor(clusterParameters[3].colour, this);
	pal.setColor(QPalette::Active ,QColorGroup::ButtonText, clusterParameters[3].colour);
	buttonCol4->setPalette(pal);
	replot();
	plot->update();
}

void QCCWindow::changeColourBg()
{
	QPalette pal=buttonBgColour->palette();
	bgColour=QColorDialog::getColor(bgColour, this);
	pal.setColor(QPalette::Active ,QColorGroup::ButtonText, bgColour);
	buttonBgColour->setPalette(pal);
	replot();
	plot->update();
}

void QCCWindow::replot()
{
	PLINT old_strm;
	
	PLFLT min_x, max_x, min_y, max_y, min_z, max_z;
	
	PLFLT t_x[]={clusterParameters[0].position[0],
			clusterParameters[1].position[0],
			clusterParameters[2].position[0],
			clusterParameters[3].position[0]};
	PLFLT t_y[]={clusterParameters[0].position[1],
			clusterParameters[1].position[1],
			clusterParameters[2].position[1],
			clusterParameters[3].position[1]};
	PLFLT t_z[]={clusterParameters[0].position[2],
			clusterParameters[1].position[2],
			clusterParameters[2].position[2],
			clusterParameters[3].position[2]};

	min_x=*(min_element(t_x, t_x+sizeof(t_x)/sizeof(t_x[0])));
	max_x=*(max_element(t_x, t_x+sizeof(t_x)/sizeof(t_x[0])));
	min_y=*(min_element(t_y, t_y+sizeof(t_y)/sizeof(t_y[0])));
	max_y=*(max_element(t_y, t_y+sizeof(t_y)/sizeof(t_y[0])));
	min_z=*(min_element(t_z, t_z+sizeof(t_z)/sizeof(t_z[0])));
	max_z=*(max_element(t_z, t_z+sizeof(t_z)/sizeof(t_z[0])));
	
	if(min_x>=max_x || min_y>=max_y || min_z>=max_z) return;

	double length=max_x-min_x; min_x-=0.1*length; max_x+=0.1*length;
	length=max_y-min_y; min_y-=0.1*length; max_y+=0.1*length;
	length=max_z-min_z; min_z-=0.1*length; max_z+=0.1*length;

	plgstrm(&old_strm);
	plsstrm(strm);
	pladv(0);

	plvpor(0., 1., 0., 1.);
	plwind(-1., 0.8, -0.8, 1.8);

	plw3d(1., 1., 1.,
		min_x, max_x,
		min_y, max_y,
		min_z, max_z,
	     alt, az);

	PLFLT cubex[10], cubey[10], cubez[10];

	plcol0(getColourIndex(bgColour));
	// Plot facet x=min_x
	cubex[0]=cubex[1]=cubex[2]=cubex[3]=min_x;
	cubey[0]=cubey[1]=min_y; cubey[2]=cubey[3]=max_y;
	cubez[0]=cubez[3]=min_z; cubez[1]=cubez[2]=max_z;
	plfill3(4, cubex, cubey, cubez);
	
	// Plot facet x=max_x
	cubex[0]=cubex[1]=cubex[2]=cubex[3]=max_x;
	plfill3(4, cubex, cubey, cubez);
	
	// Plot facet y=min_y
	cubey[0]=cubey[1]=cubey[2]=cubey[3]=min_y;
	cubex[0]=cubex[1]=min_x; cubex[2]=cubex[3]=max_x;
	cubez[0]=cubez[3]=min_z; cubez[1]=cubez[2]=max_z;
	plfill3(4, cubex, cubey, cubez);
	
	// Plot facet y=max_y
	cubey[0]=cubey[1]=cubey[2]=cubey[3]=max_y;
	plfill3(4, cubex, cubey, cubez);
	
	// Plot facet z=min_z
	cubez[0]=cubez[1]=cubez[2]=cubez[3]=min_z;
	cubex[0]=cubex[1]=min_x; cubex[2]=cubex[3]=max_x;
	cubey[0]=cubey[3]=min_y; cubey[1]=cubey[2]=max_y;
	plfill3(4, cubex, cubey, cubez);
	
	// Plot facet z=max_z
	cubez[0]=cubez[1]=cubez[2]=cubez[3]=max_z;
	plfill3(4, cubex, cubey, cubez);
	
	plcol0(0);
	
	//draw box contour
	pllsty(2);
	
	cubex[0]=cubex[3]=cubex[4]=cubex[5]=cubex[6]=cubex[9]=min_x;
	cubex[1]=cubex[2]=cubex[7]=cubex[8]=max_x;
	cubey[0]=cubey[1]=cubey[4]=cubey[5]=cubey[8]=cubey[9]=min_y;
	cubey[2]=cubey[3]=cubey[6]=cubey[7]=max_y;
	cubez[0]=cubez[1]=cubez[2]=cubez[3]=cubez[4]=min_z;
	cubez[5]=cubez[6]=cubez[7]=cubez[8]=cubez[9]=max_z;
	plline3(10, cubex, cubey, cubez);
	
	cubex[0]=cubex[1]=min_x;
	cubey[0]=cubey[1]=max_y;
	cubez[0]=min_z; cubez[1]=max_z;
	plline3(2, cubex, cubey, cubez);
	
	cubex[0]=cubex[1]=max_x;
	plline3(2, cubex, cubey, cubez);
	
	cubey[0]=cubey[1]=min_y;
	plline3(2, cubex, cubey, cubez);
	
	// plot axes
	pllsty(1);
	
	plbox3("bnstu", (QString("x ")+frame).latin1(), 0, 0,
	       "bnstu", (QString("y ")+frame).latin1(), 0, 0,
			 "bnstu", (QString("z ")+frame).latin1(), 0, 0);
	
	for(int i=0; i<4; i++)
	{
		double x=(projection[0]?min_x:clusterParameters[i].position[0]);
		double y=(projection[1]?min_y:clusterParameters[i].position[1]);
		double z=(projection[2]?min_z:clusterParameters[i].position[2]);
		
		plssym(0.,4.);
		plcol0(getColourIndex(clusterParameters[i].colour));
		plpoin3(1, &x, &y, &z, 17);
// 		plptex3(x, y, z, 1., 0., 0., 0., 0., 0., 0.5, "X");
// 		plcol0(0);
		plwid(QtWidth2PlWidth(2));
		plssym(0.,1.);
		plcol0(0);
		plpoin3(1, &x, &y, &z, '1'+i);
		plwid(QtWidth2PlWidth(1));
	}
	
	// draw tetrahedron
	plcol0(0);
	
	cubex[0]=(projection[0]?min_x:clusterParameters[0].position[0]);
	cubey[0]=(projection[1]?min_y:clusterParameters[0].position[1]);
	cubez[0]=(projection[2]?min_z:clusterParameters[0].position[2]);
	
	cubex[1]=(projection[0]?min_x:clusterParameters[1].position[0]);
	cubey[1]=(projection[1]?min_y:clusterParameters[1].position[1]);
	cubez[1]=(projection[2]?min_z:clusterParameters[1].position[2]);
	
	cubex[2]=(projection[0]?min_x:clusterParameters[2].position[0]);
	cubey[2]=(projection[1]?min_y:clusterParameters[2].position[1]);
	cubez[2]=(projection[2]?min_z:clusterParameters[2].position[2]);
	
	cubex[3]=(projection[0]?min_x:clusterParameters[0].position[0]);
	cubey[3]=(projection[1]?min_y:clusterParameters[0].position[1]);
	cubez[3]=(projection[2]?min_z:clusterParameters[0].position[2]);
	
	cubex[4]=(projection[0]?min_x:clusterParameters[3].position[0]);
	cubey[4]=(projection[1]?min_y:clusterParameters[3].position[1]);
	cubez[4]=(projection[2]?min_z:clusterParameters[3].position[2]);
	
	cubex[5]=(projection[0]?min_x:clusterParameters[2].position[0]);
	cubey[5]=(projection[1]?min_y:clusterParameters[2].position[1]);
	cubez[5]=(projection[2]?min_z:clusterParameters[2].position[2]);
	
	plline3(6, cubex, cubey, cubez);
	
	cubex[0]=(projection[0]?min_x:clusterParameters[1].position[0]);
	cubey[0]=(projection[1]?min_y:clusterParameters[1].position[1]);
	cubez[0]=(projection[2]?min_z:clusterParameters[1].position[2]);
	
	cubex[1]=(projection[0]?min_x:clusterParameters[3].position[0]);
	cubey[1]=(projection[1]?min_y:clusterParameters[3].position[1]);
	cubez[1]=(projection[2]?min_z:clusterParameters[3].position[2]);
	
	plline3(2, cubex, cubey, cubez);
	
	// add title
	
	//TODO
	
	// add phi and theta
	
	plschr(130., 1.);
	char text[40];
	sprintf(text, "#gh=%d", (int)(90.-alt));
	plmtex("rv", -1., 0.05, 1., text);
	sprintf(text, "#gf=%d", (int)az);
	plmtex("rv", -1., 0.08, 1., text);
	
	plschr(130., 1.5);
	plmtex("t", -1., 0.5, 0.5, title.latin1());
	
	plsstrm(old_strm);
}

void QCCWindow::resetColourIndex()
{
	PLINT iR,iG,iB;
	
	userR.clear();
	userG.clear();
	userB.clear();
	for(register int i=0;i<16;++i)
	{
		plgcol0(i,&iR,&iG,&iB);
		userR.push_back(iR);
		userG.push_back(iG);
		userB.push_back(iB);
	}
}

int QCCWindow::getColourIndex(const QColor& c)
{
	// find PGPlot colour index closest to this colour
	int R=c.red();
	int G=c.green();
	int B=c.blue();
	int min_diff=200000; // bigger than 255*255*3
	//int colIndex=0;
		
	int indexNearest=0;
		
	for(register int i=0;i<userR.size();i++)
	{
		int diff=(R-userR[i])*(R-userR[i]) + (G-userG[i])*(G-userG[i]) + (B-userB[i])*(B-userB[i]);
		if(diff==0)
		{
			return i; // found in existing table
		}
		else if(diff<min_diff)
		{
			min_diff = diff;
			indexNearest=i;
		}
	}
		
	if( userR.size() < 40)
	{
		// create new colour and set in local copy
		
		userR.push_back(R);
		userG.push_back(G);
		userB.push_back(B);
	}
	else
	{
		// use nearest colour
		return indexNearest;
	}
	
	// set up colour table as new colour added
	plscmap0n(userR.size());
	plscol0(userR.size()-1, R, G, B);
	
	return userR.size()-1;
}

// void QCCWindow::savePlot()
// {
// 
// #if QT_VERSION < 0x040000
// 	QString saveName=QFileDialog::getSaveFileName(QString::null, QString("Image file (*.png *.jpg *.jpeg *.tif *.tiff *.ps *.eps)"), this);
// #else
// #if QT_VERSION < 0x040300 // svg not available
// 	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", QString(), "Image file (*.png *.jpg *.jpeg *.tif *.tiff *.ps *.eps)");
// #else
// 	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", QString(), "Image file (*.png *.jpg *.jpeg *.tif *.tiff *.ps *.eps *.svg)");
// #endif
// #endif
//         
// 	if(saveName.isEmpty()) return;
// #if QT_VERSION < 0x040000
// #define CASE_INSENSITIVE FALSE
// #else
// #define CASE_INSENSITIVE Qt::CaseInsensitive
// #endif
// 	char format[5];
// 	if(saveName.contains("jpg", CASE_INSENSITIVE) || saveName.contains("jpeg", CASE_INSENSITIVE))
// 		sprintf(format, "%s","JPEG");
// 	else
// 	{
// 		if(saveName.contains("png", CASE_INSENSITIVE))
// 		{
// 			sprintf(format, "%s","PNG");
// 		}
// 		else
// 		{
// 			if(saveName.contains("tiff", CASE_INSENSITIVE) || saveName.contains("tif", CASE_INSENSITIVE))
// 			{
// 				sprintf(format, "%s","TIFF");
// 			}
// 			else
// 			{
// #if QT_VERSION >= 0x040300
// 				if(saveName.contains("svg", CASE_INSENSITIVE))
// 				{
// 					sprintf(format, "%s", "SVG");
// 				}
// 				else
// #endif
// 					sprintf(format, "%s","PS");
// 			}
// 		}
// 	}
//         
// 	if(strcmp(format, "PS")==0)
// 	{
// 		QPrinter p(QPrinter::HighResolution);
// 		p.setOutputFileName(saveName);
// 		p.setColorMode(QPrinter::Color);
// 		p.setOrientation(QPrinter::Landscape);
// #if QT_VERSION < 0x040400
// 		p.setPageSize(QPrinter::A4);
// #else
// 		p.setPaperSize(QPrinter::A4);
// #endif
// 
// 		plot->print(p);
// 	}
// 	else
// 	{
// #if QT_VERSION >= 0x040300
// 		if(strcmp(format, "SVG")==0)
// 		{
// 			plot->saveSVG(saveName, 600, 600);
// 		}
// 		else
// #endif
// 			plot->savePixmap(saveName, format, 600, 600, 85);
// 	}
// }

void QCCWindow::savePlotPNG()
{
#if QT_VERSION < 0x040000
	QString saveName=QFileDialog::getSaveFileName(QString::null, QString("PNG file (*.png)"), this);
#else
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", QString(), "Image file (*.png)");
#endif
	if(saveName.isEmpty()) return;
	if(!saveName.endsWith(".png", CASE_INSENSITIVE))
	{
		saveName.append(".png");
	}
	plot->savePixmap(saveName, "PNG", 600, 600, 85);
}

void QCCWindow::savePlotJPG()
{
#if QT_VERSION < 0x040000
	QString saveName=QFileDialog::getSaveFileName(QString::null, QString("JPG file (*.jpg *.jpeg)"), this);
#else
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", QString(), "Image file (*.jpg *.jpeg)");
#endif
	if(saveName.isEmpty()) return;
	if(!saveName.endsWith(".jpg", CASE_INSENSITIVE) && !saveName.endsWith(".jpeg", CASE_INSENSITIVE))
	{
		saveName.append(".jpg");
	}
	plot->savePixmap(saveName, "JPEG", 600, 600, 85);
}

void QCCWindow::savePlotPS()
{
#if QT_VERSION < 0x040000
	QString saveName=QFileDialog::getSaveFileName(QString::null, QString("Encapsulated Postscript file (*.eps)"), this);
#else
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", QString(), "Encapsulated Postscript file (*.eps)");
#endif
	if(saveName.isEmpty()) return;
	
	if(!saveName.endsWith(".ps", CASE_INSENSITIVE) && !saveName.endsWith(".eps", CASE_INSENSITIVE))
	{
		saveName.append(".eps");
	}
	
	QPrinter p;
	p.setResolution(150);
	p.setOutputFileName(saveName);
	p.setOrientation(QPrinter::Landscape);
	p.setColorMode(QPrinter::Color);
	
#if QT_VERSION >= 0x040000
	p.setOutputFormat(QPrinter::PostScriptFormat);
#endif
	
#if QT_VERSION < 0x040400
	p.setPageSize(QPrinter::A5);
#else
	p.setPaperSize(QPrinter::A5);
#endif

	plot->print(p);
}

void QCCWindow::savePlotPDF()
{
#if QT_VERSION < 0x040000
	QString saveName=QFileDialog::getSaveFileName(QString::null, QString("PDF document (*.pdf)"), this);
#else
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", QString(), "PDF document (*.pdf)");
#endif
	if(saveName.isEmpty()) return;
	
	if(!saveName.endsWith(".pdf", CASE_INSENSITIVE))
	{
		saveName.append(".pdf");
	}
	
	QPrinter p;
	p.setResolution(150);
	p.setOutputFileName(saveName);
	p.setOrientation(QPrinter::Landscape);
	p.setColorMode(QPrinter::Color);

#if QT_VERSION >= 0x040000
	p.setOutputFormat(QPrinter::PdfFormat);
#endif
	
#if QT_VERSION < 0x040400
	p.setPageSize(QPrinter::A5);
#else
	p.setPaperSize(QPrinter::A5);
#endif

	plot->print(p);
}


void QCCWindow::savePlotSVG()
{
#if QT_VERSION >= 0x040300
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", QString(), "SVG file (*.svg)");
	if(saveName.isEmpty()) return;
	if(!saveName.endsWith(".svg", CASE_INSENSITIVE))
	{
		saveName.append(".svg");
	}
	plot->saveSVG(saveName, 600, 600);
#endif
}

void QCCWindow::printPlot()
{
	QPrinter printer;
	printer.setResolution(150);
	printer.setPageSize(QPrinter::A5);
	printer.setOrientation(QPrinter::Landscape);
	printer.setColorMode(QPrinter::Color);
	QPrintDialog printDialog(&printer, this);
	if (printDialog.exec() == QDialog::Accepted)
	{
		plot->print(printer);
	}
}


