#include "PlotWindow.h"

#include <cmath>

#if QT_VERSION < 0x040000
#include <qmessagebox.h>
#define CASE_INSENSITIVE FALSE

std::list<PlotWindow*> GlobalPlotWindows;
#else
#include <QMessageBox>
#define CASE_INSENSITIVE Qt::CaseInsensitive

QList<PlotWindow*> GlobalPlotWindows;
#endif
	 
PlotWindow::PlotWindow(int id, int width, int height, double saveWidth, double saveHeight, int resolution, QPrinter::PageSize pageSize, QPrinter::Orientation orientation, QWidget* parent)
	:QMainWindow(parent)
{
   
#if QT_VERSION < 0x040000
	QPopupMenu * plotMenu = new QPopupMenu( this );
#else
	QMenu * plotMenu = new QMenu( this );
#endif
	menuBar()->insertItem("Plot", plotMenu);
	plotMenu->insertItem("Save Plot As PNG", this, SLOT(savePlotPNG()));
	plotMenu->insertItem("Save Plot As JPG", this, SLOT(savePlotJPG()));
#if QT_VERSION >= 0x040300
	plotMenu->insertItem("Save Plot As SVG", this, SLOT(savePlotSVG()));
#endif
#if defined(Q_OS_MACX) || QT_VERSION >= 0x040000
	plotMenu->insertItem("Save Plot As PDF", this, SLOT(savePlotPDF()));
#endif
#if !(defined(Q_OS_MACX) && QT_VERSION < 0x040000)
	plotMenu->insertItem("Save Plot As EPS", this, SLOT(savePlotPS()));
#endif
	plotMenu->insertItem("Print", this, SLOT(printPlot()));
	
	this->saveWidth=saveWidth;
	this->saveHeight=saveHeight;
	this->pageSize=pageSize;
	this->orientation=orientation;
	this->resolution=resolution;
	
	plot=new QtPLWidget(width, height, saveWidth, saveHeight, this);
	setCentralWidget(plot);
	connect(plot, SIGNAL(resized(double)), this, SLOT(changeDisplayedRatio(double)));
        resize(width, height);
        setId(id);
        plmkstrm(&strm);
        plsetqtdev((void*)plot);
        plsdev ("qt");
        plinit();
	
	connect(plot, SIGNAL(clicked(double, double, int, int)), this, SIGNAL(clicked(double, double, int, int))); // redirection of widget "clicked" events
	
	intervalDialog=new IntervalDialog(this);
	intervalDialog->hide();
#if QT_VERSION < 0x040000
#else
	intervalDialog->setVisible(false);
#endif
	
	connect(intervalDialog, SIGNAL(rescale(double, double)), this, SIGNAL(rescale(double, double)));
}

void PlotWindow::setPLFocus()
{
    plsstrm(strm);
}

void PlotWindow::setTitle(const QString& title)
{
	this->title=title;
	
#if QT_VERSION < 0x040000
	setCaption(title);
#else
	setWindowTitle(title);
#endif
}

void PlotWindow::changeDisplayedRatio(double r)
{
#if QT_VERSION < 0x040000
	setCaption(QString("%1 (%2%)").arg(title).arg((int)(r*100.)));
#else
	setWindowTitle(QString("%1 (%2%)").arg(title).arg((int)(r*100.)));
#endif
}

void PlotWindow::save(){std::cout << "TODO PlotWindow: save" << std::endl;}
                        
void PlotWindow::unsave(){std::cout << "TODO PlotWindow: unsave" << std::endl;}
                        
void PlotWindow::setSmooth(bool b)
{
    if(plot) plot->setSmooth(b);
}

void PlotWindow::setId(int id){this->id=id;}
                        
int PlotWindow::getId(){return id;}

void PlotWindow::setSize(double width, double height)
{
	bool mustResize=(width!=saveWidth||height!=saveHeight);
    saveWidth=width;
    saveHeight=height;
    PLINT old_strm;
    plgstrm(&old_strm);
    plsstrm(strm);
    plot->setNewSaveSize(width, height);
    plsstrm(old_strm);
    if(mustResize)
    {
    	resize((width<1000?(int)floor(width+0.5):1000), (height<700?(int)(height+0.5):700));
    }
}

void PlotWindow::savePlotPNG()
{
#if QT_VERSION < 0x040000
	QString saveName=QFileDialog::getSaveFileName(QString::null, QString("PNG file (*.png)"), this);
#else
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", QString(), "PNG file (*.png)");
#endif
	if(saveName.isEmpty()) return;
	if(!saveName.endsWith(".png", CASE_INSENSITIVE))
	{
		saveName.append(".png");
	}
	plot->savePixmap(saveName, "PNG", (int)saveWidth, (int)saveHeight, 85);
}

void PlotWindow::savePlotJPG()
{
#if QT_VERSION < 0x040000
	QString saveName=QFileDialog::getSaveFileName(QString::null, QString("JPG file (*.jpg *.jpeg)"), this);
#else
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", QString(), "JPG file (*.jpg *.jpeg)");
#endif
	if(saveName.isEmpty()) return;
	if(!saveName.endsWith(".jpg", CASE_INSENSITIVE) && !saveName.endsWith(".jpeg", CASE_INSENSITIVE))
	{
		saveName.append(".jpg");
	}
	plot->savePixmap(saveName, "JPEG", (int)saveWidth, (int)saveHeight, 85);
}

void PlotWindow::savePlotPS()
{
#if QT_VERSION < 0x040000
	QString saveName=QFileDialog::getSaveFileName(QString::null, QString("Encapsulated Postscript file (*.eps)"), this);
#else
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", QString(), "Encapsulated Postscript file (*.eps)");
#endif
	if(saveName.isEmpty()) return;
	
	if(!saveName.endsWith(".ps", CASE_INSENSITIVE) && !saveName.endsWith(".eps", CASE_INSENSITIVE))
	{
		saveName.append(".eps");
	}
	
	QPrinter p;
	p.setResolution(resolution);
	p.setOutputFileName(saveName);
	p.setOrientation(orientation);
	p.setColorMode(QPrinter::Color);
	
#if QT_VERSION >= 0x040000
	p.setOutputFormat(QPrinter::PostScriptFormat);
#endif
	
#if QT_VERSION < 0x040400
	p.setPageSize(pageSize);
#else
	p.setPaperSize(pageSize);
#endif

	plot->print(p);
}

void PlotWindow::savePlotPDF()
{
#if QT_VERSION < 0x040000
	QString saveName=QFileDialog::getSaveFileName(QString::null, QString("PDF document (*.pdf)"), this);
#else
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", QString(), "PDF document (*.pdf)");
#endif
	if(saveName.isEmpty()) return;
	
	if(!saveName.endsWith(".pdf", CASE_INSENSITIVE))
	{
		saveName.append(".pdf");
	}
	
	QPrinter p;
	p.setResolution(resolution);
	p.setOutputFileName(saveName);
	p.setOrientation(orientation);
	p.setColorMode(QPrinter::Color);

#if QT_VERSION >= 0x040000
	p.setOutputFormat(QPrinter::PdfFormat);
#endif
	
#if QT_VERSION < 0x040400
	p.setPageSize(pageSize);
#else
	p.setPaperSize(pageSize);
#endif

	plot->print(p);
}


void PlotWindow::savePlotSVG()
{
#if QT_VERSION >= 0x040300
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...", QString(), "SVG file (*.svg)");
	if(saveName.isEmpty()) return;
	if(!saveName.endsWith(".svg", CASE_INSENSITIVE))
	{
		saveName.append(".svg");
	}
	plot->saveSVG(saveName, (int)saveWidth, (int)saveHeight, resolution);
#endif
}

void PlotWindow::printPlot()
{
	QPrinter printer;
	printer.setResolution(resolution);
	printer.setPageSize(pageSize);
	printer.setOrientation(orientation);
	printer.setColorMode(QPrinter::Color);
	QPrintDialog printDialog(&printer, this);
#if !defined(WINDOWS) && !defined(Q_OS_MACX)
	if(QSAS_profiles->current()->UseKPrinter)
	{
		printer.setPrintProgram(QString("kprinter"));
		printer.setPrinterName("foo"); // Requires a name. Ideally a printer name, but anything is fine in practice.
		plot->print(printer);
	}
	else
	{
		if (printDialog.exec() == QDialog::Accepted)
		{
			printer.setPrintProgram(QString("lpr"));
			plot->print(printer);
		}
	}
#else
	if (printDialog.exec() == QDialog::Accepted)
	{
		plot->print(printer);
	}
#endif

	
}

PlotWindow* newPlotWindow(double width, double height, int* id, int resolution, QPrinter::PageSize pageSize, QPrinter::Orientation orientation)
{
	// find free id
	
	bool found_id=false;
	bool this_one_ok;
	*id=1;
	while(!found_id)
	{
		this_one_ok=true;
#if QT_VERSION < 0x040000
		for(std::list<PlotWindow*>::iterator i=GlobalPlotWindows.begin(); i!=GlobalPlotWindows.end() && this_one_ok; ++i)
#else
		for(QList<PlotWindow*>::iterator i=GlobalPlotWindows.begin(); i!=GlobalPlotWindows.end() && this_one_ok; ++i)
#endif
		{
			this_one_ok=this_one_ok && ((*i)->getId()!=*id);
		}
		if(this_one_ok)
		{
			found_id=true;
		}
		else
		{
			(*id)++;
		}
	}

	PlotWindow* res=new PlotWindow(*id, (width<1000?(int)width: 1000), (height<700?(int)height:700), width, height, resolution, pageSize, orientation, NULL);
#if QT_VERSION < 0x040000
	GlobalPlotWindows.push_back(res);
#else
	GlobalPlotWindows.append(res);
#endif
	res->setTitle(QString("QSAS Plot Window #%1").arg(*id));
	res->show();
	return res;
}

PlotWindow* focusPlotWindow(int id, double width, double height, int resolution, QPrinter::PageSize pageSize, QPrinter::Orientation orientation)
{
	PlotWindow * res;
#if QT_VERSION < 0x040000
	for(std::list<PlotWindow*>::iterator i=GlobalPlotWindows.begin(); i!=GlobalPlotWindows.end(); ++i)
#else
		for(QList<PlotWindow*>::iterator i=GlobalPlotWindows.begin(); i!=GlobalPlotWindows.end(); ++i)
#endif
	{
		if((*i)->getId()==id)
		{
			(*i)->show();
			if(width>0. && height > 0.)
			{
				(*i)->setSize(width, height);
				(*i)->resolution=resolution;
				(*i)->pageSize=pageSize;
				(*i)->orientation=orientation;
			}
			(*i)->setPLFocus();
			return (*i);
		}
	}
	
	// If reached this point, the window does not exist.
	// Force creation of Window with this id
	
	res=new PlotWindow(id, (width<1000?(int)width: 1000), (height<700?(int)height:700), width, height, resolution, pageSize, orientation, NULL);
#if QT_VERSION < 0x040000
	GlobalPlotWindows.push_back(res);
	//res->setCaption(QString("QSAS Plot Window #%1").arg(id));
#else
	GlobalPlotWindows.append(res);
	//res->setWindowTitle(QString("QSAS Plot Window #%1").arg(id));
#endif
	res->setTitle(QString("QSAS Plot Window #%1").arg(id));
	res->show();
	return res;
}

void clearPlotWindows()
{
#if QT_VERSION < 0x040000
    for(std::list<PlotWindow*>::iterator i=GlobalPlotWindows.begin(); i!=GlobalPlotWindows.end(); ++i)
#else
	    for(QList<PlotWindow*>::iterator i=GlobalPlotWindows.begin(); i!=GlobalPlotWindows.end(); ++i)
#endif
    {
        PlotWindow * temp=*i;
	temp->close();
        delete temp;
    }
    GlobalPlotWindows.clear();
}

