// floatspinbox.h
//
// Floating Point Spinner Class for Qt - Interface File
//
//

#ifndef FLOATSPINBOX_H_
#define FLOATSPINBOX_H_

#include <qobject.h>
#include <qspinbox.h>
#include <qvalidator.h>
#include <qevent.h>

class FloatSpinBox : public QSpinBox
{
	Q_OBJECT
	public:
		FloatSpinBox( QWidget* parent = 0, const char* name = 0 );
		FloatSpinBox( float minValue, float maxValue, int decimals = 2, float step = 1.0,
			      QWidget* parent = 0, const char* name = 0 );
		~FloatSpinBox();

		virtual QSize sizeHint () const;

		float value() const;
		float minValue() const;
		float maxValue() const;

		void setRange( double minValue, double maxValue );

	public slots:
		void setValue( double value );

	signals:
		void valueChanged( double value );
		void focusOut();
		void returnPressed();

	protected:
		virtual QString mapValueToText ( int value );
		virtual int mapTextToValue ( bool * ok = 0 );
		virtual void valueChange ();
		virtual void keyPressEvent(QKeyEvent*);
  
		bool eventFilter(QObject*,QEvent*);

	private:
		QDoubleValidator* dVal;

		int dec;
};

#endif  //  FLOATSPINBOX_H_

