
#include "FARdb.h"

#include <fstream>

#include <string.h>   // for C string functions
#include <stdlib.h>   // for C atof functions
#include <math.h>     // for C fabs functions

namespace FAR {

void Record::diag_print(void) const
{
  int i;
  vector<Field>::const_iterator p = fields.begin();
  
  for( i=0; p!=fields.end(); i++,p++ )
  {
    cout << "  field " << i <<": <" << p->c_str() << ">" << endl;
  }
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void FARData::diag_print(void) const
{
  int i, j;
  vector<Record>::const_iterator rp = records.begin();
  
  for( i=0; rp!=records.end(); i++,rp++ )
  {
    cout << "R:" << i << " ";
    for( j=0; j<records[i].n_fields(); j++ )
      cout << "<" << records[i][j] << "> ";
    cout << endl;
  }
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


/*void Record::add_fields( vector<string>& ff )
{
  vector<string>::const_iterator p=ff.begin();
  for( ; p!=ff.end(); p++ )
    fields.push_back( *p ); 
}*/

ostream& operator<<(ostream& os,const Record& rec)
{
  string separator=",";
  for( int ifld=0; ifld<rec.n_fields(); ifld++ )
  {
    
    if( ifld != 0 )
      os << separator.c_str();
    int spos = rec[ifld].find( separator );
    if( spos == string::npos )
      os << rec[ifld].c_str();
    else
      os << "\"" << rec[ifld].c_str() << "\"";
  }
} 

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

bool FARData::is_comment_line( const char* s )
{
//  cout << "check comment: <" << s << ">" << endl;
  return ( strlen(s)>0 && s[0]=='#' );
}


// trim leading and trailing whitespace from a string

/*void FARData::trim_whitespace( string& s )
{
  string::iterator p1 = s.begin();
  string::iterator p2 = s.end();

  while( s.size() > 0 && isspace(s[0]) )
    s.erase(0,1);
  while( s.size() > 0 && isspace(s[s.size()-1]) )
    s.erase(s.size()-1);
}*/
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */


int FARData::xtokenize_string( const char *s, const char *delimiters, vector<string>& tokens )
{
  char *ss, *ss1, *delimiter_pos;
  int i, n_tokens;
  
  vector<char *> tokensp;

  if( strlen(s)==0 )
    return 0;
  ss1 = ss = strdup(s);

  n_tokens=1;
  tokensp.push_back(ss1);

  while( (delimiter_pos = strpbrk( ss1, delimiters ) ) != NULL )
  {
    ss1 = delimiter_pos + 1;
    tokensp.push_back( ss1 );
    *delimiter_pos = '\0';
  }

  n_tokens = tokensp.size();
  for(i=0;i<n_tokens;i++)
  {
    tokens.push_back( tokensp[i] );
    trim_whitespace( tokens[i] );
  }
  free( ss );  
  return n_tokens;
} 
/* ---------------------------------------------------------------------- */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */


/*int FARData::tokenize_string( const char *s, const char *delimiters,
                                  vector<string>& token_list )
{
  string token;
  int s_len = strlen(s);

  if( s_len == 0 )
    return 0;

  enum state_e {
    token_not_started, token_started_normal,
    token_started_quoted, token_ended_quoted } state;

  int ip = -1;
  state = token_not_started;

  while( ip++ < s_len )
    switch( state )
    {
    case token_not_started:
      if( !isspace(s[ip]) )
        if( s[ip] == '"' )
          state = token_started_quoted;
        else if( s[ip] == ',' )          // another comma already!
        {
          trim_whitespace( token );
          token_list.push_back(token);
          token.erase();
          state = token_not_started;
        } else
        {
          state = token_started_normal;
          token += s[ip];
        }
      break;
    
    case token_started_normal:
      if( strchr( delimiters, s[ip] ) != NULL )  // if delimiter ...
      {
        trim_whitespace( token );
        token_list.push_back(token);
        token.erase();
        state = token_not_started;
      } else
        token += s[ip];
      break;

    case token_started_quoted:
      if( s[ip] == '"' )
      {
        ip++;
        if( s[ip] == '"' )  // double quote
        {
          token += '"';
        } else
        {
          ip--;
          trim_whitespace( token );
          token_list.push_back(token);
          token.erase();
          state = token_ended_quoted;
        }
      } else
      {
        token += s[ip];
      }
      break;

    case token_ended_quoted:
     if( strchr( delimiters, s[ip] ) != NULL )  // if delimiter ...
       state = token_not_started;
     break;
    }

  if( state == token_started_normal || state == token_started_quoted )
  {
    trim_whitespace( token );
    token_list.push_back(token);
    token.erase();
  }

} */
/* ---------------------------------------------------------------------- */

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

/*void FARData::write_to_file( string file_name,
    FARWriteFileMode write_mode, string separator )
{

// probably gnu c++ specific!

  ios::openmode open_mode;
  switch( write_mode )
  {
    default:
    case FARWRITEFILE_NEW:
      open_mode = ios::out;
      break;
    case FARWRITEFILE_APPEND:
      open_mode = ios::app;
      break;
  }

  std::ofstream outfile( file_name.c_str(), open_mode );

  if( !outfile )  // baddd!
  {
     cout << " FAR database error: output file could not be opened.\n" ;
     return;
  }

  for(int ir=0; ir<records.size(); ir++ )
  {
    for( int ifld=0; ifld<records[ir].n_fields(); ifld++ )
    {
      if( ifld != 0 )
        outfile << separator.c_str();
      int spos = records[ir][ifld].find( separator );
      if( spos == string::npos )
        outfile << records[ir][ifld].c_str();
      else
        outfile << "\"" << records[ir][ifld].c_str() << "\"";
    }

    outfile << endl;
  }
  outfile.close();
}*/

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
/*void FARData::read_from_file( string file_name, FARReadFileMode read_mode, string separator )
{
  switch( read_mode )
  {
  default:
  case FARREADFILE_REPLACE:
    delete_all_records();
    append_from_file( file_name, separator );
    break;
  case FARREADFILE_APPEND:
    append_from_file( file_name, separator );
    break;
  }
}*/

/* ---------------------------------------------------------------------- */
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
/*void FARData::append_from_file( string file_name, string separator )
{
  const int n_line_buf = 5000;
  char line_buf[n_line_buf];
  std::ifstream infile( file_name.c_str() );

  if( !infile )  // baddd!
  {
     cout << " FAR database error: input file could not be opened.\n" ;
     return;
  }
// readline, eol discarded
  while( !infile.eof() )
  {
    infile.getline( line_buf, n_line_buf );

// Apr 14, 1999:
//    decided to drop idea of comment lines! because of wanting
//    to preserve them when writing out a database to a file
//   
//    if( is_comment_line( line_buf ) )
//      continue;

// create a new empty record 
   records.push_back( Record() );

//   cout << line_buf << endl;
     
     vector<string> tokens;
     int n_tokens = tokenize_string( line_buf, ",", tokens );
//     for( int i=0; i < tokens.size(); i++ )
//       cout << "token "<<i<<": <"<<tokens[i].c_str()<<">"<<endl;
     records[records.size()-1].add_fields( tokens );
  }
  infile.close();
}*/
/* ---------------------------------------------------------------------- */

/*int FARData::match_record_by_field( int field, string s, int r_start, int r_end ) const
{
  if( r_end == -1 ) r_end = records.size()-1;
  if( r_end<0 || r_start<0 ) return -1;
  if( r_end >= records.size() || r_start >= records.size()  ) return -1;
  if( r_start > r_end ) return -1;
  if( field < 0 ) return -1;

  bool match_found = false;
  int r;
  for( r=r_start; r <= r_end; r++ )
  {
    if( field >= records[r].n_fields() )
      continue;
    if( records[r][field] == s )
    {
      match_found = true; break;
    }
  }
  return match_found ? r : -1;
}*/
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// Delete specified field from all records in database

void  FARData::delete_field( int f )
{
  vector<Record>::iterator rp = records.begin();
  for( ; rp!=records.end(); rp++ )
    rp->delete_field( f );
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int FARData::match_record_by_field( int field, double x, int r_start, int r_end ) const
{
  if( r_end == -1 ) r_end = records.size()-1;
  if( r_end<0 || r_start<0 ) return -1;
  if( r_end >= records.size() || r_start >= records.size()  ) return -1;
  if( r_start > r_end ) return -1;
  if( field < 0 ) return -1;

  bool match_found = false;
  int r;
  for( r=r_start; r <= r_end; r++ )
  {
    if( field >= records[r].n_fields() )
      continue;
    if( fabs(atof(records[r][field].c_str()) - x) < 1.0e-4 )
    {
      match_found = true; break;
    }
  }
  return match_found ? r : -1;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int cmp_nocase( const string& s1, const string& s2 )
{
  string::const_iterator p1 = s1.begin();
  string::const_iterator p2 = s2.begin();
 

}

} // namespace FAR

